#include "TPhSensorMapDBSensorItem.h"

#include <string.h>

// These are the hardcoded numbers corresponding to the number
// of rows and columns and bondingpattern size in the various sensor 
// types. 
// The convention is: Type 1 - 5, Oct, In VTX, Out VTX, Ring 
const Int_t kRows[] =    {     0,  22,   5,  8,  4,  4,   4,   4,   2, 8 };
const Int_t kColumns[] = {     0,  70, 100, 64, 64, 64,  32, 128, 128, 8 };  
const Int_t kBondPatSize[] = { 0, 128,  64, 64, 64, 64, 128, 128,  64, 64};

ClassImp(TPhSensorMapDBSensorItem)

  /////////////////////////////////////////////////////////////////////////
  //
  // This class stores the (FEC, String, Col) and Bonding Pattern 
  // information for a given Sensor. It is being used by TPhSensorMapDB
  //
  // $Id: TPhSensorMapDBSensorItem.cxx,v 1.4 1999/01/13 14:50:55 decowski Exp $
  //
  /////////////////////////////////////////////////////////////////////////

TPhSensorMapDBSensorItem::TPhSensorMapDBSensorItem()
{
  // Default constructor
  fType = 0;
  fRows = 0;
  fColumns = 0;
  fMap = NULL;
}

TPhSensorMapDBSensorItem::TPhSensorMapDBSensorItem(Char_t *SensName, 
	       				   Int_t Type)
{
  // Normal constructor
  // Set the detector type (1 - 5 in Spectr, 6 = Octagon, 7 = In VTX, 
  // 8 = Out VTX, 9 = RingMult).

  SetName(SensName);
  fType = Type;
  fRows = kRows[Type];
  fColumns = kColumns[Type];
  fBondPatSize = kBondPatSize[Type];

  fMap = new FECMap_t[fRows * fColumns];
  // Clear memory
  memset(fMap, 0, sizeof(FECMap_t) * fRows * fColumns);

}

TPhSensorMapDBSensorItem::~TPhSensorMapDBSensorItem()
{
  // Default destructor

  if(fMap)
    delete [] fMap;
  fRows = 0;
  fColumns = 0;
  fType = 0;
  fBondPatSize = 0;
}

void TPhSensorMapDBSensorItem::Store(Int_t Row, Int_t Col, 
	     Int_t FEC, Int_t String, Int_t ChPos, Int_t BondPat)
{
  // Store the (FEC, String, ChPos) + BondPattern information 
  // 
  // This routine is called when the table is initially populated
  // 
  // The following convention is followed for "special pads" in
  // Type 2 and the Octagon:
  //       (Row, Column):
  //       (-1, -1)  unconnected
  //       (-1, -2)  connected through a capacitor to the Guard ring
  //       (-1, -3)  connected to a dummy pad

  if(Row > fRows || Row < 0 || Col > fColumns || Col < 0) {
    if(Row == -1 && (Col == -1 || Col == -2 || Col == -3))
      return;
    Error("Store", "%s Row or Col not good! Row = %d ;Col = %d", 
	  GetName(), Row, Col);
  }
  if(FEC > 60 || FEC < -1) {
    Error("Store", "%s FEC not good! FEC = %d", GetName(), FEC);
    return;
  }
  if(String > 7 || String < -1) {
    Error("Store", "%s String not good! String = %d", GetName(), String);
    return;
  }
  if(ChPos > 768 || ChPos < -1) {
    Error("Store", "%s ChPos not good! ChPos = %d", GetName(), ChPos);
    return;
  }

  FECMap_t data;
  data.FEC = Short_t(FEC);
  data.String = Char_t(String);
  data.ChannelPos = Short_t(ChPos);
  data.BondPat = Char_t(BondPat);

  fMap[Index(Row, Col)] = data;

}

Bool_t TPhSensorMapDBSensorItem::Lookup(Int_t Row, Int_t Col, 
	     Int_t *FEC, Int_t *String, Int_t *ChPos, 
	     Int_t *BondPat)
{
  // Lookup  the (FEC, String, Ch) & BondPat information for a given 
  // Row and Column
  //
  // Returns: kTRUE on success and kFALSE on failure. 

  FECMap_t data;

  if(Row >= fRows || Col >= fColumns || Row < 0 || Col < 0)
    return kFALSE;

  data = fMap[Index(Row, Col)];
  *FEC = data.FEC;
  *String = data.String;
  *ChPos = data.ChannelPos;
  *BondPat = data.BondPat;

  return kTRUE;
}

//___________________________________________________________
void TPhSensorMapDBSensorItem::Streamer(TBuffer &R__b)
{
  // Stream an object of class TPhSensorMapDBSensorItem.
  // This function was hand edited.

  Int_t i, NumEntries;

  if (R__b.IsReading()) {
    Version_t R__v = R__b.ReadVersion(); if (R__v) { }
    TNamed::Streamer(R__b);
    R__b >> fRows;
    R__b >> fColumns;
    R__b >> fType;
    R__b >> fBondPatSize;

    // Allocate memory for the FECMap_t struct and read in data
    NumEntries = fRows * fColumns;
    fMap = new FECMap_t [NumEntries];

    for(i = 0; i < NumEntries; i++) {
      R__b >> fMap[i].FEC;
      R__b >> fMap[i].String;
      R__b >> fMap[i].ChannelPos;
      R__b >> fMap[i].BondPat;
    }
  } else {
    R__b.WriteVersion(TPhSensorMapDBSensorItem::IsA());
    TNamed::Streamer(R__b);
    R__b << fRows;
    R__b << fColumns;
    R__b << fType;
    R__b << fBondPatSize;
    NumEntries = fRows * fColumns;
    for(i = 0; i < NumEntries; i++) {
      R__b << fMap[i].FEC;
      R__b << fMap[i].String;
      R__b << fMap[i].ChannelPos;
      R__b << fMap[i].BondPat;
    }

  }
}
