
#include "TPhAbsRawHitFEC.h"

ClassImp(TPhAbsRawHitFEC)
/////////////////////////////////////////////////////////////////////////
// 
// Abstract Base class providing the functionality for the FEC data 
// storage. This class does not store any data, but describes how the 
// data from the FEC is accessible to the user. 
//
// The TPhRawHitFECn0 (non 0-suppressed) and TPhRawHitFECs0 (0-supp.)
// classes implement the data storage.
//
// Created by: Patrick Decowski
// 
// $Id: TPhAbsRawHitFEC.cxx,v 1.5 1999/02/24 23:34:04 olszewsk Exp $
//
/////////////////////////////////////////////////////////////////////////

TPhAbsRawHitFEC::TPhAbsRawHitFEC()
{
// Default Constructor

}

TPhAbsRawHitFEC::TPhAbsRawHitFEC(Int_t FECNum)
{
  // Normal constructor
  // Construct a name that can be used for lookup of this 
  // buffer in a hashtable

  Char_t Name[20];
  Int_t i;

  SetName(MakeFECName(FECNum, Name));
  fFECNumber = FECNum;

  fTrailer[kFECNumPos] |= FECNum & 0x00ff;

  for(i = 0; i < 8; i++)
    fRawHits[i] = NULL;
}

TPhAbsRawHitFEC::~TPhAbsRawHitFEC()
{
// Normal destructor

  Int_t i;

  for(i = 0; i < 8; i++) {
    if(fRawHits[i])
      delete[] fRawHits[i];
  }

}

Char_t *TPhAbsRawHitFEC::MakeFECName(Int_t FECNum, Char_t *String)
{
  // construct a string out of the FEC name. This string can be used for 
  // fast FEC buffer lookup in a hashcontainer. The user has to provide 
  // sufficient space for the string. 
  //
  // Return: pointer to the string with the FEC name.

  sprintf(String, "FEC_%d", FECNum);
  return String;
}

Char_t* TPhAbsRawHitFEC::Conv2Base(UChar_t Byte, Char_t *Str) const
{
  // Converts a Byte into 2 base notation and writes it into Str
  // Caller allocates space, the start of Str is returned back.

  Int_t i;
  Int_t k = 0;
  UChar_t digit; 

  for(i = 7; i >= 0; i--) {
    digit = Byte/UChar_t(TMath::Power(2, i));
    Byte -= digit * UChar_t(TMath::Power(2, i));
    Str[k++] = digit + '0';
  }
  Str[k++] = '\0';
  return Str;
}

UChar_t TPhAbsRawHitFEC::GetMonitorVal(Int_t Index)
{
  // Return one of the 32 monitor values
  //
  // The values are stored in 16 15bit words. Even index returns
  // an 8-bit monitor value; odd index returns a 7-bit monitor value.

  Int_t i;

  if(fTrailer == NULL) {
    Error("GetMonitorVal", "fTrailer is NULL!\n");
    return 0; // got to return something
  }
    

  if(Index > 31) {
    Error("GetMonitorVal", "Index out of bounds! Index = %d\n", Index);
    return 0;
  }

  i = Index / 2;
  // make sure that this is a header (bit 15 set)
  if(!fTrailer[i] & 0x8000) {
    Error("GetMonitorVal", "Trailer does not have the required  bit set!\n");
    return 0;
  }

  // get the value by indexing the trailer
  if(Index%2 == 0) 
    return (fTrailer[i] >> 8) & 0x007f;
  else
    return fTrailer[i] & 0x00ff;
}


FECRawHit_t*  TPhAbsRawHitFEC::GetRawHitArray(Int_t string)
{
  // Return a pointer pointing to the array of hits

  return fRawHits[string];
}

ERunMode TPhAbsRawHitFEC::GetMode()
{
  // Return the runmode

  return ERunMode(fTrailer[kRunModePos] >> 8);
}

Int_t TPhAbsRawHitFEC::GetError()
{
  // Get the error. The bit positions indicates which string
  // raised the error. Currently we OR the FIFO Err and Shift Out Err
  // together.

  return (fTrailer[kFIFOErrorPos] 
	  | fTrailer[kShiftOutErrorPos]) & 0x00ff;

}

void TPhAbsRawHitFEC::Add(Int_t String, FECRawHit_t *hit)
{
  // Add a hit to a buffer on a specific string on the FEC
  //
  // FOR FAKE DATA GENERATION!
  //
  // Note that for the time being we assume that all the fake data
  // is generated in one session. eg. making some fake data, writing it
  // out and making some more fake data will NOT work! i don't think that 
  // that is too much of a problem.  

  Int_t i;

  if(String >= kNumStrings) {
    Error("Add", "String too large: String = %d", String);
    return;
  }

  if(hit->Channel >= kMaxLength) {
    Error("Add", "Channel too large: Channel = %d", hit->Channel);
    return;
  }

  if(fRawHits[String] == NULL) {
    fRawHits[String] = new FECRawHit_t[kMaxLength];
    fRawHitBufSize[String] = 0;
    // zero array where we store the hits
    for(i = 0; i < kMaxLength; i++) {
      fRawHits[String][i].Channel = 0;
      fRawHits[String][i].ADC = 0;
    }
  }

  if(fRawHitBufSize[String] <= hit->Channel)
    fRawHitBufSize[String] = hit->Channel + 1;
  // Copy the hit into the buffer
  fRawHits[String][hit->Channel] = *hit;
}

void TPhAbsRawHitFEC::SetMonitorVal(Int_t Index, UChar_t Val)
{
  // Set a monitor value to a specific value. 
  //
  // FOR FAKE DATA GENERATION!

  Int_t i;

  if(Index > 31) {
    Error("SetMonitorVal", "Index out of bounds! Index = %d\n", Index);
    return;
  }

  i = Index / 2;

  // set the value of the monitor at a specific location
  UShort_t Save;

  Save = fTrailer[i];

  if(Index%2 == 0) {
    // The lower byte of the trailer is an 8-bit value
    fTrailer[i] = Save & 0x00ff;
    fTrailer[i] |= (Val << 8);
  }
  else {
    // The upper byte of the trailer is a 7-bit value + bit 15 set
    fTrailer[i] = Save & 0xff00;
    fTrailer[i] |= Val;
  }  
  SETBIT(fTrailer[i], 15);
}

void TPhAbsRawHitFEC::SetMode(ERunMode Mode)
{
  // Set the running mode label
  // 
  // FOR FAKE DATA GENERATION!

  UShort_t Save;
  
  Save = fTrailer[kRunModePos];
  fTrailer[kRunModePos] = 0;
  fTrailer[kRunModePos] |= (Save & 0x00ff);
  fTrailer[kRunModePos] |= (Mode << 8);
}

void TPhAbsRawHitFEC::SetFIFOError(Int_t String)
{
  // Set the FIFO Error flag for a given string
  // 
  // FOR FAKE DATA GENERATION!

  if(String >= kNumStrings) {
    Error("SetFIFOError", "The specified string is too large!\n");
    return;
  }
  
  SETBIT(fTrailer[kFIFOErrorPos], String);
}

void TPhAbsRawHitFEC::SetShiftOutError(Int_t String)
{
  // Set the Shift Out Error flag for a given string
  //
  // FOR FAKE DATA GENERATION!

  if(String >= kNumStrings) {
    Error("SetShiftOutError", "The specified string is too large!\n");
    return;
  }
  
  SETBIT(fTrailer[kShiftOutErrorPos], String);

}

void TPhAbsRawHitFEC::CalcTotalWords()
{
  // Calculate the total number of words in the buffer

  Int_t String;
  fTotalWords = 0;

  // figure out what the total number of words is.
  for(String = 0; String < kNumStrings; String++) {
    if(fRawHits[String] != NULL)
      fTotalWords += fRawHitBufSize[String];
  }

  fTotalWords += kTrailerLength;
}

void TPhAbsRawHitFEC::Print(Option_t *option)
{
  // Print the unpacked content

  Int_t i, j, nhits;
  UInt_t channel;
  Int_t adc;

  if(!fRawHits) 
    Warning("Print","use PrintPacked method to list packed contents\n");;

  // Prints the unpacked hits 
  printf("FEC# %d - unpacked hits:\n", GetFECNum());
  for(i = 0; i < kNumStrings; i++) {
    nhits = fRawHitBufSize[i];
    if(!nhits) continue;
    printf("String %d:\n", i);
    for(j = 0; j < nhits; j++) {
      channel = fRawHits[i][j].Channel;
      adc = fRawHits[i][j].ADC;
      if( channel || adc )
	printf("             Channel: %5d ADC: %5d\n", 
	       channel, adc);
    }
  }
}

void TPhAbsRawHitFEC::Streamer(TBuffer &R__b)
{
  // Stream an object of class TPhAbsRawHitFEC.

  if (R__b.IsReading()) {
    Version_t R__v = R__b.ReadVersion(); if (R__v) { }
    TPhDataObject::Streamer(R__b);
    R__b >> fFECNumber;
  } else {
    R__b.WriteVersion(TPhAbsRawHitFEC::IsA());
    TPhDataObject::Streamer(R__b);
    R__b << fFECNumber;
  }
}
