#include <math.h>
#include <stdlib.h>
#include "TPhDST.h"

//*-*-* ROOT includes

#include "TROOT.h"
#include "TFile.h"
#include "TTree.h"
#include "TBranch.h"
#include "TKey.h"
#include "TClonesArray.h"
#include "TList.h"


ClassImp(TPhDST)

////////////////////////////////////////////////////////////////
//
// TPhDST is a PHOBOS object providing services for accessing 
// TPhEvent objects stored in a ROOT file. 
// TPhEvent provides member function for to open and close one
// file and to loop over all events in the file using the 
// GetNextEvent() member function.
//
// $Id: TPhDST.cxx,v 1.6 1999/01/19 18:33:30 decowski Exp $
//
////////////////////////////////////////////////////////////////


TPhDST::TPhDST()
{
// Constructor. Set fNumEvent and fRootFile to NULL
    fNumEvent  = 0;
    fVerbose = 0;
    fRootFile  = NULL; 
    fNextKey = NULL;
}

TPhDST::TPhDST(Char_t *FileName)
{
// Constructor. Set fNumEvent and fRootFile to NULL
// and open the file *FileName

    fNumEvent  = 0;
    fVerbose = 0;
    fRootFile  = NULL; 
    Open(FileName);
}

TPhDST::~TPhDST()
{
// Destructor. Close the file properly if it is still open

  if(fRootFile) {
    if(fRootFile->IsOpen())
      fRootFile->Close();
    delete fRootFile;
  }

  SafeDelete(fNextKey);
}

Bool_t TPhDST::Open(char *FileName)
{
// Open the ROOT file *FileName. Return kFALSE if the 
// file can not be found. Otherwise create a list of 
// all keys to the events and reset counter for the 
// event loop.

  fNumEvent = 0;

  if(fVerbose)
    printf("Open root file %s\n",FileName);
  fRootFile    = new TFile(FileName);
  if(fRootFile) {
    if(!fRootFile->IsOpen()) {
      if(fVerbose)
	Warning("Open", "Couldn't open file %s\n");
      return kFALSE;
    }
    fRootFile->cd();
    fListOfKeys = gDirectory->GetListOfKeys();
    if(fVerbose) {
      printf("Got List of keys (%d)\n", fListOfKeys->GetSize());
      printf("Create Iterator\n");
    }
    fNextKey    = new TIter(fListOfKeys);
  }
  else {
    if(fVerbose)
      Warning("Open", "fRootFile = NULL\n");
    return kFALSE;
  }
  return kTRUE;
}

Bool_t TPhDST::Close()
{
// Close the current file (fRootFile). Return kFALSE
// if fRootFile is NULL. 

  if(fRootFile) {
    fRootFile->Close();
    delete fRootFile;  
    fRootFile = NULL;
  }
  else {
    if(fVerbose)
      Warning("Close", "fRootFile = NULL\n");
    return kFALSE;
  }

  // Get rid of any key operator
  SafeDelete(fNextKey);

  return kTRUE;
}


TPhEvent *TPhDST::GetNextEvent(Int_t skip)
{
// Read one event from the data file and return 
// a pointer to the TPhEvent object. If the skip
// argument is provided, skip over the corresponding
// number of keys before reading an event.
//
// Note that the memory allocated for the event has 
// to be freed by the program calling GetNextEvent().

 
  TKey *key;

  if(!fNextKey) {
    if(fVerbose)
      Error("GetNextEvent", "No Keys!");
    return NULL;
  }

  for(Int_t i=0;i<skip;i++) {
    key = (TKey *) fNextKey->operator()();
    if( !key ) 
      {
	if(fVerbose)
	  printf("No keys left - skipped past EOF\n");
	return NULL;
      }
  }

  if( (key = (TKey *) fNextKey->operator()()) != NULL) {
    // Make sure that this is an event, skip over all other objects.
    while(strcmp(key->GetClassName(), "TPhEvent") != 0) {
      key = (TKey *) fNextKey->operator()();
      if(!key) break;
    }
  }

  if(key) {
    fRootFile->cd();
    // TKey::Read got renamed to TKey::ReadObj in Root V1.03/09
    TPhEvent *Event = (TPhEvent *) key->ReadObj();
    if(Event == NULL)
      if(fVerbose)
	Warning("GetNextEvent", "key->ReadObj() returned NULL\n");
    fNumEvent++;
    return Event;
  }
  else {
    if(fVerbose)
      printf("No keys left - reached EOF\n");
    return NULL;
  }
}

TPhDetector *TPhDST::GetDetector(Char_t *name)
{
// Get detector (named name) from the data file

   TPhDetector *detector;
   if(fRootFile) detector = (TPhDetector *) fRootFile->Get(name);
   if( strcmp(detector->ClassName(),"TPhDetector") ) {
     if(fVerbose>0) printf("%s object - not a TPhDetector\n",name);
     return 0;
   }
   return detector;
}

TPhDetector *TPhDST::GetDetector()
{
// Get detector object (first found) from the data file
//
   TKey *key;
   TIter nextKey(fListOfKeys);
   if((key = (TKey *) nextKey()) != NULL ) {
     // Look for TPhDetector object
     while(strcmp(key->GetClassName(), "TPhDetector") != 0) {
       key = (TKey *) nextKey();
       if(!key) break;
     }
   }
   
   if(key) {
     fRootFile->cd();
     TPhDetector *detector = (TPhDetector *) key->ReadObj();
     return detector;
   }
   else {
     if(fVerbose)
       printf("TPhDetector object not found on %s DST\n",GetFile()->GetName());
     return NULL;
   }
}

void TPhDST::Reset()
{
// Reset eventcounter to allow looping from beginning of data
// file.

  fNextKey->Reset();
  fNumEvent = 0;
}





