#ifndef __TPhString__
#define __TPhString__


//////////////////////////////////////////////////////////////////////////////
//                                                                          //
//    TPhString                                                             //
//                                                                          //
//    Class storing information on structure and properties of electronic   //
//    channels withing one electronic string.                               //
//    FEC is divided into strings of channels, that may come from different //
//    sensors. A string may come from from several chips, each of them      //
//    may be connected to a various number of pads.                         //
//                                                                          //
//    Author  : Andrzej Olszewski (andrzej.olszewski@ifj.edu.pl)            //
//    Created : 29/12/98                                                    //
//    Version : 1.0                                                         //
//    Changed :                                                             //
//                                                                          //
// $Id: TPhString.h,v 1.2 1999/03/05 11:54:18 olszewsk Exp $
//                                                                          //
//////////////////////////////////////////////////////////////////////////////

// ROOT classes
#include "TObjArray.h"
#include "TArrayS.h"
#include "TArrayI.h"
#include "TArrayD.h"

// Phat classes
#include "TPhHashContainer.h"
#include "TPhChip.h"
#include "TPhChannelCalAbs.h"
#include "TPhStringCalAbs.h"

struct FECRawHit_t;
class TPhRawHitFECs0;

enum EDescriptionFEC {
  kMaxLength =   768,      // The (maximum) length of one string
  kNumStrings =    8,      // The (maximum) number of strings on an FEC
  kMaxChipNumber =12,      // The (maximum) number of chips in a string 
  kMaxChipSize = 128       // The (maximum) number of pads connected to a chip
};

enum EvaModes_t {
  kPedPreMode,             // pedestal pre-evaluation mode
  kPedMode,                // pedestal evaluation mode
  kNoiseMode,              // noise evaluation mode
  kZeroSuppMode,           // zero suppresion mode
  kCalibratedMode          // zero suppresion mode
};

class TPhString : public TObject {

public:

  TPhString();
  TPhString(Int_t fecnum, Int_t stringno, Int_t stringsize,
	    Int_t chipno, Int_t *chipsizes);
  virtual ~TPhString();

  UShort_t      GetFECNo()                { return fFECNo;      }
  UShort_t      GetStringNo()             { return fStringNo;   }
  UShort_t      GetStringSize()           { return fStringSize; }
  UShort_t      GetNoChips()              { return fNoChips;    }
  TPhChip      *GetChip(Int_t no);
  TPhChannelCalAbs* GetChannelCalObject(Int_t no); 

  // assumed non-zero-suppressed strings here
  // ----------------------------------------
  Int_t         PedPreProcess(FECRawHit_t *hitstring);  
  Int_t         PedProcess(FECRawHit_t *hitstring);
  Int_t         CMNProcess(FECRawHit_t *hitstring);
  Int_t         NoiseProcess(FECRawHit_t *hitstring);
  Int_t         ZeroSuppress(FECRawHit_t *hitstring);
  Int_t         ZeroSuppress(FECRawHit_t *inhitstring, TPhRawHitFECs0 *rawBufs0);
  // ----------------------------------------------------------------------------
  void          Calibrate(Bool_t &status);
  Bool_t        Calibrate(FECRawHit_t *hitstring, Int_t strlength);
  Bool_t        CalibrateOnString(FECRawHit_t *hitstring, Int_t strlength);

  Bool_t        AddChannelCalObject(Int_t channo, TPhChannelCalAbs *chobj);
  Bool_t        MakeStringCalObject(TPhStringCalAbs *strcal);
  TPhStringCalAbs *GetStringCalObject()         { return fCalStringObj;                   }
  void          MakeHitArrays(TPhHashContainer *HitArrayCont);
  void          DeleteCalChannelArray()  { if(fCalChannelArray) delete fCalChannelArray;  }
  void          DeleteCalStringObject()  { if(fCalStringObj)    delete fCalStringObj;     }

  void          DeletePedPreArrays();
  void          CreatePedPreArrays();
  void          DeletePedArrays();
  void          CreatePedArrays();
  void          DeleteCMNArrays();
  void          CreateCMNArrays();
  void          DeleteNoiseArrays();
  void          CreateNoiseArrays();
  void          DeleteZZArrays();
  void          CreateZZArrays();
  void          DeleteSuppArrays();
  void          CreateSuppArrays();

  void          DumpPedPreMean(Int_t no=0)      {  DumpArrayD(fPedPreMean, no);           }
  void          DumpPedMean(Int_t no=0)         {  DumpArrayF(fPedMean, no);              }
  void          DumpPedRMS(Int_t no=0)          {  DumpArrayD(fPedRMS, no);               } 
  void          DumpCMNMean(Int_t no=0)         {  DumpArrayF(fCMNMean, no);              }
  void          DumpCMNFlag(Int_t no=0)         {  DumpArrayS(fCMNFlag, no);              }
  void          DumpNoise(Int_t no=0)           {  DumpArrayD(fNoise, no);                }
  void          DumpZSSHitChan(Int_t no=0)      {  DumpArrayS(fZSSHitChan, no);           }
  void          DumpZSSHitADC(Int_t no=0)       {  DumpArrayF(fZSSHitADC, no);            }
  void          DumpEvents(Int_t no=0)          {  DumpArrayI(fEvents, no);               }

  //            Setting/getting parameters for pedestal/noise routines
  void          SetMIN_NUM_PED_EVENTS(Int_t value)   { MIN_NUM_PED_EVENTS = value;        }
  void 	        SetMIPADC_CUTOFF(Int_t value)        { MIPADC_CUTOFF = value;             }
  void          SetMIN_CHANNELS_CMN_EST(Int_t value) { MIN_CHANNELS_CMN_EST = value;      }
  void          SetMIN_NUM_NOISE_EVENTS(Int_t value) { MIN_NUM_NOISE_EVENTS = value;      }
  void          SetTHRESH_CMN(Float_t value)         { THRESH_CMN = value;                }
  void          SetTHRESH_MULTIPLIER(Float_t value)  { THRESH_MULTIPLIER = value;         }

  Int_t         GetMIN_NUM_PED_EVENTS()         { return MIN_NUM_PED_EVENTS;              }
  Int_t 	GetMIPADC_CUTOFF()              { return MIPADC_CUTOFF;                   }
  Int_t         GetMIN_CHANNELS_CMN_EST()       { return MIN_CHANNELS_CMN_EST;            }
  Int_t         GetMIN_NUM_NOISE_EVENTS()       { return MIN_NUM_NOISE_EVENTS;            }
  Float_t       GetTHRESH_CMN()                 { return THRESH_CMN;                      }
  Float_t       GetTHRESH_MULTIPLIER()          { return THRESH_MULTIPLIER;               }

  Bool_t        GetDoHistograms()               { return fDoHistograms;                   }     
  void          SetDoHistograms(Bool_t onoff)   { fDoHistograms = onoff;                  }

  void          DeleteHistograms();
  void          AddPedPreHistogram(Int_t no, Float_t xlow, Float_t xhigh, Int_t nbins); 
  TH1F         *GetPedPreHistogram(Int_t no)    { return GetHistogram(fPedPreHistos, no); }
  void          DrawPedPreHistogram(Int_t no)
                            { DrawHistogram(fPedPreHistos, no);                           }
  TH1F         *CreatePedPreSHisto();
  TH1F         *GetPedPreSHistogram()           { return fPedPreSHisto;                   }
  void          DrawPedPreSHisto();
  void          AddPedMeanHistogram(Int_t no, Float_t xlow, Float_t xhigh, Int_t nbins);
  TH1F         *GetPedMeanHistogram(Int_t no)   { return GetHistogram(fPedMeanHistos, no);}
  void          DrawPedMeanHistogram(Int_t no)
                            { DrawHistogram(fPedMeanHistos, no);                          }
  TH1F         *CreatePedMeanSHisto();
  TH1F         *GetPedMeanSHistogram()          { return fPedMeanSHisto;                  }
  void          DrawPedMeanSHisto();
  void          AddPedRMSHistogram(Int_t no, Float_t xlow, Float_t xhigh, Int_t nbins);
  TH1F         *GetPedRMSHistogram(Int_t no)    { return GetHistogram(fPedRMSHistos, no); }
  void          DrawPedRMSHistogram(Int_t no)
                            { DrawHistogram(fPedRMSHistos, no);                           }
  TH1F         *CreatePedRMSSHisto();
  TH1F         *GetPedRMSSHistogram()           { return fPedRMSSHisto;                   }
  void          DrawPedRMSSHisto();

  void          AddCMNEventHistogram(Int_t no, Float_t xlow, Float_t xhigh, Int_t nbins);
  TH1F         *GetCMNEventHistogram(Int_t no)  { return GetHistogram(fCMNEventHistos, no);}
  void          DrawCMNEventHistogram(Int_t no);

  void          AddCMNMeanHistogram(Int_t no, Float_t xlow, Float_t xhigh, Int_t nbins);
  TH1F         *GetCMNMeanHistogram(Int_t no)   { return GetHistogram(fCMNMeanHistos, no);}
  void          DrawCMNMeanHistogram(Int_t no)
                            { DrawHistogram(fCMNMeanHistos, no);                          }
  TH1F         *CreateCMNMeanSHisto();
  TH1F         *GetCMNMeanSHistogram()          { return fCMNMeanSHisto;                  }
  void          DrawCMNMeanSHisto();
  void          AddCMNFHistogram(Int_t no, Float_t xlow, Float_t xhigh, Int_t nbins);
  TH1F         *GetCMNFHistogram(Int_t no)      { return GetHistogram(fCMNFHistos, no);   }
  void          DrawCMNFHistogram(Int_t no)
                            { DrawHistogram(fCMNFHistos, no);                             }
  TH1F         *CreateCMNFSHisto();
  TH1F         *GetCMNFSHistogram()             { return fCMNFSHisto;                     }
  void          DrawCMNFSHisto();
  void          AddNoiseHistogram(Int_t no, Float_t xlow, Float_t xhigh, Int_t nbins);
  TH1F         *GetNoiseHistogram(Int_t no)     { return GetHistogram(fNoiseHistos, no);  }
  void          DrawNoiseHistogram(Int_t no)
                            { DrawHistogram(fNoiseHistos, no);                            }
  TH1F         *CreateNoiseSHisto();
  TH1F         *GetNoiseSHistogram()            { return fNoiseSHisto;                    }
  void          DrawNoiseSHisto();
  void          AddZSSHistogram(Int_t no, Float_t xlow, Float_t xhigh, Int_t nbins);
  TH1F         *GetZSSHistogram(Int_t no)       { return GetHistogram(fZSSHistos, no);    }
  void          DrawZSSHistogram(Int_t no)
                            { DrawHistogram(fZSSHistos, no);                              }
  TH1F         *CreateZSSSHisto();
  TH1F         *GetZSSSHistogram()              { return fZSSSHisto;                      }
  void          DrawZSSSHisto();
  void          AddCalHistogram(Int_t no, Float_t xlow, Float_t xhigh, Int_t nbins);
  TH1F         *GetCalHistogram(Int_t no)
                            { return GetHistogram(fCalHistos, no);                        }
  void          DrawCalHistogram(Int_t no)
                            { DrawHistogram(fCalHistos, no);                              }

  virtual void  ls(Option_t *option = "");
  virtual void  Print(Option_t *option = "");

  Int_t         GetVerbose() const        { return fVerbose; }
  void          SetVerbose(Int_t verbose) { fVerbose = verbose; }
  void          SetVerboseDown(Int_t verbose) 
                            { fVerbose = verbose;
		              fCalChannelArray->ForEach(TPhChannelCalAbs,SetVerbose)(verbose); }

protected:

  Bool_t        BuildChipList(Int_t chipno, Int_t *chipsizes);
  void          DumpArrayS(TArrayS *ar, Int_t no);
  void          DumpArrayF(TArrayF *ar, Int_t no);
  void          DumpArrayD(TArrayD *ar, Int_t no);
  void          DumpArrayI(TArrayI *ar, Int_t no);

  TH1F         *GetHistogram(TObjArray *list, Int_t no);
  void          DrawHistogram(TObjArray *list, Int_t no);

  UShort_t      fFECNo;                 // FEC number
  UShort_t      fStringNo;              // String number
  UShort_t      fStringSize;            // No of channels in this string
  UShort_t      fNoChips;               // Number of chips
  Bool_t        fDoHistograms;          //!Switch to turn histo fill on/off 
  TPhStringCalAbs *fCalStringObj;       // Calibration object with parameters for all channels
  TObjArray    *fCalChannelArray;       // Array of calibration objects for all channels
  TObjArray    *fChips;                 // Array of chip objects
  TArrayD      *fPedPreMean;            // Buffer to store mean of signals+pedestals
  TArrayF      *fPedMean;               // Pedestal means evaluated for channels in this string
  TArrayD      *fPedRMS;                // Pedestal RMSs evaluated for channels in this string
  TArrayF      *fCMNMean;               // CMN means evaluated for channels in this string
  TArrayS      *fCMNFlag;               // CMN validity flags for channels in this string
  TArrayD      *fNoise;                 // Noise levels evaluated for this string
  TArrayS      *fZSSHitChan;            // Channel numbers of raw hits after zero suppressions
  TArrayF      *fZSSHitADC;             // ADCs of raw hits after zero suppressions

  TArrayF      *fSignal;                //! Temporary raw ADC-Pedestal
  TArrayI      *fEvents;                //! Number of events gathered
  TArrayF      *fSum;                   //! Buffer to store sum of signals
  TArrayD      *fSumSq;                 //! Buffer to store sum of squares of signals

  TArrayC      *fPedPreHistoChannels;   //! A list of channels to histogram (-1 all in one)
  TArrayC      *fPedMeanHistoChannels;  //! A list of channels to histogram (-1 all in one)
  TArrayC      *fPedRMSHistoChannels;   //! A list of channels to histogram (-1 all in one)
  TArrayC      *fCMNEventHistoChannels; //! A list of channels to histogram (-1 all in one)
  TArrayC      *fCMNMeanHistoChannels;  //! A list of channels to histogram (-1 all in one)
  TArrayC      *fCMNFHistoChannels;     //! A list of channels to histogram (-1 all in one)
  TArrayC      *fNoiseHistoChannels;    //! A list of channels to histogram (-1 all in one)
  TArrayC      *fZSSHistoChannels;      //! A list of channels to histogram (-1 all in one)
  TArrayC      *fCalHistoChannels;      //! A list of channels to histogram (-1 all in one)

  TObjArray    *fPedPreHistos;          //! A list of histograms
  TObjArray    *fPedMeanHistos;         //! A list of histograms
  TObjArray    *fPedRMSHistos;          //! A list of histograms
  TObjArray    *fCMNEventHistos;         //! A list of histograms
  TObjArray    *fCMNMeanHistos;         //! A list of histograms
  TObjArray    *fCMNFHistos;            //! A list of histograms
  TObjArray    *fNoiseHistos;           //! A list of histograms
  TObjArray    *fZSSHistos;             //! A list of histograms
  TObjArray    *fCalHistos;             //! A list of histograms

  TH1F         *fPedPreSHisto;          //! A summary histogram
  TH1F         *fPedMeanSHisto;         //! A summary histogram
  TH1F         *fPedRMSSHisto;          //! A summary histogram
  TH1F         *fCMNMeanSHisto;         //! A summary histogram
  TH1F         *fCMNFSHisto;            //! A summary histogram
  TH1F         *fNoiseSHisto;           //! A summary histogram
  TH1F         *fZSSSHisto;             //! A summary histogram

  EvaModes_t    fMode;                  //! Pedestal, Noise, or ZeroSupp modes possible
                                        //! (CMN can be evaluated on the event-by-event basis)

  Int_t         MIN_NUM_PED_EVENTS;     // pedestal routine parameter
  Int_t 	MIPADC_CUTOFF;          // MIP parameter
  Int_t         MIN_CHANNELS_CMN_EST;   // CMN noise routine parameter
  Int_t         MIN_NUM_NOISE_EVENTS;   // Channel noise routine parameter
  Float_t       THRESH_CMN;             // Threshold in CMN routine
  Float_t       THRESH_MULTIPLIER;      // Threshold in zero-suppression routine

  Int_t         fVerbose;               //! verbose flag
public:

  ClassDef(TPhString,1)	// String and string channel's properties

};


#endif // __TPhString__
