#include "TPhDetector.h"
#include "TPhSubDetector.h"
#include "TPhSensorMapDB.h"
#include "TPhMagnet.h"
#include "TPhEDigi.h"
#include "TPhEvent.h"
#include "TPhObjectContainer.h"
#include "TPhHashContainer.h"
#include "TPhRawHitFECn0.h"
#include "TPhRawHitFECs0.h"
#include "TPhHitArray.h"
#include "TPhHit.h"

#include "TPhSpectSensor.h"
#include "TPhOctSensor.h"
#include "TPhVTXSensor.h"
#include "TPhTOFSensor.h"
#include "TPhRingSensor.h"
#include "TPhTriggerSensor.h"
#include "TPhHighEtaSensor.h"
#include "TPhPaddleSensor.h"

#include "TPhSensorMapDBSensorItem.h"
#include "TPhSensorMapDBFECItem.h"

#include "TFile.h"
#include "TView.h"
#include "TCanvas.h"

#include <iostream.h>
#include <math.h>

//#define DEBUG
#undef DEBUG

// here are some consts
const Int_t kMaxSubdetectors = 7;            //! Number of PHOBOS subdetectors 
const Int_t kMaxDetectorPlanes = 682;        //! Number of PHOBOS planes

ClassImp(TPhDetector)
////////////////////////////////////////////////////////////////////
//                                                                //
// The TPhDetector class provides an interface to the PHOBOS      //
// geometry. It is currently filled by the TPhASCIIData and       //
// TPhPMC classes, which are interface to the Geant MC, PMC       // 
//                                                                //
// The TPhDetector object manages a list of TPhSubdetector        // 
// objects, which in turn manage TPhLayer objects which manage    //
// the TPhSensor objects. The TPhSensor objects hold the actual   //
// geometry information and own the unpack TPhHit objects.        //
//                                                                //
// $Id: TPhDetector.cxx,v 1.20 1999/02/25 00:33:04 olszewsk Exp $
//                                                                //
////////////////////////////////////////////////////////////////////

TPhDetector *gDetector;

Bool_t TPhDetector::fgDetIsInitialized = kFALSE;

TPhDetector *TPhDetector::GetInstance()
{
// Returns the singleton instance of the Detector object.

    return gDetector;
}

TPhDetector *TPhDetector::CreateInstance()
{
// Returns the singleton instance of the Detector object. If one is already
// present it will be deleted and a new one is created. 
// Creates the following subdetector objects:
//    "RingMult" Ring Multiplicity detectors
//    "Octagon"  Octagon Mult detector detector
//    "Vertex"   Vertex detector
//    "SpectP"   Positive Spectrometer arm
//    "SpectN"   Negative Spectrometer arm
//    "TOFP"     Positive Time of flight detector
//    "TOFN"     Positive Time of flight detector
//    "Trigger"  Trigger detector

    if(fgDetIsInitialized)
    {
     fgDetIsInitialized = kFALSE;
     delete gDetector;
    }
    gDetector = new TPhDetector();

    // The reason for doing the following memory allocation for the 
    // SubDetector List and Sensor List in this way is, that we can't 
    // directly call the datamembers, since this is a static member 
    // function.
    THashTable *subdlist = new THashTable(kMaxSubdetectors);
    THashList  *senslist = new THashList(kMaxDetectorPlanes);

    gDetector->SetSubDetectorList( subdlist );
    gDetector->SetSensorList( senslist ); 

    TPhSubDetector *subDetector;

    subDetector = new TPhSubDetector("RingMult");
    gDetector->AddSubDetector(subDetector);
    subDetector = new TPhSubDetector("Octagon");
    gDetector->AddSubDetector(subDetector);
    subDetector = new TPhSubDetector("Vertex");
    gDetector->AddSubDetector(subDetector);
    subDetector = new TPhSubDetector("SpectP");
    gDetector->AddSubDetector(subDetector);
    subDetector = new TPhSubDetector("SpectN");
    gDetector->AddSubDetector(subDetector);
    subDetector = new TPhSubDetector("TOFP");
    gDetector->AddSubDetector(subDetector);
    subDetector = new TPhSubDetector("TOFN");
    gDetector->AddSubDetector(subDetector);
    subDetector = new TPhSubDetector("Trigger");
    gDetector->AddSubDetector(subDetector);

    return gDetector;
}

void TPhDetector::DeleteInstance()
{
// Returns the singleton instance of the Detector object. If one is already
// present it will be deleted and a new one is created. 

    if(fgDetIsInitialized)
    {
     fgDetIsInitialized = kFALSE;
     delete gDetector;
    }
    gDetector = NULL;
    return;
}

TPhDetector *TPhDetector::GetOrCreateInstance()
{
// Returns the singleton instance of the Detector object. If it is not
// yet initialized, it will be created. 

    if(!fgDetIsInitialized)
    {
        gDetector = CreateInstance();
    }
    return gDetector;
}

TPhDetector *TPhDetector::ReadLatestGeometry()
{
// Reads the TPhDetector object named "LatestGeometry" 
// from a standard file $PHATHOME/init/test/Detector.root.
 
    Char_t FileName[100];
    const Char_t *phathome = gSystem->Getenv("PHATHOME");
    sprintf(FileName,"%s/init/test/%s",phathome,"Detector.root");
    TFile fileDet(FileName);

    DeleteInstance();
    gDetector = (TPhDetector *) fileDet.Get("LatestGeometry");
    if(gDetector) 
      printf("TPhDetector ver.%f found in %s\n", 
              gDetector->GetVersion(), FileName);
    else 
      printf("Error in TPhDetector::ReadLatestGeometry()" \
             "LatestGeometry object not found in %s file\n", FileName);

    return gDetector;
}

TPhDetector::TPhDetector()
{
// Constructor. Sets fLayerList to NULL. To fill the 
// the layer lists and the sensor lists,
// use the ReadASCIIGeometry() call. So far this only fills 
// one subdetector object, namely the right-hand side
// spectrometer arm.

    if(!fgDetIsInitialized)
    {
        fSubDetectorList  =  NULL;
        fSensorList =  NULL;
        fEvent      =  NULL;
        fMagnet     =  NULL;
	fVerbose    = 0;
        ReadInDBMap();

        fVersion    = 0;
        fgDetIsInitialized = kTRUE;
        gDetector  = this;
    }
    else
    {
        Error("TPhDetector()", "TPhDetector already exists." \
              "Use TPhDetector::GetInstance to access it.\n");
    }
}

TPhDetector::~TPhDetector()
{
// Destructor. Deletes TPhLayer and TPhSensor objects
// Reset singleton flag.                             

    gDetector = NULL;
    fgDetIsInitialized = kFALSE;

    delete fSensorMapDB;

    // The TPhSubDetectors and the TPhSensors both belong to the TPhDetector
    // class. The TPhLayers inside a TPhSubDetector also have pointers to 
    // the TPhSensors, but they don't _OWN_ them, this means that their 
    // destructor does not delete the TPhSensor objects. 
    if(fSubDetectorList) {
      fSubDetectorList->Delete();
      delete fSubDetectorList;
    }

    if(fSensorList) {
      fSensorList->Delete();
      delete fSensorList;
    }
    SafeDelete(fBrik);
    SafeDelete(fNode);

}

Bool_t TPhDetector::AddSensor(Int_t detno, Char_t *detname, Int_t shapeno, 
                    Float_t *center, Int_t nvolpar, Float_t *volpars, 
                    TRotMatrix *rotmatp, Int_t padtype, Float_t *padpars)
{
  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  //                  Creation of sensor/layer objects                        !
  //                 TRotMatrix must be created outside                       !
  // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

   Int_t layernumber;

   if(!rotmatp) {
     Warning("AddSensor", 
             "Rotation matrix no %d does not exist - cannot add sensor %s!\n", 
              detname);
     return 0;
   }

   TPhSensor *sensor;

   TPhSpectSensor *SpectSensor = NULL;
   TPhOctSensor *OctSensor = NULL;
   TPhVTXSensor *VTXSensor = NULL;
   TPhRingSensor *RingSensor = NULL;
   TPhTOFSensor *TOFSensor = NULL;
   TPhBoxSiSensor *BoxSiSensor = NULL;
   TPhTriggerSensor *TriggerSensor = NULL;
   TPhHighEtaSensor *HighEtaSensor = NULL;
   TPhPaddleSensor *PaddleSensor = NULL;

   TPhSubDetector *subdetector = NULL;
   sensor = NULL;
       // Look at the GEANT shape number. 1 corresponds to 'BOX':
       if( (Int_t)shapeno == 1 ) {
	 //if the name of the detector starts with 'T', 
	 //then make a TPhTOFSensor
	 if(!strncmp(detname, "T", 1)) {
	   TOFSensor = new TPhTOFSensor();
	   if(center[0] > 0)
	     subdetector = GetSubDetector("TOFP");
	   else
	     subdetector = GetSubDetector("TOFN");
	   // The Layer is set according to the third letter 
	   // in the detector name, eg:
	   // A = layer 0
	   TOFSensor->SetLayer(Int_t(detname[2] - 'A'));
	   TOFSensor->SetLabel(detno);
	   TOFSensor->SetName(detname);

	   // FIXME: test for nvolpar number as add. check - what is the #? 
	   if(nvolpar > 888 /* dummy! */)
	      Warning("AddSensor", "Integrity check failed!");

	   // Set the rotation matrix
           // Only one rotation matrix is used, because since Sept 1997 
	   // PMC writes to AOU
           // only one, final rotation matrix for each detector.
	   TOFSensor->SetMatrix( rotmatp );
	   // Set the position of the sensor
	   TOFSensor->SetPosition(center[0], center[1], center[2]);
	   // Set the half dimensions of the sensor
	   TOFSensor->SetDim(volpars[0], volpars[1], volpars[2]);
	   // Calculate the outer edges of the sensor
	   TOFSensor->CalcBoxPoints();
	   sensor = TOFSensor; 
	 }
	 // 'MO' for Multiplicity
	 else if (!strncmp(detname, "MO", 2)) {
 	   OctSensor = new TPhOctSensor();
	   OctSensor->SetLayer(Int_t(detname[2] - 'A'));
	   BoxSiSensor = OctSensor;
	   subdetector = GetSubDetector("Octagon");
	 }
	 // 'V' for Vertex
	 // We use the following convention:
	 // Layer 0 = Top Inner vertex
	 // Layer 1 = Bottom Inner vertex
	 // Layer 2 = Top Outer vertex
	 // Layer 3 = Bottom Outer vertex

	 else if (!strncmp(detname, "V", 1)) {
 	   VTXSensor = new TPhVTXSensor();
	   if(detname[1] == 'T') 
	     layernumber = 0;
	   else 
	     layernumber = 1;
	   layernumber += 2*Int_t(detname[2] - 'A');
	   VTXSensor->SetLayer(layernumber);
	   BoxSiSensor = VTXSensor;
	   subdetector = GetSubDetector("Vertex");
	 }
	 // All the other detector are Spectrometers
	 // Naming of the spectrometers was changed:
	 // version 3.8 and below spectrometers started with 'S'
	 // From version 4.0 the first letter of the spectrometer 
	 // is either 'P' or 'N'
	 else {
	   SpectSensor = new TPhSpectSensor();
	   if(center[0] > 0)
	     subdetector = GetSubDetector("SpectP");
	   else
	     subdetector = GetSubDetector("SpectN");
	   // we changed the names of the spectrometer modules 
	   // above ugcdata.txt version 4.0
	   // the layer number is the 2nd letter there
	   if(fVersion > 3.8)
	     SpectSensor->SetLayer(Int_t(detname[1] - 'A'));
	   else
	     // version 3.8 and below it is the 3rd letter
	     SpectSensor->SetLayer(Int_t(detname[2] - 'A'));
	   BoxSiSensor = SpectSensor;
	 }
	 if(BoxSiSensor) {
	   BoxSiSensor->SetLabel(detno);
	   BoxSiSensor->SetName(detname);

  	   // Set the rotation matrix
	   BoxSiSensor->SetMatrix( rotmatp );
	   // Set the position of the sensor
	   BoxSiSensor->SetPosition(center[0], center[1], center[2]);
	   // Set the half dimensions of the sensor
	   BoxSiSensor->SetDim(volpars[0], volpars[1], volpars[2]);
	   // Calculate the outer edges of the sensor
	   BoxSiSensor->CalcBoxPoints();

	   BoxSiSensor->SetPxlWidth(padpars[0]);
	   BoxSiSensor->SetPxlHeight(padpars[1]);

	   if(SpectSensor)
	     SpectSensor->CalcDistVar();
	   sensor = BoxSiSensor; 
	 } 
       }
       // If the Sensor name starts with a G, then it is a trigger sensor
       else if(detname[0] == 'G') {
	 // if shape parameter is 2, it is a paddle trigger counter
	 if ((Int_t)shapeno == 2) {
	   PaddleSensor = new TPhPaddleSensor();
	   PaddleSensor->SetName(detname);
	   PaddleSensor->SetParameters(volpars[0], volpars[1], 
                                       volpars[2], volpars[3]);
	   TriggerSensor = PaddleSensor;
	 }
	 // if shape parameter is 5, it is a high eta trigger counter
	 else if ((Int_t)shapeno == 5) {
	   HighEtaSensor = new TPhHighEtaSensor();
	   HighEtaSensor->SetName(detname);
	   HighEtaSensor->SetParameters(volpars[0], volpars[1], volpars[2]);
	   TriggerSensor = HighEtaSensor;
	 }
	 else
	   TriggerSensor = NULL;
	 if(TriggerSensor) {
	   TriggerSensor->SetLabel(detno);	 
	   TriggerSensor->SetMatrix( rotmatp );
	   TriggerSensor->SetPosition(center[0], center[1], center[2]);
	   subdetector = GetSubDetector("Trigger");
	   // layers 0, 1 are the positive arm trigger sensors
	   // layers 2, 3 are the neg. arm trigger sensors
	   if(detname[1] == 'P') 
	     layernumber = 0;
	   else
	     layernumber = 2;
	   layernumber += Int_t(detname[2] - 'A');
	   TriggerSensor->SetLayer(layernumber);
	   sensor = TriggerSensor;
	 }
       }
       // If the GEANT shape parameter is 6, it is a TUBS shape
       else if ((Int_t)shapeno == 6 ) {
	 // Make sure that it is a multiplicity sensor
	 if (!strncmp(detname, "M", 1)) {
	   RingSensor = new TPhRingSensor();
	   RingSensor->SetLabel(detno);
	   RingSensor->SetName(detname);

	   RingSensor->SetMatrix( rotmatp );
	   RingSensor->SetPosition(center[0], center[1], center[2]);
	   RingSensor->SetParameters(volpars);
	   //	   RingSensor->SetPxlWidth( padpars[0] );
	   //	   RingSensor->SetPxlHeight( padpars[1] ); 
	   // Variable pad sizes are set in TPhRingSensors
           RingSensor->SetPadParameters(padpars);
	   subdetector = GetSubDetector("RingMult");
	   RingSensor->SetLayer(Int_t(detname[2] - 'A'));
	   sensor = RingSensor;
	 }
       }
       // Check if this is the pre UGCData V4.0 implementation 
       else if ((Int_t)shapeno == 4 ) {
	 if (!strncmp(detname, "M", 1))
	   Warning("CreateDetector", 
                   "Old version of ring sensor goes here...\n");
       }

       if(sensor) {

	 // Depending on the pixelization type set the number of rows/columns
	 if((Int_t)padtype==6) {
	   // For rings number of pixels in radial direction is fixed
	   sensor->SetRows( 8 );
           // No of columns = Delta phi of tube segment/ Delta phi of a pad 
	   sensor->SetColumns(TMath::Nint((volpars[4]-volpars[3])/padpars[0]));
	 } else {
	   // Pixelization type 1 (and others if they happen)
           // Calculate the number of Rows and Columns by looking at 
	   // the padsize
           if(padpars[0] > 0 ) 
	     sensor->SetColumns(TMath::Nint(2.0*volpars[0]/padpars[0]) );
	   else 
	     sensor->SetColumns( 1 );
           if(padpars[1] > 0) 
  	     sensor->SetRows(TMath::Nint(2.0*volpars[1]/padpars[1]) );
	   else 
	     sensor->SetRows( 1 );
	 }
	 // add sensor to the given subdetector
	 subdetector->AddSensor(sensor);
	 // The GEANT sensor number gives the location of the sensor in our list
         // and its
         if(fSensorList->At(detno)) {
	   Error("AddSensor",
                 "Location %d demanded for sensor %s is already occupied", 
                  detno, detname);
           return kFALSE;
	 } else fSensorList->AddAt( sensor,detno );
#ifdef DEBUG
         printf("Sensor %3d: %4s Layer %3d : \t%d %d\n",
                detno, sensor->GetName(), sensor->GetLayer(),
		sensor->GetRows(),
		sensor->GetColumns());
#endif
         return kTRUE;
       }
       else {
	   Warning("CreateDetector", "Unknown sensor: %s", detname);
	   return kFALSE;
       }
       return kTRUE;
}


Bool_t TPhDetector::ReadInSensorGeometry()
{
  // This function will load standard geometry of sensors
  if( fSubDetectorList ) {
    delete fSubDetectorList; 
    fSubDetectorList = NULL;
  }
  Char_t SensorGeometryFileName[100];
  const Char_t *phathome = gSystem->Getenv("PHATHOME");
  sprintf(SensorGeometryFileName,"%s/init/%s",phathome,"SensorGeometry.root");
  return 
    ReadInSensorGeometry(SensorGeometryFileName);
}

Bool_t TPhDetector::ReadInSensorGeometry(TFile *tfile)
{
  // This function should be used to load custom geometry of sensors
  if(!tfile) return kFALSE;
  if(!tfile->IsOpen()) return kFALSE;
 
  if( fSubDetectorList ) {
    delete fSubDetectorList; 
    fSubDetectorList = NULL;
  }

  fSubDetectorList = (THashTable*) tfile->Get("SubDetList"); 
  if(!fSubDetectorList) return kFALSE;

  // Build additional lists of sensors in the Detector/SubDetector/Layer structures
  return BuildSensorList();
}

Bool_t TPhDetector::ReadInSensorGeometry(Char_t *filename)
{
  // This function should be used to load custom geometry of sensors
  TFile *tfile = new TFile(filename);
  return 
    ReadInSensorGeometry(tfile);
}

Bool_t TPhDetector::ReadInDBMap()
{
  // This function should be used after reading detector from a file, as the 
  // streamer function of TPhSensorMapDB object is not yet implemented.
  if( fSensorMapDB ) {
    delete fSensorMapDB; 
    fSensorMapDB = NULL;
  }
  Char_t SensorMapDBName[100];
  const Char_t *phathome = gSystem->Getenv("PHATHOME");
  sprintf(SensorMapDBName,"%s/init/%s",phathome,"SensorDB.root");
  return 
    ReadInDBMap(SensorMapDBName);
}

Bool_t TPhDetector::ReadInDBMap(TFile *tfile)
{
  // This function should be used to load custom sensor<->electronics map
  if(!tfile) return kFALSE;
  if(!tfile->IsOpen()) return kFALSE;
 
  if( fSensorMapDB ) {
    delete fSensorMapDB; 
    fSensorMapDB = NULL;
  }

  fSensorMapDB = (TPhSensorMapDB*) tfile->Get("SensDB"); 
  if(!fSensorMapDB) return kFALSE;
  return kTRUE;
}

Bool_t TPhDetector::ReadInDBMap(Char_t *filename)
{
  // This function should be used to load custom sensor<->electronics map
  TFile *tfile = new TFile(filename);
  return 
    ReadInDBMap(tfile);
}

Bool_t TPhDetector::ReadInFECGeometry()
{
  // This function will load standard electronics geometry description
  if( fFECDetector ) {
    delete fFECDetector; 
    fFECDetector = NULL;
  }
  Char_t FECDetectorFileName[100];
  const Char_t *phathome = gSystem->Getenv("PHATHOME");
  sprintf(FECDetectorFileName,"%s/init/%s",phathome,"FECDetector.root");
  return 
    ReadInFECGeometry(FECDetectorFileName);
}

Bool_t TPhDetector::ReadInFECGeometry(TFile *tfile)
{
  // This function should be used to load custom electronics structure
  if(!tfile) return kFALSE;
  if(!tfile->IsOpen()) return kFALSE;
 
  if( fFECDetector ) {
    delete fFECDetector; 
    fFECDetector = NULL;
  }

  fFECDetector = (TPhFECDetector*) tfile->Get("LatestFECGeometry"); 
  if(!fFECDetector) return kFALSE;
  return kTRUE;
}

Bool_t TPhDetector::ReadInFECGeometry(Char_t *filename)
{
  // This function should be used to load custom electronics structure
  TFile *tfile = new TFile(filename);
  return 
    ReadInFECGeometry(tfile);
}

Bool_t TPhDetector::ReadInCalibration()
{
  // This function will fill FECDetector with calibration parameters
  // from an ASCII file
  if( !fFECDetector ) return kFALSE;

  Char_t CalibrationFileName[100];
  const Char_t *phathome = gSystem->Getenv("PHATHOME");
  sprintf(CalibrationFileName,"%s/init/%s",phathome,"Calibration.txt");
  return 
    ReadInCalibration(CalibrationFileName);
}

Bool_t TPhDetector::ReadInCalibration(Char_t *filename)
{
  // This function should be used to init calibration from a text file
  // containing calibration parameters
  if(!fFECDetector) return kFALSE;

  return 
    fFECDetector->ReadinStringCalibrationTB98File(filename); 
}

Bool_t TPhDetector::BuildSensorList()
{
  // Build a hashtable of sensors in the detector out of the sensors
  // contained inside the subdetectors. 

  if(!fSubDetectorList)
    return kFALSE;

  fSensorList = new THashList();

  // Iterate over all subdetectorss and add the sensors to the list
  Int_t sensno;
  TPhSubDetector *subdet;
  TPhSensor *Sensor;
  TIter NextSDetector(fSubDetectorList);
  while ((subdet = (TPhSubDetector*) NextSDetector())) {
    TIter NextSensor(subdet->GetSensorIter());
    while ((Sensor = (TPhSensor*) NextSensor())) {
      sensno = Sensor->GetLabel();
      if(fSensorList->At(sensno)) {
	Error("BuildSensorList","More then one sensor with label %d\n",sensno);
	return kFALSE;
      }
      fSensorList->AddAt(Sensor,sensno);
    }
  }
  return kTRUE;
}

Bool_t TPhDetector::BuildFECDetector()
{
  // Build FEC structures as a copy of those in the sensor map data base,
  // currently connected to the Detector object
  Int_t i, j;
  Char_t fecdetname[50];
  TPhSensorMapDBFECItem *fecitem;
  TPhSensorMapDBSensorItem *sensoritem;
  TPhFEC *fec=NULL;
  Int_t FECNum, numstrings=0, strchipsize, currentchip, prevchip;
  Int_t stringsizes[kNumStrings], chipnos[kNumStrings];
  Int_t chipsizes[kMaxChipNumber*kNumStrings];
  Int_t stringno, channel;
  Char_t sensor[10];
  Int_t row, col, bondPat;

  if(fFECDetector) {
    Warning("BuildFECDetector","Old FEC structure removed, new one is being built\n");
    delete fFECDetector;
  }
  fFECDetector = new TPhFECDetector();
  sprintf(fecdetname,"FECDet_V0.0");
  fFECDetector->SetName(fecdetname);
  sprintf(fecdetname,"FEC Detector made out of SensorMap data base");
  fFECDetector->SetTitle(fecdetname);

  // Assume that there may occure no unoccupied holes in the FEC numbering
  for(i=0; i<kMaxFECs; i++) {
    FECNum = i;
    fecitem = fSensorMapDB->GetFEC(FECNum);
    if(!fecitem) continue;
    // if FEC is placed not according to its number, ERROR!
    if( fecitem->GetFEC()!=FECNum ) {
      Error("BuildFECDetector","Wrong FEC at %d position in map data base\n",i);
      return kFALSE;
    }
    prevchip = 0; currentchip = 0;
    for(j=0;j<kNumStrings;j++) {
      // this time assume no holes in string positions
      // this means first 0-length string bounds the last string
      if( !(stringsizes[j] = fecitem->GetStringLength(j)) ) break;
      stringno = j;
      channel = stringsizes[j];
      // loop to find connected channel
      while( channel>-1 &&
	     !fecitem->Lookup(stringno, channel, sensor, &row, &col, &bondPat) ) 
	channel--;
      // info for this fec/string has not been found
      if(channel<0) return kFALSE;
      // one could use bondPad value for the last channel as number of pads on a chip
      // for the chips from this string, but better is to get it directly 
      sensoritem = (TPhSensorMapDBSensorItem*) fSensorMapDB->GetSensor(sensor);
      if(sensoritem) strchipsize = sensoritem->GetBondPatSize();
      else {
	Error("BuildFECDetector","Could not determine chip size: String %d FEC %d\n",
	      stringno,FECNum);
	return kFALSE;
      }
      if(strchipsize) currentchip += (chipnos[j] = stringsizes[j]/strchipsize);
      else {
	Error("BuildFECDetector","Zero chip size: String %d FEC %d\n",
	      stringno,FECNum);
	return kFALSE;
      }
      // cause of warning left on purpose, to remind that we may want to
      // do here better
      for(prevchip; prevchip<currentchip; prevchip++) {
	chipsizes[prevchip] = strchipsize;
      }
      numstrings = j+1;
    }
    // having all data needed construct FEC object
    fec = new TPhFEC(FECNum, 
		     numstrings, stringsizes, chipnos, chipsizes);
    fFECDetector->AddFEC(fec);
  }
  return kTRUE;
}

void TPhDetector::WriteOutSensorGeometry()
{
  // This function will write out standard geometry of sensors
  Char_t SensorGeometryFileName[100];
  
  const Char_t *phathome = gSystem->Getenv("PHATHOME");
  sprintf(SensorGeometryFileName,"%s/init/%s",phathome,"SensorGeometry.root");
  WriteOutFECGeometry(SensorGeometryFileName);
  return;
}

void TPhDetector::WriteOutSensorGeometry(TFile *tfile)
{
  // This function should be used to write out custom geometry of sensors
  if(!tfile) {
    Warning("WriteOutSensorGeometry","NULL TFile pointer\n");
    return;
  }
  if(!tfile->IsOpen()) {
    Warning("WriteOutSensorGeometry","TFile could not be opened for writing\n");
    return;
  }
 
  tfile->cd();
  this->Write();
  tfile->Close();
  return;
}

void TPhDetector::WriteOutSensorGeometry(Char_t *filename)
{
  // This function should be used to write out custom geometry of sensors
  TFile *tfile = new TFile(filename);
  WriteOutSensorGeometry(tfile);
  return;
}

void TPhDetector::WriteOutDBMap()
{
  // This function should be used after reading detector from a file, as the 
  // streamer function of TPhSensorMapDB object is not yet implemented.
  // not done
  Char_t SensorDBMapFileName[100];
  
  const Char_t *phathome = gSystem->Getenv("PHATHOME");
  sprintf(SensorDBMapFileName,"%s/init/%s",phathome,"SensorDB.root");
  WriteOutFECGeometry(SensorDBMapFileName);
  return;
}

void TPhDetector::WriteOutDBMap(TFile *tfile)
{
  // This function should be used to write out custom sensor<->electronics map
  if( !fSensorMapDB ) {
    Warning("WriteOutDBMap","No SensorMapDB detector object, nothing written\n");
    return;
  }
  if(!tfile) {
    Warning("WriteOutDBMap","NULL TFile pointer\n");
    return;
  }
  if(!tfile->IsOpen()) {
    Warning("WriteOutDBMap","TFile could not be opened for writing\n");
    return;
  }
 
  tfile->cd();
  fSensorMapDB->Write();
  tfile->Close();
  return;
}

void TPhDetector::WriteOutDBMap(Char_t *filename)
{
  // This function should be used to write out custom sensor<->electronics map
  TFile *tfile = new TFile(filename, "create");
  WriteOutDBMap(tfile);
  return;
}

void TPhDetector::WriteOutFECGeometry()
{
  // This function will write out standard electronics geometry description
  Char_t FECDetectorFileName[100];
  
  const Char_t *phathome = gSystem->Getenv("PHATHOME");
  sprintf(FECDetectorFileName,"%s/init/%s",phathome,"FECDetector.root");
  WriteOutFECGeometry(FECDetectorFileName);
  return;
}

void TPhDetector::WriteOutFECGeometry(TFile *tfile)
{
  // This function should be used to write out custom electronics structure
  if( !fFECDetector ) {
    Warning("WriteOutFECGeometry","No FEC detector object, nothing written\n");
    return;
  }
  if(!tfile) {
    Warning("WriteOutFECGeometry","NULL TFile pointer\n");
    return;
  }
  if(!tfile->IsOpen()) {
    Warning("WriteOutFECGeometry","TFile could not be opened for writing\n");
    return;
  }
 
  tfile->cd();
  fFECDetector->Write("LatestFECGeometry");
  tfile->Close();
  return;
}

void TPhDetector::WriteOutFECGeometry(Char_t *filename)
{
  // This function should be used to write out custom electronics structure

  TFile *tfile = new TFile(filename,"create");
  WriteOutFECGeometry(tfile);
  return;
}


void TPhDetector::SetSensorsVerbose(Int_t verbose)
{
 // Iterates over all Sensors, and sets the Verbose status
 
   TPhSensor *sensor;

   TIter NextSensor(fSensorList);
   while((sensor = (TPhSensor *) NextSensor()) != NULL)
     sensor->SetVerbose(verbose);
}

TPhSensor *TPhDetector::GetSensor(const Int_t index) const
{
  // There are two ways of getting a TPhSensor. One is with this call, where
  // one has to provide the numeric value, or by name where one uses the 
  // sensor name to get the sensor. 

  return (TPhSensor *) fSensorList->At(index);
}

TPhSensor *TPhDetector::GetSensor(const Char_t *Name) const
{
  // There are two ways of getting a TPhSensor. One is with this call, where
  // one has to provide the name of the sensor. 
  // The other one is by providing the sensor number. 

  return (TPhSensor *) fSensorList->FindObject(Name);
}

void TPhDetector::Draw(Option_t* /* Option */)
{
  // Draw the phobos detector. This is done by calling all the subdetectors.
  // Use TPhSubDetector::TurnOn(), TPhSubDetector::TurnOff() to control what
  // subdetectors are drawn. 

  // Draw only if there is a canvas for it. 
  if(gPad) {
    fBrik = new TBRIK("MotherBRIK", "MotherBRIK", "void", 100, 100, 100);
    fNode = new TNode("MotherNode", "MotherNode", "MotherBRIK");

    fNode->cd();
    fSubDetectorList->Draw();
    fNode->SetVisibility(2);
    fNode->Draw();
  }
}

void TPhDetector::DeleteHits()
{
// For the time being, the hits are owned by the Sensor elements.
// TPhDetector::DeleteHits() loops over all sensors and asks
// them to delete the hits in their hitbuffers
// If you are absolutely sure about what you are doing,
// you can also use TPhDetector::ClearHits() to just
// clear the sensor hitlists without deleting the hits
// and then take care of deleting the hits yourself in
// some other place.

  fSubDetectorList->ForEach(TPhSubDetector,DeleteHits)();
  
}

TPhSubDetector *TPhDetector::GetSubDetector(const Char_t *Name) const
{ 
// Return a pointer to the SubDetector with the given name. 
// Use TPhDetector::ls() or TPhDetector::Print() to find out what SubDetectors
// are there

  return (TPhSubDetector*) fSubDetectorList->FindObject(Name);
}

TPhSubDetector *TPhDetector::GetSubDetectorOfSensor(const Char_t *Name) const
{
  // Returns a pointer to the Subdetector that contains the sensor of the
  // given name

  TIter NextSubDetector(fSubDetectorList);
  TPhSubDetector *SubDetector;
  while((SubDetector = (TPhSubDetector*) NextSubDetector()) != NULL)
    if(SubDetector->IsThere(Name))
      return SubDetector;
  return NULL;
}

TPhSubDetector *TPhDetector::GetSubDetectorOfSensor(const Int_t SensNum) const
{
  // Returns a pointer to the Subdetector that contains the sensor of the
  // given sensor label (a number)

  TPhSensor *Sensor = GetSensor(SensNum);
  if(Sensor)
    return GetSubDetectorOfSensor(Sensor->GetName());
  else
    return NULL;
}

void TPhDetector::ClearHits()
{
// TPhDetector::ClearHits() loops over all layers and asks
// them to clear their hitbuffers, without actually deleting
// the hits. if you don't delete the hits yourself somewhere
// else, this results in a beautiful mempry leak.

   TPhSubDetector *subdetector;

   TIter NextSubDetector(fSubDetectorList);
    while((subdetector = (TPhSubDetector *) NextSubDetector()) != NULL)
        subdetector->ClearHits();
}

void TPhDetector::ClearHitArrays()
{
  // Zero BUT DO NOT DELETE the hitarrays held inside of TPhSiSensor (such 
  // as is the case after using FillWithEventHitArrays(). Note that you need 
  // to be careful with this function, if you don't know what you are doing
  // then you will cause a memory leak!. Have a look at the destructor of 
  // TPhEvent to see how it is used. 

  TPhSiSensor *SiSensor;  

  TIter NextSensor(GetSensorList());
  while((SiSensor = (TPhSiSensor*) NextSensor()) != NULL) {
    // Make sure that the Sensor is of the TPhSiSensor kind that knows
    // about HitArrays!
    if(SiSensor->InheritsFrom("TPhSiSensor"))
      SiSensor->ClearHitArray();
    // FIXME: place holder for TOF!
  }  
}

void TPhDetector::Print() const
{
// Geometry listing. List all the SubDetectors and their layers and sensors;

  printf("Detector geometry V%s. It has the following SubDetectors:\n", 
	  GetName());
  fSubDetectorList->ForEach(TPhSubDetector,Print)();
}

void TPhDetector::ls(Option_t *Option) const
{
// Print some information about the geometry tree

  IndentLevel();
  cout << "Detector " << GetName() << " contains " 
       << fSubDetectorList->GetSize() << " subdetectors." << endl;
  TObject::IncreaseDirLevel();
  fSubDetectorList->ls(Option);
  
  TObject::DecreaseDirLevel();
}

Int_t TPhDetector::MakeHitArraysOutOfHits(Char_t *HitConName, 
					  Char_t *HitArrayConName)
{
  // Iterates over SubDetectors, to create HitArrays out of sensors' hit lists. 
  // An event object should be set in the detector to receive hitarray objects, 
  // and to provide information on hits. Hits from the HitConName event data 
  // container are automatically added to the detector sensors' hit lists with 
  // the FillWithEventHits() function.
  // The sensor hit lists are cleared just before function returns. 
                    //is this what we want?
  //
  // Return: number of created HitArrays.
 
   Int_t counter = 0;
   
   // See if we have an event set in the detector
   TPhEvent *event = GetEvent();
   if(!event) {
     if(fVerbose > 0)
       Warning("MakeHitArraysOutOfHits", 
               "No Event connected to the detector - cannot get hits!\n");
     return 0;
   }

   // Check if we have HitConName container in event. 
   TPhObjectContainer *HitContainer = NULL;
   if( HitConName ) {
     HitContainer = (TPhObjectContainer*) event->GetObject(HitConName);
     if(!HitContainer) {
       if(fVerbose > 0) 
       Warning("MakeHitArraysOutOfHits", 
               "No %s data container in the event - cannnot get hits\n", 
                HitConName);
       return 0;
     }

     // If the hit container in the event contains (MC for now)Hits, 
     // fill the detector with hits from this container.
     TIter NextObject(HitContainer->GetIterator());
     TObject *object = NextObject();
     if( object ) {
       if(strcmp(object->ClassName(),"TPhMCHit")) {
           if(fVerbose > 0) 
           Warning("MakeHitArraysOutOfHits", 
                   "%s data container does not contain (MC)Hit objects\n", 
                    HitConName);
           return 0;
       }
     }
     FillWithEventHits(HitConName);
   }

   // If there is HitArrayConName container in the event, ask for another name 
   // for this container. 
   TPhHashContainer *hitArrCon = 
     (TPhHashContainer*) event->GetObject(HitArrayConName);

   if(hitArrCon) {
     if(fVerbose > 0) 
       Warning("MakeHitArraysOutOfHits", 
	       "%s data container already present in the event - " \
	       "choose another name\n", HitArrayConName);
       return 0;
   } else {
     hitArrCon = new TPhHashContainer(HitArrayConName,
                        "HitArrays made from MonteCarlo hits");
     event->AddObject(hitArrCon);
   }

   TPhSubDetector *subdetector;

   TIter NextSubDetector(fSubDetectorList);
   while((subdetector = (TPhSubDetector *) NextSubDetector()) != NULL)
     counter += subdetector->MakeHitArraysOutOfHits(HitConName, 
						    HitArrayConName);
  
   ClearHits();
   return counter;
}

Int_t TPhDetector::MakeHitsOutOfHitArrays(Char_t *HitArrayConName,
					  Char_t *HitConName)
{
  // Iterates over sensors, (selects only BoxSiSensors for now), to create hits 
  // out hitarrays. 
  // An event object should be set in the detector to receive hit objects, 
  // and to provide information on hit arrays. HitArrays from the 
  // HitArrayConName event data container are automatically set for the detector
  // sensors with the FillWithEventHitArrays() function.
  //
  // Return: number of created hits.
 
   Int_t counter = 0;
   Char_t title[200];
   
   // See if we have an event set in the detector
   TPhEvent *event = GetEvent();
   if(!event) {
     if(fVerbose > 0)
       Warning("MakeHitsOutOfHitArrays", 
               "No Event connected to the detector - cannot get hits!\n");
     return 0;
   }

   // Check if we have HitArrayConName container in event. 
   TPhHashContainer *HitArrayContainer = NULL;
   if( HitArrayConName ) {
     HitArrayContainer = (TPhHashContainer*) event->GetObject(HitArrayConName);
     if(!HitArrayContainer) {
       if(fVerbose > 0) 
       Warning("MakeHitsOutOfHitArrays", 
               "No %s data container in the event - cannnot get hitarrays\n", 
                HitArrayConName);
       return 0;
     }

     // If the hitarray container in the event contains hitarrays, 
     // fill the detector with hitarrays from this container.
     TIter NextObject(HitArrayContainer->GetIterator());
     TObject *object = NextObject();
     if( object ) {
       if(strcmp(object->ClassName(),"TPhHitArray")) {
           if(fVerbose > 0) 
           Warning("MakeHitsOutOfHitArrays", 
                   "%s data container does not contain HitArray objects\n", 
                    HitArrayConName);
           return 0;
       }
     }
     FillWithEventHitArrays(HitArrayConName);
   }

   // If there is HitConName container in the event, ask for another name 
   // for this container. 
   TPhObjectContainer *hitCon = NULL;
   if((hitCon = (TPhObjectContainer*) event->GetObject(HitConName))) {
     if(fVerbose > 0) 
       Warning("MakeHitsOutOfHitArrays", 
	       "%s data container already present in the event - " \
	       "choose another name\n", HitConName);
       return 0;
   } else {
     sprintf(title,"Hits made from <%s> hitarrays",HitArrayContainer->GetTitle());
     hitCon = new TPhObjectContainer(HitConName,title);
     event->AddObject(hitCon);
   }

   TPhSubDetector *subdetector;

   TIter NextSubDetector(fSubDetectorList);
   while((subdetector = (TPhSubDetector *) NextSubDetector()))
     counter += subdetector->MakeHitsOutOfHitArrays(HitArrayConName, 
						    HitConName);
  
   return counter;
}

Int_t TPhDetector::MakeHitArrays(Char_t *RawHitFECConName, 
                                 Char_t *HitArrayConName)
{
  // Iterates over all TPhRawHitFEC type objects (suppressed or not) 
  // to create HitArrays. 
  //
  // Return: number of created HitArrays.
 
  Int_t counter = 0;

  if( !fEvent ) {
    if( fVerbose > 0 ) 
      Warning("MakeHitArrays",
	      "Event not set in the detector object, HitArrays not created!\n");
    return 0;
  }

  Int_t i, j;
  Int_t FEC, String, ChPos;
  Char_t SensorName[10];

  TPhSensor *sensor;
  TPhHitArray *HitArray = NULL;
  TPhAbsRawHitFEC *RawBuf;
  FECRawHit_t *Data;

  TPhAbsDataContainer *RawBufCont = 
        (TPhAbsDataContainer *) fEvent->GetObject(RawHitFECConName);
  if( !RawBufCont ) {
    if( fVerbose > 0 ) 
      Warning("MakeHitArrays",
	      "%s container not found, HitArrays not created!", 
	      RawHitFECConName);
    return 0;
  }

  TPhHashContainer *HitArrayCont = 
    (TPhHashContainer*) fEvent->GetObject(HitArrayConName);
  if( HitArrayCont ) {
    if( fVerbose > 0 ) 
      Warning("MakeHitArrays", "%s container is already present." \
        "It is deleted and a new one is created!\n", HitArrayConName);
    fEvent->RemoveObject(HitArrayCont);
    HitArrayCont->Delete();
  }

  Char_t title[100];
  sprintf(title,"HitArrays made from %s container objects",RawHitFECConName);
  HitArrayCont = new TPhHashContainer(HitArrayConName,title);
  fEvent->AddObject(HitArrayCont);

  TIter NextFEC(RawBufCont->GetIterator());

  // Loop over all FECs
  while((RawBuf = (TPhAbsRawHitFEC *) NextFEC()) != NULL) {
    FEC = RawBuf->GetFECNum();
    for(String = 0; String < 8; String++) {
      Data = RawBuf->GetRawHitArray(String);
      for(ChPos = 0; ChPos < RawBuf->GetRawHitArraySize(String); ChPos++) {
        if(fSensorMapDB->Lookup(FEC, String, ChPos, SensorName, &i, &j)) {
          Int_t ADC = Data[ChPos].ADC;
          HitArray = NULL;
          HitArray = (TPhHitArray*) HitArrayCont->GetObject(SensorName);
          if( !HitArray ) {
            sensor = GetSensor(SensorName);
	    if( sensor ) {
	      if( sensor->InheritsFrom("TPhSiSensor") ) {
                 HitArray = new TPhHitArray(sensor, SensorName, SensorName);
                 HitArrayCont->Add(HitArray);
	         counter++;
	      } else {
                  if( fVerbose ) 
		    Warning("MakeHitArrays",
		    "Attempt to make HitArrays for a non TPhSiSensor object!");
                  sensor = NULL;
	      }
	    } else {
                  if( fVerbose ) 
                      Warning("MakeHitArrays",
			      "Sensor name %s given by TPhSensorMapDB::Lookup" \
			      "not found in the geometry!");
	    }
	  }
          if( HitArray ) HitArray->SetADC( i, j, ADC );
        } else {
          if( fVerbose ) Warning("MakeHitArrays",
				 "Couldn't find sensor,row,column numbers" \
				 "for FEC=%d, String=%d, Chan=%d\n",
				 FEC,String,ChPos);
	}
      }
    }
  }
  TIter NextHitArray(HitArrayCont->GetIterator());
  while ((HitArray = (TPhHitArray *)NextHitArray()) != NULL) 
    HitArray->CalcNHits();

  return counter;
}

Int_t TPhDetector::MakeHits(Char_t *RawHitFECn0ConName, Char_t *HitConName,
			    Char_t *PedFileName, Char_t *CalFileName)
{
 // Iterates over all TPhRawHitFECn0 objects to create Hits. 
 //
 // Return: number of created Hits.
 
  Int_t counter = 0;

  if( !fEvent ) {
    if( fVerbose > 0 ) 
      Warning("MakeHits",
	      "Event not set in the detector object, Hits not created!\n");
    return 0;
  }

  Int_t i, j;
  Double_t hx,hy,hz;
  Int_t FEC, String, ChPos;
  Char_t SensorName[10];
  Int_t ADC;

  TPhSensor *sensor;
  TPhHit *Hit = NULL;
  TPhRawHitFECn0 *RawBuf = NULL;

  TPhAbsDataContainer *RawBufCont = 
        (TPhAbsDataContainer *) fEvent->GetObject(RawHitFECn0ConName);
  if( !RawBufCont ) {
    if( fVerbose > 0 ) 
      Warning("MakeHits",
	      "%s container not found, Hits not created!", RawHitFECn0ConName);
    return 0;
  }

  TPhObjectContainer *HitCont = 
    (TPhObjectContainer*) fEvent->GetObject(HitConName);
  if( HitCont ) {
    if( fVerbose > 0 ) 
      Warning("MakeHits", "%s container is already present." \
        "It is deleted and a new one is created!\n", HitConName);
    fEvent->RemoveObject(HitCont);
    HitCont->Delete();
  }

  Char_t title[100];
  sprintf(title,"Hits made from %s container objects",RawHitFECn0ConName);
  HitCont = new TPhObjectContainer(HitConName,title);
  fEvent->AddObject(HitCont);

  // Create ADC->E object and calibrate it with values found in files
  TPhEDigi *digiToE = new TPhEDigi();

  Short_t pedestal;
  if(PedFileName) {
    FILE *pedfile = fopen(PedFileName,"r");
    if(pedfile) {
      if(fscanf(pedfile,"%hd",&pedestal)) digiToE->SetPedestal(pedestal);
      fclose(pedfile);
    }
    else
      if(fVerbose) 
	Warning("MakeHits", "File with pedestal cut-off can not be opened.");
  }
  Double_t erange, energy;
  if(CalFileName) {
    FILE *calfile = fopen(CalFileName,"r");
    if(calfile) {
      if(fscanf(calfile,"%f",&erange)) digiToE->SetFullERange(erange);
      fclose(calfile);
    }
    else
      if(fVerbose) 
	Warning("MakeHits", "File with calibration info can not be opened.");
  }

  TIter NextFEC(RawBufCont->GetIterator());

  // Loop over all FECs
  while((RawBuf = (TPhRawHitFECn0*) NextFEC())) {
    FEC = RawBuf->GetFECNum();
    for(String = 0; String < 8; String++) {
      FECRawHit_t *Data = RawBuf->GetRawHitArray(String);
      for(ChPos = 0; ChPos < RawBuf->GetRawHitArraySize(String); ChPos++) {
        if(fSensorMapDB->Lookup(FEC, String, ChPos, SensorName, &i, &j)) {
          ADC = Data[ChPos].ADC;
          Hit = NULL;
	  sensor = GetSensor(SensorName);
	  if( sensor ) {
	    if( sensor->InheritsFrom("TPhSiSensor") ) {
	      Hit = new TPhHit();
              Hit->SetSensorLabel(sensor->GetLabel());
              Hit->SetLayer(sensor->GetLayer());
	      // BoxSiSensor for now, since we have no localization function elsewhere
	      if( sensor->InheritsFrom("TPhBoxSiSensor") ) {
		((TPhBoxSiSensor *) sensor)->GetCoords(j,i,&hx,&hy,&hz);
		Hit->SetX((Float_t) hx);
		Hit->SetY((Float_t) hy);
		Hit->SetZ((Float_t) hz);
	      }
              Hit->SetPadX(i);
              Hit->SetPadY(j);
	      energy = digiToE->ADCtoE(ADC);
	      if(energy>=0) {
		Hit->SetDe(energy);
		// don't know how to interprete time data
		Hit->SetDt(0);
		HitCont->Add(Hit);
		counter++;
	      } else delete Hit;
	    } else {
	      if( fVerbose ) 
		Warning("MakeHits",
			"Attempt to make Hits for a non TPhSiSensor object!");
	      sensor = NULL;
	    }
	  } else {
	    if( fVerbose ) 
	      Warning("MakeHits","Sensor name %s given by TPhSensorMapDB::Lookup" \
		      " not found in the geometry!");
	  }
        } else {
          if( fVerbose ) Warning("MakeHits", "Couldn't find sensor,row,column" \
				 "numbers for FEC=%d, String=%d, Chan=%d\n",
				 FEC,String,ChPos);
	}
      }
    }
  }

  delete digiToE;
  return counter;
}

void TPhDetector::AddHit(TPhHit *Hit)
{
  // Add a hit into the detector. This member function finds the 
  // appropriate sensor and then calls the sensor AddHit member. 

  TPhSensor *sensor = GetSensor(Hit->GetSensorLabel());
  // if the sensor is not defined, just return. 
  if(!sensor) {
    if(fVerbose) 
      Warning("AddHit", "Can't find sensor %s\n", Hit->GetSensorLabel());
    return;
  }

  sensor->AddHit(Hit);
}

void TPhDetector::FillWithEventHits(Char_t *HitConName)
{
  // Add hits from the container of the event set in the detector
  // to the Sensors

  if(!fEvent) return;

  TPhObjectContainer *ObjectContainer = (TPhObjectContainer *) 
    fEvent->GetObject(HitConName);

  if(!ObjectContainer)
    return;

  // Loop over all sensors and clear the HitLists, to ensure that there
  // are no remnants from the last event!
  TPhSensor *Sensor;
  TIter NextSensor(GetSensorList());
  while((Sensor = (TPhSensor*) NextSensor()) != NULL)
    Sensor->ClearHits();
  
  TIter NextHit(ObjectContainer->GetIterator());
  TPhHit *hit;
  while ((hit = (TPhHit*) NextHit()) != NULL) 
    AddHit(hit);
}

void TPhDetector::MakeHitLists(const Char_t *HitsConName)
{
  // Make the hit lists in the sensors by calling the 
  // TPhSiBoxSensor::MakeHitList() member function, when looping over
  // all sensors. The hits are made from
  // HitArrays already attached to the sensors with something like 
  // TPhDetector::FillWithEventHitArrays() - DO NOT FORGET THIS! the hits
  // are also added to a new container name called HitsConName.

  //FIXME: this should be extended to all the silicon sensors! 

  TPhSensor *Sensor;

  // Loop over all sensors, check that it is a TPhBoxSiSensor
  TIter NextSensor(GetSensorList());
  while((Sensor = (TPhSensor*) NextSensor()) != NULL) {
    //FIXME: const const const !!!!

    if(Sensor->InheritsFrom(TPhBoxSiSensor::Class()->GetName())) 
      ((TPhBoxSiSensor*)Sensor)->MakeHitList((Char_t*)HitsConName);
  }
}

void TPhDetector::FillWithEventHitArrays(Char_t *HitArrConName)
{
  // Add hit arrays from the container of the event set in the detector to
  // the Sensors
  
  if(!fEvent) return;

  TPhHashContainer *HashContainer = (TPhHashContainer *) 
    fEvent->GetObject(HitArrConName);

  if(!HashContainer)
    return;

  TPhSiSensor *SiSensor;  
  TPhHitArray *HitArray;

  TIter NextSensor(GetSensorList());
  while((SiSensor = (TPhSiSensor*) NextSensor()) != NULL) {
    // Make sure that the Sensor is of the TPhSiSensor kind that knows
    // about HitArrays!
    if(SiSensor->InheritsFrom("TPhSiSensor")) {
      HitArray = (TPhHitArray*)HashContainer->GetObject(SiSensor->GetName());
      // If there is no such HitArray in the event, then we store NULL
      SiSensor->AddHitArray(HitArray);
    }
  }  
}

Int_t TPhDetector::MakeRawHitFECn0(Char_t *HitArrayConName, 
				   Char_t *RawHitFECn0ConName)
{
  // Iterates over all sensors from turned on subdetectors, 
  // to create the RawHitFECn0 objects. 
  //
  // Return: number of created RawHitFECn0's.

  Int_t counter = 0;

  Int_t i, j;
  Int_t FEC, String, ChPos;
  Char_t FECName[10];

  TPhAbsDataContainer *hitArrayCont;
  TPhHitArray *HitArray;
  TPhRawHitFECn0 *RawBuf;
  FECRawHit_t Hit;

  if( !fEvent ) {
    if( fVerbose > 0 ) 
      Warning("MakeHitArrays", "Event not set in the detector object; "\
	            "HitArrays not created!\n");
    return 0;
  }

  hitArrayCont = (TPhAbsDataContainer*) fEvent->GetObject(HitArrayConName);
  if( !hitArrayCont ) {
    if( fVerbose > 0 ) 
       Warning("MakeRawHitFECn0",
               "%s container not found, HitArrays not created!",
                HitArrayConName);
    return 0;
  }

  TPhHashContainer *RawBufCont = 
      (TPhHashContainer*) fEvent->GetObject(RawHitFECn0ConName);
  if( RawBufCont ) {
    if( fVerbose > 0 ) 
       Warning("MakeRawHitFECn0","%s container is already present" \
	       "It is deleted and a new one is created!\n", 
	       RawHitFECn0ConName);
    fEvent->RemoveObject(RawBufCont);
    RawBufCont->Delete();
  }
  Char_t title[100];
  sprintf(title,"RawHitFECn0 buffers made from %s container objects",
                 HitArrayConName);
  RawBufCont = new TPhHashContainer(RawHitFECn0ConName,title);
  fEvent->AddObject(RawBufCont);

  TIter NextHitArray(hitArrayCont->GetIterator());

  while((HitArray = (TPhHitArray*) NextHitArray()) != NULL) {
    for(i = 0; i < HitArray->GetRows(); i++) {
      for(j = 0; j < HitArray->GetColumns(); j++) {
        if(fSensorMapDB->Lookup((Char_t *)HitArray->GetName(), i, j, 
                                 &FEC, &String, &ChPos)) {
	  sprintf(FECName, "FEC_%d", FEC);
#ifdef DEBUG
          printf("%s %d %d %d %d %d\n", 
		 HitArray->GetName(), i, j, FEC, String, ChPos);
#endif
          if( fVerbose ) {
             if(ChPos > 768)
                printf("Channel no problem: %s %d %d %d %d %d\n", 
                        HitArray->GetName(), i, j, FEC, String, ChPos);
               if(FEC < 0) 
                  printf("FEC negative problem: %s %d %d %d %d %d\n", 
                        HitArray->GetName(), i, j, FEC, String, ChPos);
	  }
          RawBuf = (TPhRawHitFECn0*) RawBufCont->GetObject(FECName);
          if(RawBuf == NULL) {
            if( fVerbose > 1 ) printf("Making FEC# %d\n", FEC);
            RawBuf = new TPhRawHitFECn0(FEC);
            RawBufCont->Add(RawBuf);
            counter++;
          }
          Hit.ADC = HitArray->GetADC(i, j);
          Hit.Channel = ChPos;
          RawBuf->Add(String, &Hit);
        }
      }
    }
  }
  return counter;
}

Bool_t TPhDetector::MakeRawHitFECs0(Char_t *RawHitFECn0ConName, 
				    Char_t *RawHitFECs0ConName)
{
  // Iterates over all TPhRawHitFECn0 objects, does zero suppression, 
  // and creates the RawHitFECs0 objects. 
  //
  // Return: status.

  Bool_t status=kTRUE;
  Int_t fechits=0, strhits;
  Int_t hitstrlength;
  Int_t FEC, String;

  TPhString *str;
  TPhHashContainer *fecn0Cont;
  TPhHashContainer *fecs0Cont;
  TPhRawHitFECn0 *rawBufn0;
  TPhRawHitFECs0 *rawBufs0;
  FECRawHit_t *rawhitstring;

  if( !fEvent ) {
    if( fVerbose > 0 ) 
       Error("MakeRawHitFECs0", "event not set in the detector, no input data !");
    return kFALSE;
  }
  if( !fFECDetector ) {
    if( fVerbose > 0 ) 
       Error("MakeRawHitFECs0", "FECDetector info missing !");
    return kFALSE;
  }

  fecn0Cont = (TPhHashContainer*) fEvent->GetObject(RawHitFECn0ConName);
  if( !fecn0Cont ) {
    if( fVerbose > 0 ) 
       Warning("MakeRawHitFECs0", "%s container not found !",
                RawHitFECn0ConName);
    return kFALSE;
  }

  // Compatibility check here
  // commented out for TestBeam98
  //  if( !CompatibilityCheck(RawHitFECn0ConName) ) {
  //    if( fVerbose > 0 ) 
  //       Error("MakeRawHitFECs0", "%s raw hits not compatible with present FECDetector !",
  //	     fecn0ConName);
  //    return kFALSE;
  //  }

  fecs0Cont = 
      (TPhHashContainer*) fEvent->GetObject(RawHitFECs0ConName);
  if( fecs0Cont ) {
    if( fVerbose > 0 ) 
       Warning("MakeRawHitFECs0","%s container is already present" \
	       "It is deleted and a new one is created!\n", 
	       RawHitFECs0ConName);
    fEvent->RemoveObject(fecs0Cont);
    delete fecs0Cont;
  }
  Char_t title[100];
  sprintf(title,"RawHitFECs0 buffers made from %s container objects",
                 RawHitFECn0ConName);
  fecs0Cont = new TPhHashContainer(RawHitFECs0ConName,title);
  fEvent->AddObject(fecs0Cont);

  TIter NextRawBufn0(fecn0Cont->GetIterator());
  while((rawBufn0 = (TPhRawHitFECn0*) NextRawBufn0())) {
    fechits = 0;
    FEC = rawBufn0->GetFECNum();
    rawBufs0 = new TPhRawHitFECs0(FEC);
    for(String=0; String<kNumStrings; String++) {
      hitstrlength = rawBufn0->GetRawHitArraySize(String);
      if( !hitstrlength ) continue;
      rawhitstring = rawBufn0->GetRawHitArray(String);
      str = fFECDetector->GetString(FEC,String);
      // strings of raw hits are separated here, fed into TPhString object
      // for zero-suppression. 
      // The TPhRawHitFECs0 buffer is filled at the string level 
      // and added to the RawHitFECs0ConName container only if non-empty
      if(str) {
	strhits = str->ZeroSuppress(rawhitstring, rawBufs0);
	if(strhits >= 0) fechits += strhits;
	else             status   = kFALSE;
      } else {
	Error("MakeRawHitFECs0","Unknown string structure in FEC no %d\n",FEC);
	status = kFALSE;
      }
    }
    if(fechits) fecs0Cont->Add(rawBufs0);
    else delete rawBufs0;
  }
  return status;
}

Bool_t TPhDetector::MakeCalRawHitFECs0(Char_t *RawHitFECs0ConName, 
				       Char_t *CalRawHitFECs0ConName)
{
  // Iterates over all TPhRawHitFECs0 objects, makes their copy
  // and calibrates this copy. 
  //
  // Return: status.

  Bool_t status=kTRUE;

  TPhHashContainer *fecs0Cont;
  TPhHashContainer *calfecs0Cont;

  if( !fEvent ) {
    if( fVerbose > 0 ) 
       Error("CalibrateRawHitFECs0", "event not set in the detector, no input data !");
    return kFALSE;
  }
  if( !fFECDetector ) {
    if( fVerbose > 0 ) 
       Error("CalibrateRawHitFECs0", "FECDetector info missing !");
    return kFALSE;
  }

  fecs0Cont = (TPhHashContainer*) fEvent->GetObject(RawHitFECs0ConName);
  if( !fecs0Cont ) {
    if( fVerbose > 0 ) 
       Warning("CalibrateRawHitFECs0", "%s container not found !",
                RawHitFECs0ConName);
    return kFALSE;
  }

  calfecs0Cont = 
      (TPhHashContainer*) fEvent->GetObject(CalRawHitFECs0ConName);
  if( calfecs0Cont ) {
    if( fVerbose > 0 ) 
       Warning("CalibrateRawHitFECs0","%s container is already present" \
	       "It is deleted and a new one is created!\n", 
	       CalRawHitFECs0ConName);
    fEvent->RemoveObject(calfecs0Cont);
    delete calfecs0Cont;
  }

  // A new copy of the zero-suppressed container will be made by a Clone() function
  // and will have name and title changed.
  Char_t title[100];
  sprintf(title,"Calibrated RawHitFECs0 buffers made from %s container objects",
                 RawHitFECs0ConName);
  calfecs0Cont  = (TPhHashContainer *) fecs0Cont->Clone(); 
  calfecs0Cont->SetName(CalRawHitFECs0ConName);
  calfecs0Cont->SetTitle(title);
  fEvent->AddObject(calfecs0Cont);

  //  status = Calibrate(CalRawHitFECs0ConName); 
  status = CalibrateOnStrings(CalRawHitFECs0ConName); 

  return status;
}

Bool_t TPhDetector::PedPreProcess(Char_t *rawContainerName)
{
  // PedPreProcess on the whole event raw container
  Bool_t currstatus, status=kTRUE;
  TPhHashContainer *rawcontainer=NULL;
  TPhRawHitFECn0* fecevent=NULL;
  if( !fEvent ) {
    Error("PedPreProcess","No event connected to the detector\n");
    return kFALSE;
  }
  if( !(rawcontainer=(TPhHashContainer *) fEvent->GetObject(rawContainerName)) ) {
    Error("PedPreProcess","Raw data container %s not found\n",rawContainerName);
    return kFALSE;
  }
  TIter nextFECEvent(rawcontainer->GetIterator());
  while( (fecevent=(TPhRawHitFECn0 *) nextFECEvent()) ) {
    currstatus = fFECDetector->PedPreProcess(fecevent);
    status = status && currstatus;
  }
  return status;
}

Bool_t TPhDetector::PedProcess(Char_t *rawContainerName)
{
  // PedProcess on the whole event raw container
  Bool_t currstatus, status=kTRUE;
  TPhHashContainer *rawcontainer=NULL;
  TPhRawHitFECn0* fecevent=NULL;
  if( !fEvent ) {
    Error("PedProcess","No event connected to the detector\n");
    return kFALSE;
  }
  if( !(rawcontainer=(TPhHashContainer *) fEvent->GetObject(rawContainerName)) ) {
    Error("PedProcess","Raw data container %s not found\n",rawContainerName);
    return kFALSE;
  }
  TIter nextFECEvent(rawcontainer->GetIterator());
  while( (fecevent=(TPhRawHitFECn0 *) nextFECEvent()) ) {
    currstatus = fFECDetector->PedProcess(fecevent);
    status = status && currstatus;
  }

  return status;
}

Bool_t TPhDetector::CMNProcess(Char_t *rawContainerName)
{
  // CMN6SigmaProcess on the whole event raw container
  Bool_t currstatus, status=kTRUE;
  TPhHashContainer *rawcontainer=NULL;
  TPhRawHitFECn0* fecevent=NULL;
  if( !fEvent ) {
    Error("CMN6SigmaProcess","No event connected to the detector\n");
    return kFALSE;
  }
  if( !(rawcontainer=(TPhHashContainer *) fEvent->GetObject(rawContainerName)) ) {
    Error("CMN6SigmaProcess","Raw data container %s not found\n",rawContainerName);
    return kFALSE;
  }
  TIter nextFECEvent(rawcontainer->GetIterator());
  while( (fecevent=(TPhRawHitFECn0 *) nextFECEvent()) ) {
    currstatus = fFECDetector->CMNProcess(fecevent);
    status = status && currstatus;
  }

  return status;
}

Bool_t TPhDetector::NoiseProcess(Char_t *rawContainerName)
{
  // NoiseProcess on the whole event raw container
  Bool_t currstatus, status=kTRUE;
  TPhHashContainer *rawcontainer=NULL;
  TPhRawHitFECn0* fecevent=NULL;
  if( !fEvent ) {
    Error("NoiseProcess","No event connected to the detector\n");
    return kFALSE;
  }
  if( !(rawcontainer=(TPhHashContainer *) fEvent->GetObject(rawContainerName)) ) {
    Error("NoiseProcess","Raw data container %s not found\n",rawContainerName);
    return kFALSE;
  }
  TIter nextFECEvent(rawcontainer->GetIterator());
  while( (fecevent=(TPhRawHitFECn0 *) nextFECEvent()) ) {
    currstatus = fFECDetector->NoiseProcess(fecevent);
    status = status && currstatus;
  }

  return status;
}

Bool_t TPhDetector::ZeroSuppress(Char_t *rawContainerName)
{
  // ZeroSuppress on the whole event raw container
  Bool_t currstatus, status=kTRUE;
  TPhHashContainer *rawcontainer=NULL;
  TPhRawHitFECn0* fecevent=NULL;
  if( !fEvent ) {
    Error("ZeroSuppress","No event connected to the detector\n");
    return kFALSE;
  }
  if( !(rawcontainer=(TPhHashContainer *) fEvent->GetObject(rawContainerName)) ) {
    Error("ZeroSuppress","Raw data container %s not found\n",rawContainerName);
    return kFALSE;
  }
  TIter nextFECEvent(rawcontainer->GetIterator());
  while( (fecevent=(TPhRawHitFECn0 *) nextFECEvent()) ) {
    currstatus = fFECDetector->ZeroSuppress(fecevent);
    status = status && currstatus;
  }

  return status;
}

Bool_t TPhDetector::Calibrate(Char_t *rawContainerName)
{
  // calibrate the whole event raw container
  Bool_t currstatus, status=kTRUE;
  TPhHashContainer *rawcontainer=NULL;
  TPhAbsRawHitFEC* fecevent=NULL;
  if( !fEvent ) {
    Error("Calibrate","No event connected to the detector\n");
    return kFALSE;
  }
  if( !(rawcontainer=(TPhHashContainer *) fEvent->GetObject(rawContainerName)) ) {
    Error("Calibrate","Raw data container %s not found\n",rawContainerName);
    return kFALSE;
  }
  TIter nextFECEvent(rawcontainer->GetIterator());
  while( (fecevent=(TPhRawHitFECn0 *) nextFECEvent()) ) {
    currstatus = fFECDetector->Calibrate(fecevent);
    status = status && currstatus;
  }

  return status;
}

Bool_t TPhDetector::CalibrateOnStrings(Char_t *rawContainerName)
{
  // calibrate on the whole event raw container
  Bool_t currstatus, status=kTRUE;
  TPhHashContainer *rawcontainer=NULL;
  TPhAbsRawHitFEC* fecevent=NULL;
  if( !fEvent ) {
    Error("Calibrate","No event connected to the detector\n");
    return kFALSE;
  }
  if( !(rawcontainer=(TPhHashContainer *) fEvent->GetObject(rawContainerName)) ) {
    Error("Calibrate","Raw data container %s not found\n",rawContainerName);
    return kFALSE;
  }
  TIter nextFECEvent(rawcontainer->GetIterator());
  while( (fecevent=(TPhRawHitFECn0 *) nextFECEvent()) ) {
    currstatus = fFECDetector->CalibrateOnStrings(fecevent);
    status = status && currstatus;
  }

  return status;
}

Int_t TPhDetector::MakeHitArrays(Char_t *hitArrayConName)
{
  // Make and fill hit arrays with raw hits left after zero-suppression
  // Suppressed raw data are taken from buffers in FECDetector structures
  // previously filled with a call to ZeroSuppress() method.
  if( !fEvent ) {
    Error("MakeHitArrays","No event connected to the detector\n");
    return 0;
  }
  TPhHashContainer *HitArrayCont = 
    (TPhHashContainer*) fEvent->GetObject(hitArrayConName);
  if( HitArrayCont ) {
    if( fVerbose > 0 ) 
      Warning("MakeHitArrays", "%s container is already present." 
        "It is deleted and a new one is created\n", hitArrayConName);
    fEvent->RemoveObject(HitArrayCont);
    HitArrayCont->Delete();
  }

  Char_t title[100];
  sprintf(title,"HitArrays made from (zero-suppressed) FECDetector buffers");
  HitArrayCont = new TPhHashContainer(hitArrayConName,title);
  fEvent->AddObject(HitArrayCont);

  fFECDetector->MakeHitArrays(HitArrayCont);
  TPhHitArray * HitArray;
  TIter NextHitArray(HitArrayCont->GetIterator());
  while ( (HitArray = (TPhHitArray *)NextHitArray()) ) HitArray->CalcNHits();

  return HitArrayCont->GetEntries();
}

Bool_t TPhDetector::CompatibilityCheck(Char_t *rawContainerName)
{
  // Check compatibility between event raw hit buffers and the 
  // current FEC detector object
  TPhHashContainer* container = 
    (TPhHashContainer *) fEvent->GetObject(rawContainerName);

  return fFECDetector->CompatibilityCheck(container);
}

Bool_t TPhDetector::CompatibilityCheck(TPhRawHitFECn0* rawhit)
{
  // Check compatibility between the given raw hit buffer and the 
  // current FEC detector object

  return fFECDetector->CompatibilityCheck(rawhit);
}

#undef DEBUG
