#include "TPhSensor.h"
#include "TPhDetector.h"
#include "TPhHit.h"
#include "TPhEvent.h"

#include "TMath.h"
#include "TGeometry.h"
#include "TRotMatrix.h"

ClassImp(TPhSensor)

///////////////////////////////////////////////////////////////////////////
//
// This is the Base class for all Phobos Sensor Classes. 
// The actual sensor classes will typically inherit from a class 
// that describes the geometry (such as TPhBoxSensor or 
// TPhRingSensor). These geometry classes inherit from this base class.
//
// This class for instance deals with how to make HitArrays out of 
// RawHitBuffers and vice-versa for Fake Data creation. 
// 
// $Id: TPhSensor.cxx,v 1.6 1999/01/12 22:35:48 decowski Exp $
//
///////////////////////////////////////////////////////////////////////////

TPhSensor::TPhSensor() : TNode()
{
// Constructor for TPhSensor
  fLabel = 0;
  fWing = 0;
  fLayer = 0;
  fVerbose    = 0;
  fHitList    = NULL;
  fDetector   = TPhDetector::GetInstance();

}

TPhSensor::~TPhSensor()
{
// Destructor for TPhSensor. 

  SafeDelete(fMatrix);

  //Clear fHitList without deleting hits (they are owned by the event object).
  if(fHitList) {
    fHitList->Clear();
    delete fHitList;
  }
}


void TPhSensor::DeleteHits()        
{
  // Resets the hits in the hitlist and also DELETES the hits. 

  if(fHitList) {
    fHitList->Delete();  
    delete fHitList;
  }
  fHitList = NULL;
}

void TPhSensor::ClearHits()        
{
  // Resets the array containing the Hits in the sensor. DOES NOT delete
  // the hits!

  if(fHitList) {
    fHitList->Clear();
    delete fHitList;
  }
  fHitList = NULL;
}

void TPhSensor::SetRotMtx(Double_t theta1, Double_t phi1, 
			  Double_t theta2, Double_t phi2, 
			  Double_t theta3, Double_t phi3)
{
  // Set the rotation matrix of the sensor. The angles have to be in degrees.
  
  Char_t name[40];
  sprintf(name, "Rot_%s", GetName());
  fMatrix = new TRotMatrix(name, name, theta1, phi1, theta2, 
			   phi2, theta3, phi3);
}

void TPhSensor::Draw(Option_t * /* option */)
{
  // The default sensor drawing method

  SetLineColor(kBlack);
  fParent = gGeometry->GetCurrentNode();
  if(fParent) {
    fParent->BuildListOfNodes();
    fParent->GetListOfNodes()->Add(this);
  } else
    Warning("Draw", "No Mother Volume Set!\n");

}

void TPhSensor::Print()
{
  // List TPhSpectSensor object info

   printf("Sensor %s no %d layer %d \n", GetName(), GetLabel(), GetLayer());       
}


void TPhSensor::AddHit(TPhHit *hit)
{
  // Add a Hit to this sensor
    if(!fHitList)
        fHitList = new TObjArray();
    fHitList->Add(hit);
}

void TPhSensor::Streamer(TBuffer &R__b)
{
  // Stream an object of class TPhSensor.
  // Adjusted for ClassVersion 1 & 2. 

  // Stream our old data into this dummy variable
  Double_t dummy;

  if (R__b.IsReading()) {
    Version_t R__v = R__b.ReadVersion(); 
    TNode::Streamer(R__b);
    R__b >> fLabel;
    R__b >> fLayer;
    R__b >> fWing;

    // Take care of ClassVersion 1 where there was an 'efficiency' variable
    if (R__v == 1)
      R__b >> dummy;
  } else {
    R__b.WriteVersion(TPhSensor::IsA());
    TNode::Streamer(R__b);
    R__b << fLabel;
    R__b << fLayer;
    R__b << fWing;
  }
}




