#include "TPhSpectSensor.h"
#include "TPhHit.h"
#include "TPhMCHit.h"
#include "TPhHitArray.h"
#include "TPhObjectContainer.h"
#include "TPhEDigi.h"
#include "TPhEvent.h"
#include "TPhDetector.h"

#include "TMath.h"
#include "TVirtualPad.h"
#include "TInterpreter.h"
#include "TROOT.h"

#include <math.h>

ClassImp(TPhSpectSensor)

////////////////////////////////////////////////////////////////////////////
//                                                                        //
// TPhSpectSensor handles the management of geometry information          // 
// for one Spectrometer detector element. It also provides                //
// functionality for the TPhSwim detector simulation class.               //
// The position of the sensors is specified by the coordinates            //
// of two corners of the sensor in the PHOBOS coordinate system.          //
// The sensor also contains two data members specifying the               //
// pixel width and pixel height.                                          //
// The position and pixel information are used in several                 //
// member functions of the TPhSpectSensor class.                          //
// TPhSpectSensor::DistToSensor(TPhHit *) uses the position information   //
// of the sensor to calculate the distance of a hit from the              //
// sensor. This is used in the TPhSwim class. Also used in TPhSwim        //
// is the TPhSpectSensor::Pixelize(TPhHit *) call, which takes a hit      //
// on the sensor, pixelises the hit coordinates and returns a             //
// modified hit. This call also does hit merging.                         //
//                                                                        //
// At some point this class should be integrated with the                 //
// geometry package inside ROOT (something for '97)                       //
//                                                                        //
//
// $Id: TPhSpectSensor.cxx,v 1.9 1999/03/10 12:38:57 decowski Exp $
//
////////////////////////////////////////////////////////////////////////////

#define dEBUG

TPhSpectSensor::TPhSpectSensor()
{
// Constructor for TPhSpectSensor.

  SetTitle("Spectrometer Sensor");
}

TPhSpectSensor::~TPhSpectSensor()
{
// Destructor for TPhSpectSensor. 

}

void TPhSpectSensor::Draw(Option_t *option)
{ 
  // Draw the Spectrometer elements in Magenta

  TPhSensor::Draw(option); 
  SetLineColor(kMagenta); 
}

void TPhSpectSensor::MoveHitToSensor(TPhMCHit *hit)
{
// Move hit from to position after last Runge-Kutta
// step to position on the surface of the sensor.

    double a;
    double b;
    double c;
    double d;
    double x;    
    double z;


    a = (GetZ1() - GetZ0())/(GetX1() - GetX0());
    b = GetZ0() - GetX0() * a;
    c = tan(hit->GetPsi());
    d = hit->GetZ() - hit->GetX() * c;

#ifdef DEBUGMV
    printf("c,d = %g %g for hit at %g %g\n",c,d,hit->GetX(),hit->GetZ());
#endif

    x = (d-b)/(a-c);
    z = a*x + b;

    hit->SetX(x);
    hit->SetZ(z);

#ifdef DEBUGMV
    printf("Intersection at x,z = %g %g \n",x,z);
#endif
}

void TPhSpectSensor::CalcDistVar()
{
// Calculate some variables that are needed to 
// determine the distance of a point to the 
// sensor (projected into the X-Z LayerNumber)

    Double_t m;
    Double_t b;
    Double_t dx;
    Double_t dz;

    if(GetX0()-GetX1() != 0)
      m = (GetZ0()-GetZ1())/(GetX0()-GetX1());
    else 
      m = 0;

    b = GetZ0() - GetX0()*m;

    fA = m;
    fB = -1;
    fC = b;
    fD = sqrt(fA*fA+fB*fB);

    dx = GetX0() - GetX1();
    dz = GetZ0() - GetZ1();
    fLen2 =  dx*dx+dz*dz;

    if(fVerbose)
       printf("%s: %g %g %g %g %g\n",GetName(),fA,fB,fC,fD,fLen2);

}

Double_t TPhSpectSensor::DistToSensor(Double_t *x) 
{
// Calculate distance of hit to surface of the 
// sensor element. Used in the TPhSwim class.
// Here we calculate only the projected distance
// in the x-z LayerNumber.

    Double_t dx;
    Double_t dz;
    Double_t d1;
    Double_t d2;

    dx = x[0] - GetX0();
    dz = x[2] - GetZ0();
    d1 = dx*dx+dz*dz;

    dx = x[0] - GetX1();
    dz = x[2] - GetZ1();
    d2 = dx*dx+dz*dz;
 
    if(d1 > fLen2 || d2 > fLen2)
    {
        if(d1>d2)
            return sqrt(d2);
        else
            return sqrt(d1);
    }
    else
    {
        return (fA*x[0]+fB*x[2]+fC)/fD;
    }
}

void TPhSpectSensor::Print()
{
  // List TPhSpectSensor object info

   printf("Sensor %s no %d layer %d - %d rows %d columns\
     (%f %f %f) (%f %f %f)\
     Thickness %f Pix width %f Pix height %f\n",
     GetName(), GetLabel(), GetLayer(), fRows, fColumns,
     GetX0(), GetY0(), GetZ0(), GetX1(), GetY1(), GetZ1(),
     fThickness, GetPadWidthX(), GetPadWidthY());       
}

void TPhSpectSensor::DisplayDialog()
{
// This displays a dialog Canvas that can be used to make a certain selection 
// of hitstograms that the user might want to see. Still in development. 

  /*   if (gPad) gROOT->SetSelectedPad(gPad->GetSelectedPad());

   TList *lc = (TList*)gROOT->GetListOfCanvases();

   // We use the interpreter to show the dialog canvas. This automatically 
   // makes sure that the canvas is entered in the event loop. 
   if (!lc->FindObject("SpectSensorDialog")) {
     char cmd[] = 
       "SpectSensorDialog = new TPhSpectSensorDialog(\"SpectSensorDialog\","\
       " \"Spectrometer Sensor Dialog\", 310, 190)";
     gInterpreter->ProcessLine(cmd);
   }
  */  
  // warn the user that this is only play...
  Warning("DisplayDialog", "The DisplayDialog class is obsolete. \n");
}
