#include "TMath.h"
#include "TPhStringCalTB98.h"
#include "TPhAbsRawHitFEC.h"

ClassImp(TPhStringCalTB98)

/////////////////////////////////////////////////////////////////////////////
//                                                                         //
// Methods to calibrate ADC signals in each separate string.               //
// Test Beam 98 specific calibration                                       //
//                                                                         //
// $Id: TPhStringCalTB98.cxx,v 1.4 1999/03/12 16:01:01 olszewsk Exp $
//                                                                         //
/////////////////////////////////////////////////////////////////////////////

TPhStringCalTB98::TPhStringCalTB98() : TPhStringCalAbs() 
{
  // Default constructor
}  

TPhStringCalTB98::TPhStringCalTB98(Int_t parno, Int_t parpointsno, Float_t boardlinearity,
				     Float_t *pars, Float_t *calpoints, 
				     Float_t *parerrors) :
                   TPhStringCalAbs(parno, parpointsno, boardlinearity, pars, calpoints, 
				    parerrors) 
{
  // Normal constructor
  Int_t i;
  if(parno != 4) {
    Error("TPhStringCalTB98","Wrong number (%d) of calibration params\n", parno); 
    fParNo = 0;                    
    fParPointsNo = 0;
    fCalLinearity = 0;
    fCalPoints.Reset();
    for(i=0; i<kMaxLength; i++) {
      fPars[i].Reset();
      fParErrors[i].Reset(); 
    }
  }
}  

TPhStringCalTB98::~TPhStringCalTB98()
{
  // Default destructor
}  
  
Bool_t TPhStringCalTB98::Calibrate(FECRawHit_t *hitstring, Int_t strlength)
{
  // Calibrate (change) ADC values according to the given formula and 
  // current parameters
  Int_t index;
  Double_t a,b,c,delta,r,r1,r2;
  Char_t sign = 1;
  
  if(!hitstring) {
    if(fVerbose) 
      Warning("Calibrate","nothing to calibrate\n");
      return kTRUE;
  }

  for(index=0; index<strlength; index++) {

    // 990318 HP change ADC from int to float 
    Float_t &ADC = (Float_t)hitstring[index].ADC;    
    Int_t channel = hitstring[index].ADC;

    // Skip the buffers "empty" slots
    if( !(ADC || channel) ) continue;

    // Convert deformed negative ADC signals back
    // (largest possible negative value cut set to kADCMaxNegative=512)
    // (this means everything above 4095-512=3583 is regarded negative)
    const Int_t negativePed = kADCMax - kADCMaxNegative;
    if(ADC > negativePed) ADC -= kADCMax;

    // Convert negative entries to positive for calibration
    // and assign previous sign before leaving
    sign = TMath::Sign((Float_t)1,ADC);
    ADC *= sign;

    Float_t par0 = fPars[index][0];
    Float_t par1 = fPars[index][1];
    Float_t par2 = fPars[index][2];
    Float_t par3 = fPars[index][3];
    if( ADC >= (Int_t) (par0 + par1*par3 + par2*par3*par3)) {
      ADC = (Int_t) fPars[index][3];
      continue;
    }
    
    // find solution for the 2nd order poly
    a = fPars[index][2]; 
    b = fPars[index][1]; 
    c = fPars[index][0] - ADC;
    delta = b*b - 4.0*a*c;
    if(delta<0) {
      ADC = 0;
      if(fVerbose) 
	Warning("Calibrate","bad parameters, Delta<0, ADC set to 0.\n");
      return kFALSE;
    }
    r1 = (-b + TMath::Sqrt(delta))/(2*a);
    r2 = (-b - TMath::Sqrt(delta))/(2*a);
    if(r1>=0 && r1<4100) r = r1;
    else 
      if(r2>=0 && r2<4100) r = r2;
      else {
	ADC = 0;
	if(fVerbose) 
	  Warning("Calibrate","signal out of range, set to 0.\n");
	return kFALSE;
      }
    // 990318 HP change ADC to float 
    ADC = (Float_t) r;
    // 990318 HP change to precision of 100 1ADC=0.01fC
    // transform DAC setting to charge (fC x 100 - to keep accuracy)
    // 2.0pF test capacitor, 21.0 voltage divider on input of preamplifier
    // fCalLinearity - calibration board linearity, 100 - scaling factor
    // 990318 HP convert ADC back to int_t by proper rounding
    // 990319 HP add in the measured fCalLinearity for the cal board used 
    // in the testbeam
    fCalLinearity = 0.9707;
    ADC = TMath::Nint( (ADC * 2.0 * fCalLinearity * 100.0 / 21.0) ); 
    ADC *= sign;
  }  

  return kTRUE;
}




