#ifndef _TPhMul_INCLUDED_
#define _TPhMul_INCLUDED_
////////////////////////////////////////////////////////////////////////
// TPhMul is a multiplicity reconstruction class that produces 
// arrays of multiplicity and occupancy in physics coordinates 
// reconstructed from charge deposited in sensor pads.
//
// File: TPhMul.h
//
// Date     Vers  Description
// 15-Jan-99 4.0  Change "MCHits" -> "PMCHits", "MCTracks" -> "PMCTracks",
//                "HitArraysFromHits" -> "PMCHitArraysFromHits" following
//                unannounced changes in Phat; make names public (RV)
//
// $Id: TPhMul.h,v 1.10 1999/01/18 19:51:07 verdier Exp $
//
////////////////////////////////////////////////////////////////////////

// 02-Jan-99 4.0  Update GetGraph and GetHist calls (RV)
// 16-Dec-98 4.0  Remove unused arguments to avoid gcc warnings (RV)
// 29-Nov-98 3.0  Add fDet to provide access to detector object;
//                add fHitArrayName, fHitArrayId, fDataFormat (RV)
// 03-Nov-98 2.0  Add Release method and invoke it in destructor (RV)
// 28-Aug-98 1.0  Make sensor data error message limits and HitArray 
//                names public (RV)
// 17-Aug-98 1.0  UseHitArrays -> fDataSource = EData; add SetVertex;
//                increase kNDispMax to allow 400 displays (RV)
// 03-Aug-98 1.0  Add fNTracksTrace for diagnostic tracing (RV)
// 14-Jul-98 1.0  Add GetGraph and GetHist and remove Display methods (RV)
// 03-Jul-98 1.0  Add fNTracksUse for diagnostic tracing (RV)
// 23-Jun-98 1.0  Access GetData diagnostics through public variables (RV)
// 17-Jun-98 1.0  Add GetDisplayIndex and GetSensMapIndex (RV)
// 01-Jun-98 1.0  Implement MaxEvents in PadMaps (RV)
// 15-May-98 0.3  Change enum's to const ints for var names in CInt (RV)
// 12-May-98 0.3  Remove det argument from Process (RV)
// 09-May-98 0.3  Add GetSensMap(name) (RV)
// 06-May-98 0.3  Add GetDisplay(name), ListDisplays, RL's (RV)
// 02-Mar-98 0.3  Add gFitDept for access to user fit functions (RV)
// 23-Feb-98 0.3  Put enumerations outside class scope (RV)
// 15-Feb-98 0.1  Add fDispDefs for access to display definitions (RV)
// 29-Jan-98 0.1  Add GetDeptP and GetFsecP methods for access to 
//                Dept and Fsec classes for fitting (RV)
// 18-Jan-98 0.1  Add GetSensMap method for access to sensor data (RV)
// 15-Jan-98 0.1  Add diagnostic pointer access methods for histos (RV)
// 12-Jan-98 0.1  Add access to SensDefs; generalize Process mode arg (RV)
// 30-Dec-97 0.1  Add primitive Set method (RV)
// 25-Nov-97 0.1  Add MC study variables (RV)
// 08-Nov-97 0.1  See GetDisplay (RV)
// 04-Nov-97 0.1  Creation (Robin Verdier)

// Root Classes

// PHOBOS Classes

// forward declarations

class TGraphErrors;        
class TH1F;        
class TH2F;        
class TObject;        
class TPhDept;
class TPhDetector;        
class TPhDispDef;        
class TPhEvent;        
class TPhFsec;
class TPhOpts;     
class TPhPadMap;     
class TPhRun;     
class TPhSensDef;
class TPhVertex;

// Constants: these could be in a separate header file

const Int_t kBarrel = 1, kRing = 2; // Sensor types

const Int_t   kNDispMax = 400; // Storage limit on # displays
const Int_t   kNSensMax = 164; // Storage limit on # sensors

const Int_t   kNDiagMax = 10;     // Storage limit on # of diagnostic objects

// Return values
enum { EOK = 0, EbadArg = -999991, Eunimplemented = -999992 };

// Values for fDataSource:

const Int_t
  EData = 0,
  EMCData = 1,
  EMCHits = 2,
  EDiagnostic = 3;

// Indices and maximum number of variables in the sensor pad maps, 
// used by GetData and Recon methods:

const Int_t
  ESEdep          = 0, // Energy deposited in pad, keV
  // The following are only for Monte Carlo data
  ESNTracksMC     = 1, // # tracks hitting pad
  ESNTracksPrimMC = 2, // # primary tracks hitting pad
  ESEdepPrimMC    = 3, // Energy deposited by primary tracks
  ESNVmax         = 4;

// Indices and maximum number of variables that can be mapped to displays,
// used by the TPhMul::Recon and possibly Display methods

// Accumulating variables, summed over an Eta-phi display cell:
const Int_t
  dEtadPhi        =  0, // Eta-Phi area of pads entered
  nPads           =  1, // Number of pads (fractional)
  Edep            =  2, // Energy deposited, limited & at normal incidence
  E1              =  3, // Global energy / track used to calculate nTracks
  nTracks         =  4, // Reconstructed # tracks incident
  nTracksLim      =  5, // Reconstructed # tracks for limited energy deposition
  FSec            =  6, // Global total / primary tracks used for nTracksPrim
  nTracksPrim     =  7, // Reconstructed # primary tracks incident

// Monte Carlo variables, used when DataSource != EData:
  nTracksMC       =  8, // MC # generated tracks incident on this pad
  nTracksPrimMC   =  9, // MC # generated primary tracks incident on this pad
  EdepPrimMC      = 10, // MC energy deposited by primary tracks

  EDNVAcc         = 11, // Total number of accumulating display variables
// At present only the first EDNVAcc variables are incremented; this will
// later be handled by flags in the display maps.

// Normalized ratios:
  Occ             = 11, // Reconstructed occupancy, tracks / pad
  OccRMS          = 12, // RMS uncertainty in occupancy
  d2NdEtadPhi     = 13, // Double differential cross section
  d2NdEtadPhiRMS  = 14, // RMS uncertainty in cross section

// Monte Carlo averaged variables:
  EdepPTMC        = 15, // MC energy deposited per track, limited & normalized
  EdepPTRMSMC     = 16, // RMS uncertainty in EdepPTMC
  EdepPPTMC       = 17, // MC energy deposited per primary track, lim & norm
  EdepPPTRMSMC    = 18, // RMS uncertainty in EdepPTMC
  FSecMC          = 19, // MC total / primary tracks, dEta*dPhi weighted  
  FSecRMSMC       = 20, // RMS uncertainty in FSecMC
  OccMC           = 21, // MC tracks / pad
  OccRMSMC        = 22, // RMS uncertainty in OccMC
  d2NdEtadPhiMC   = 23, // MC double differential cross section
  d2NdEtadPhiRMSMC = 24, // RMS uncertainty in d2NdEtadPhiMC

// Diagnostics:
  EdepPPad        = 25, // Energy deposited per pad
  E1Avg           = 26, // Average global mip weighted by dEta*dPhi
  FSecAvg         = 27, // Average global total / primary track wtd by dEdP

  EDNVmax         = 28; // Total number of display variables

class TPhMul : public TObject 

{

public:

// Public fields for developers:

  Int_t fDataSource;     // EData=0, EMCData=1, EMCHits=2, EDiagnostic=3
  Int_t fDataFormat;     // 0 for .root, 1 for .apx
  TPhDetector *fDet;     //! Pointer to detector class, NO STREAMER
  Int_t fGDNTracksTrace; // GetData: # tracks to trace in MCHits mode
  Int_t fGDTrackTrace;   // GetData: # of first track to trace
  Int_t fGDTraceSkip;    // GetData: # of tracks between traces
  Int_t fGDNDiagBins[kNDiagMax]; // GetData: #s of diag histo bins, skip if 0
  Int_t fHitArrayId;     // Index into fHitArrayName
  Char_t *fHitArrayName[3]; //! "HitArraysFromHits", "MCHitArrays", Unused
  Int_t fMaxBadIndexMsgs; // index discrepancy message limit
  Int_t fMaxInactSensMsgs; // Inactive sensor message limit
  Int_t fMaxUnkSensMsgs; // unknown sensor message limit
  Char_t *fMCHitsName;   //! Name of MCHits container in event
  Char_t *fMCTracksName; //! Name of MCTracks container in event
  Int_t fNTracksUse;     // # tracks to use, for diagnostics, NEVER for data
  Int_t fXIndexErrTol;   // Maximum allowed Z-type pad index discrepancy (MC)
  Int_t fYIndexErrTol;   // Maximum allowed Phi-type pad index discrepancy (MC)
  Int_t RLiprint;        // # of printouts since reset
  Int_t RLNPrint;        // Maximum printouts {20}
  Int_t RLGlobal;        // Report levels for various methods:
  Int_t RLProcess;
  Int_t RLGetData;
  Int_t RLGetF;
  Int_t RLRecon1;
  Int_t RLRecon2;

// The following should not have to be public; we need a place to 
// hold pointers to them. Note if we declare too few, we'll get a 
// crash without compile warnings

  TPhPadMap *Maps[kNSensMax];
  TPhPadMap *Disps[kNDispMax];

// The constructor should pass the detector object as an argument, but this 
// would break old scripts.  Instead, you can pass it via mp.fDet = xxx.
// Process will set fDet automatically if it finds it to be 0.
//  TPhMul(TPhDetector *det = (TPhDetector *)0);

  TPhMul();

  virtual ~TPhMul ();

  void GetData(Int_t isens, TPhEvent *event, TPhPadMap **Maps, 
               TPhDept &de, TPhFsec &fs);

  TPhDept *GetDeptP () { return fDept; } // Return pointer to TPhDept

  void* GetDiag(Int_t id) { return fDiagnostics[id-1]; } // Return diagnostic
    // pointer # id = 1, 2, ...

  Char_t *GetDiagName(Int_t id) { return fDiagNames[id-1]; } // Return name 
    // of diagnostic # id = 1, 2, ...

  TPhDispDef *GetDispDefs() { return fDispDefs; } // Get display definitions

  TPhPadMap *GetDisplay(Int_t idisp); // Get a display map by index

  TPhPadMap *GetDisplay (Char_t *name); // Get a display map by name

  Int_t GetDisplayIndex (Char_t *name); // Get a display map's index

  TPhPadMap **GetDisplays () { return Disps; } // Get array of display maps

  Float_t GetF(Char_t *name, Int_t arrIndex);

  Float_t GetF(/*Int_t nameIndex, Int_t arrIndex*/)
  { return Eunimplemented; }

  TPhFsec *GetFsecP () { return fFsec; } // Return pointer to TPhFsec

  TGraphErrors *GetGraph
  ( Char_t *typeName,    // Name of TGraphErrors for gROOT->GetObject access;
                         // 1st character is "s" or "S" for sensor map,
                         // anything else for display map
    Char_t *selector,    // Selector string such as "MPA3" or "MPA*"
    Int_t iVar = 0,      // Index of histo'd variable, defined in TPhMul.h
    Int_t iVarErr = -1,  // Index of error variable, -1 if none
    Float_t norm = 1.0,  // Normalizer, e.g. 360 to integrate on phi
    Int_t nXBins = 0,    // Number of X-bins in histo
    Int_t nYBins = 0,    // Number of Y-bins in histo
    Int_t nXMinSkip = 0, // Number of low X-bins to omit
    Int_t nYMinSkip = 0, // Number of low Y-bins to omit
    Int_t nXMaxSkip = 0, // Number of high X-bins to omit
    Int_t nYMaxSkip = 0  // Number of high Y-bins to omit
  );

  TGraphErrors *GetGraph
  ( Char_t *typeName,    // Name of TGraphErrors for gROOT->GetObject access;
                         // 1st character is "s" or "S" for sensor map,
                         // anything else for display map
    Int_t iMap = -1,     // Index of map from ListDisplays() or ListSensors()
    Int_t iVar = 0,      // Index of histo'd variable, defined in TPhMul.h
    Int_t iVarErr = -1,  // Index of error variable, -1 if none
    Float_t norm = 1.0,  // Normalizer, e.g. 360 to integrate on phi
    Int_t nXBins = 0,    // Number of X-bins in histo
    Int_t nYBins = 0,    // Number of Y-bins in histo
    Int_t nXMinSkip = 0, // Number of low X-bins to omit
    Int_t nYMinSkip = 0, // Number of low Y-bins to omit
    Int_t nXMaxSkip = 0, // Number of high X-bins to omit
    Int_t nYMaxSkip = 0  // Number of high Y-bins to omit
  );

  TH2F *GetHist
  ( Char_t *typeName,    // Name of TH2F for access via gROOT->GetObject;
                         // 1st character is "s" or "S" for sensor map,
                         // anything else for display map
    Char_t *selector,    // Selector string such as "MPA3" or "MPA*"
    Int_t iVar = 0,      // Index of histo'd variable, defined in TPhMul.h
    Int_t iVarErr = -1,  // Index of error variable, presently unused
    Float_t norm = 1.0,  // Normalizer, e.g. 360 to integrate on phi
    Int_t nXBins = 0,    // Number of X-bins in histo
    Int_t nYBins = 0,    // Number of Y-bins in histo
    Int_t nXMinSkip = 0, // Number of low X-bins to omit
    Int_t nYMinSkip = 0, // Number of low Y-bins to omit
    Int_t nXMaxSkip = 0, // Number of high X-bins to omit
    Int_t nYMaxSkip = 0  // Number of high Y-bins to omit
  );

  TH2F *GetHist
  ( Char_t *typeName,    // Name of TH2F for access via gROOT->GetObject;
                         // 1st character is "s" or "S" for sensor map,
                         // anything else for display map
    Int_t iMap = -1,     // Index of map from ListDisplays() or ListSensors()
    Int_t iVar = 0,      // Index of histo'd variable, defined in TPhMul.h
    Int_t iVarErr = -1,  // Index of error variable, presently unused
    Float_t norm = 1.0,  // Normalizer, e.g. 360 to integrate on phi
    Int_t nXBins = 0,    // Number of X-bins in histo
    Int_t nYBins = 0,    // Number of Y-bins in histo
    Int_t nXMinSkip = 0, // Number of low X-bins to omit
    Int_t nYMinSkip = 0, // Number of low Y-bins to omit
    Int_t nXMaxSkip = 0, // Number of high X-bins to omit
    Int_t nYMaxSkip = 0  // Number of high Y-bins to omit
  );

  Int_t GetI(/*Char_t *name, Int_t arrIndex*/) { return Eunimplemented; }

  // Int_t GetI(Int_t nameIndex, Int_t arrIndex) { return Eunimplemented; }

  Int_t GetNDiag() { return fNumberDiags; } // Number of diagnostic entries

  TPhSensDef *GetSensDefs () { return fSensDefs; } // Get sensor definitions

  TPhPadMap *GetSensMap (Int_t isens); // Get a sensor map by index

  TPhPadMap *GetSensMap (Char_t *name); // Get a sensor map by name

  Int_t GetSensMapIndex (Char_t *name); // Get a sensor map's index

  TPhPadMap **GetSensMaps () { return Maps; } // Get array of sensor maps

  TPhVertex *GetVertex (void);

  void ListDisplays (Int_t idMin = 0, Int_t idMax = kNDispMax);

  void ListSensors (Int_t idMin = 0, Int_t idMax = kNSensMax);

  void Process (TPhEvent *e, Int_t mode = 0);

  void Recon (Int_t isens, TPhPadMap **Maps, TPhDept &de, TPhFsec &fs,
              TPhPadMap **Disps);

  void Recon (TPhPadMap **Disps);

  void Release (); // Release heap storage

  Int_t SetDiag(Int_t id, Char_t *name, void* ptr); // Name and store a pntr

  Int_t SetF(Char_t *name, Int_t arrIndex, Float_t value);

  Int_t SetF(/*Int_t nameIndex, Int_t arrIndex, Float_t value*/)
  { return Eunimplemented; }

  Int_t SetI(/*Char_t *name, Int_t arrIndex, Int_t value*/)
  { return Eunimplemented; }

  /* Int_t SetI(Int_t nameIndex, Int_t arrIndex, Int_t value)
   { return Eunimplemented; }
  */
  void  SetOpts (TPhOpts *op);

  void  SetVertex (TPhVertex *v);

  void  ShowDiag(Int_t id); // Show stored objects on stdout

private:

  TPhDept     *fDept;     //! Pointer to Dept class, NO STREAMER

  TPhDispDef  *fDispDefs; //! Display definition class, NO STREAMER

 static
  Float_t      feMax;     // Deposited energy limit for one pad 
                          // This should depend on the average occupancy 
                          // in each part of the detector.

  Int_t        fEvent;    // Current event number ?

  TPhFsec     *fFsec;     //! Pointer to Fsec class, NO STREAMER

  TPhOpts     *fOpts;     //! Pointer to processing options class, NO STREAMER

  TPhSensDef  *fSensDefs; //! Sensor definitions object, NO STREAMER

  Int_t        fRunNo;    // Current run # ?

// Storage for diagnostic histograms accessed by GetDiag, ...
// This assumes fNumberDiags is created with value 0. ****
  Int_t        fNumberDiags /*= 0*/; // Current number of stored entries

  void*        fDiagnostics[kNDiagMax]; //! histo storage, NO STREAMER

  Char_t      *fDiagNames[kNDiagMax];   //! Diagnostic name pointers, NO STR

  TPhVertex   *fVertex;   //! NO STREAMER

public:

ClassDef(TPhMul,3)        // Phobos multiplicity processing class

};

#endif
