#ifndef _TPhMulDisp_INCLUDED_
#define _TPhMulDisp_INCLUDED_
////////////////////////////////////////////////////////////////////////
// TPhMulDisp is a multiplicity display class that displays raw data and 
// reconstructed physics data produced by TPhMul.
// The user instantiates a TPhMulDisp object for each plot, then calls 
// its Draw(...) methods to produce a plot.
//
// File: TPhMulDisp.h
//
// Date     Vers  Description
// 29-Jan-99 2.0  Add new constructors and Draw methods (RV)
//
// $id$
//
////////////////////////////////////////////////////////////////////////

// 02-Jan-99 1.0  Update GetGraph and GetHist calls (RV)
// 13-Nov-98 1.0  Create labels on heap (RV)
// 09-Oct-98 1.0  Complete revision (RV)
// 03-Dec-97 0.1  Creation (Robin Verdier)

// Root Classes

  class TCanvas;
  class TMarker;
  class TPaveLabel;

// PHOBOS Classes

// forward declarations

//  class TPhOpts;
  class TPhMapSet;
  class TPhMul;
  class TPhPadMap;
//  struct TPhPlotDef;
//  struct TPhPlotStyle;

  class TObject;        

  enum { EMaxPlots = 5 }; // Storage limits for Keys and Legends

/* The following struct was replaced by the MulDisp class:

// Plot definition struct defines a single display, possibly with 
// several variables and their errors:

struct TPhPlotDef

{ Char_t       *Name;      // Display variable name
  Int_t         Ivar;      // Variable index in the PadMap
  TPhPlotStyle *Style;     // Pointer to a plot style struct
  Int_t         LowXSkip;  // # bins to skip at the low end of the "X"-coord
...

  // Initialization defaults:
  // Note that unnamed variables are apparently NOT initialized to zero.

  TPhPlotDef() : Name(""), IvAR(13) {}

  TPhPlotDef(Char_t *name, Int_t iv) : Name(name), Ivar(iv), Irms(1) {}

};


// Plot style struct defines visual aspects of a display.

struct TPhPlotStyle

{
  public:

  enum Color {Black = 1, Red, Green, Blue, Yellow, Magenta, Cyan, 
              Green2, Blue2, White, Gray, Black2 };
 
  enum Style {Circle = 20, Square, TriangleUp, TriangleDown, 
              CircleOutline, SquareOutline, TriangleUpOutline, 
              TriangleDownOutline, CrossOutline, XOutline, StarOutline,
              Point };

// The following don't seem to be known; FIND THEM.

  typedef int Color_t;
  typedef int Width_t;
 
  Int_t   Ndim;        // Number of dimensions of the plot
  Int_t   Flags;       // Action flags: if it contains
                       //  1, plot graph with errors
                       //  2, make lego plot;
                       //  4, make for scatter plot
  Float_t KeyPos;      // Position of lower left corner of key in [0,1.]
  Float_t KeySize;     // Height of key pave [0,1.]
  Color_t LineColor;   // ROOT line color index
  Width_t LineWidth;   // ROOT line size, 0 means default
  Int_t   LineStyle;   // ROOT line style, 0 means default, -1 if none
  Color_t MarkerColor; // ROOT marker color index
  Width_t MarkerSize;  // ROOT marker size, 0 means default
  Int_t   MarkerStyle; // ROOT marker style index, 0 for no marker
  Int_t   Special;     // Additional user-defined flags
  Char_t *Title;       // Title, none if null
  Char_t *XLabel;      // Coordinate label string; use default if null
  Char_t *YLabel;      // Coordinate label string; use default if null

  // Additional items like Text pointers and sizes may be added.

  // Initialization defaults:

  TPhPlotStyle() : Ndim(1), Flags(1), KeyPos(0.9), KeySize(0.09),
                   LineColor(Black), LineWidth(0), LineStyle(0),
                   MarkerColor(Red), MarkerSize(0), MarkerStyle(Circle),
                   Special(0), Title(""), XLabel(""), YLabel("") {}; 

  TPhPlotStyle (Int_t nd, Int_t fl, Float_t kpos, Float_t ksize, Color_t lcol,
                Width_t lwid, Int_t lstyle, Color_t mcol, Width_t msize, 
                Int_t mstyle, Int_t spec, Char_t *t, Char_t *xlab,
                Char_t *ylab) :
                Ndim(nd), Flags(fl), KeyPos(kpos), KeySize(ksize),
                LineColor(lcol), LineWidth(lwid), LineStyle(lstyle), 
                MarkerColor(mcol), MarkerSize(msize), MarkerStyle(mstyle),
                Special(spec), Title(t), XLabel(xlab), YLabel(ylab) {}; 

};

*/

class TPhMulDisp : public TObject

{

public:

  TPhMulDisp();

  TPhMulDisp
  ( TPhMapSet *ms,        // TPhMapSet containing TPhPadMaps to be plotted
    Char_t *sel,          // Selector string for display, e.g. "*+in"
    Int_t   iMap,         // If >= 0, override selector and use display or 
                          // sensor TPhPadMap with index iMap as shown by 
                          // mp.ListDisplays() or mp.ListSensors()
    Int_t   var = -1,     // TPhPadMap variable index to plot, 0, 1, ...
                          // Indices and numbers of variables in the sensor and
                          // display padmaps are enumerated in TPhMul.h.
    Int_t   varRMS = -1,  // TGraphErrors err variable index if >= 0, else none
                          // For grHist = 1, this is varWt = -1 for 1.0
    Float_t scale = 1.,   // Total coverage <= 360 if integrating on phi
    Int_t   grHist = 0,   // 0 for TGraphErrors, 1 for TH2F
    Int_t   diSens = 0,   // 0 for display map, 1 for sensor map
    Char_t *xLabel = "Eta", // X axis text
    Char_t *yLabel = "",  // Y axis label, e.g. "dNch/dEta"
    Char_t *key = "",     // Key text, e.g. "Reconstructed"
    Int_t   mStyle = 21,  // Marker style, usually 20, 21, ...
    Int_t   mColor = 2,   // Marker color: 0-12 for none, black, red, green,
                          //   blue, yellow, magenta, cyan, green, blue,
                          //   white, gray, black.  Others may be possible.
    Float_t mSize = 0.8   // Marker size, usually 0.5 - 1.0
  );

  TPhMulDisp 
  ( TPhPadMap *pm,   // TPhPadMap containing data to be plotted
                     // Other variables are as in the first constructor:
    Int_t var = -1, Int_t varRMS = -1, Float_t scale = 1., Int_t grHist = 0,
    Int_t diSens = 0, Char_t *xLabel = "Eta", Char_t *yLabel = "",
    Char_t *key = "", Int_t mStyle = 21, Int_t mColor = 2, 
    Float_t mSize = 0.8
  );

  TPhMulDisp 
  ( TPhMul  &mp,      // TPhMul instance containing TPhPadMaps to be plotted
    Char_t  *sel,     // Selector string for display, e.g. "*+in"
    Int_t    iMap,    // If >= 0, override selector and use display or 
                      // sensor TPhPadMap with index iMap as shown by 
                      // mp.ListDisplays() or mp.ListSensors()
                      // Other variables are as in the first constructor:
    Int_t var = -1, Int_t varRMS = -1, Float_t scale = 1., Int_t grHist = 0,
    Int_t diSens = 0, Char_t *xLabel = "Eta", Char_t *yLabel = "",
    Char_t *key = "", Int_t mStyle = 21, Int_t mColor = 2, 
    Float_t mSize = 0.8
  );

  virtual ~TPhMulDisp ();

  void Delete ();

// Use varRMS = -99 for "unspecified" because -1 normally means "none"
  void Draw (Int_t var = -1, Int_t varRMS = -99, Float_t scale = -1.,
    Char_t *key = 0);

// Obsolete form for [some] backwards compatibility
  void Draw (Int_t seq, Text_t canvas);

  void SetDefaults
  ( Int_t   var = -1,     // TPhPadMap variable index to plot, 0, 1, ...
                          // Indices and numbers of variables in the sensor and
                          // display padmaps are enumerated in TPhMul.h.
    Int_t   varRMS = -1,  // TGraphErrors err variable index if >= 0, else none
                          // For grHist = 1, this is varWt = -1 for 1.0
    Float_t scale = 1.,   // Scale factor, = 360 if integrating on phi
    Int_t   grHist = 0,   // 0 for TGraphErrors, 1 for TH2F
    Int_t   diSens = 0,   // 0 for display map, 1 for sensor map
    Char_t *xLabel = "",  // X axis label, e.g. "Eta"
    Char_t *yLabel = "",  // Y axis label, e.g. "dNch/dEta"
    Char_t *key = "",     // Key text, e.g. "Reconstructed"
    Int_t   mStyle = 21,  // Marker style, usually 20, 21, ...
    Int_t   mColor = 2,   // Marker color: 0-12 for none, black, red, green,
                          //   blue, yellow, magenta, cyan, green, blue,
                          //   white, gray, black.  Others may be possible.
    Float_t mSize = 0.8   // Marker size, usually 0.5 - 1.0
  );

  void Update ();

// The fields are declared as public to allow easily-expandable user control

public:

  TCanvas *fCanvas;   //! Pointer to multiplicity canvas, NO STREAMER yet
  Char_t  *fCanvasName; //! Canvas name
  Int_t    fCanXMin;  // Canvas extent in absolute pixels
  Int_t    fCanYMin;
  Int_t    fCanXMax;
  Int_t    fCanYMax;
  Int_t    fDiSens;   // 0 for display map, 1 for sensor map
  Char_t  *fDrawOpts; //! Drawing options, e.g. "ape"  NO STREAMER
  Float_t  fExpand[2]; // Horizontal, vertical size relative to data extent
  Int_t    fIMap;     // If >= 0, override selector and use as map index
  Int_t    fGrHist;   // 0 for TGraphErrors, 1 for TH2F
  Char_t  *fKey;      //! Key text, e.g. "Reconstructed"  NO STREAMER
  Int_t    fKeyFont;  // Font number for key text
  Float_t  fKeyHSize; // Horizontal size of key in frame, 0 - 1.0
  Float_t  fKeyVSep;  // Vertical separation of keys in frame, 0 - 1.0
  Float_t  fKeyVSize; // Vertical size of key in frame, 0 - 1.0
  Float_t  fKeyH;     // Horizontal position of key in frame, 0 - 1.0
  Float_t  fKeyV;     // Vertical position of key, 0 - 1.0
  Char_t  *fLegend;   //! Key legend
  TPhPadMap *fPadMap; // Pointer to TPhPadMap, if non-zero
  TPhMapSet *fMapSet; // Pointer to TPhMapSet, if non-zero
  Int_t    fMColor;   // Marker color: 0-12 for none, black, red, green,
                      //   blue, yellow, magenta, cyan, green, blue,
                      //   white, gray, black.  Others may be possible.
  Float_t  fMSize;    // Marker size, usually 0.5 - 1.0
  Int_t    fMStyle;   // Marker style, usually 20, 21, ...
  TPhMul   fMul;      //! TPhMul instance containing TPhPadMaps, if non-zero
  Int_t    fNoNew;    // If non-0, do not create graphs, paves, or markers
  Float_t  fPlotLim[2][2]; // Limits of the plot, in plot variable coordinates
  Int_t    fReportLevel; // Report level = 1 for only fatal errors, ...
  Float_t  fScale;    // For integrating over phi
  Char_t  *fSel;      //! Selector string for display, e.g. "*+in"
  Int_t    fSeq;      // Sequence 0, 1, ... in frame
  Int_t    fSource;   // Source data: 0=TPhMapSet, 1=TPhPadMap, 2=TPhMul
  TGraphErrors *fTG; //! Created TGraph object
  TH2F    *fH2;       //! Created TH2F histogram object
  Char_t  *fTGName;   //! Created ROOT object name
  Char_t  *fTitle;    //! title ...
  Float_t  fTitleH;
  Float_t  fTitleV;
  Int_t    fTitleFont;
  Int_t    fTitleColor;
  Float_t  fTitleSize;
  TMarker *fTM[EMaxPlots]; //! Key markers
  TPaveLabel *fTP[EMaxPlots]; //! Key paves with text
  TPaveLabel *fTPDa;  //! Key pave with date
  Int_t    fVar;      // TPhPadMap variable index to plot, 0, 1, ...
                      // Indices and numbers of variables in the sensor and
                      // display padmaps are enumerated in TPhMul.h.
  Int_t    fVarRMS;   // Error variable index if >= 0, else none
  Char_t  *fXLabel;   //! X axis text, e.g. "Eta"  NO STREAMER
  Float_t  fXLabelH;  // X axis text horizontal offset
  Float_t  fXLabelV;  // FX axis text vertical offset
  Int_t    fXLabelFont;
  Int_t    fXLabelColor;
  Float_t  fXLabelSize;
  Char_t  *fYLabel;   //! Y axis label, e.g. "dNch/dEta"  NO STREAMER
  Float_t  fYLabelH;
  Float_t  fYLabelV;
  Int_t    fYLabelFont;
  Int_t    fYLabelColor;
  Float_t  fYLabelSize;

public:

ClassDef(TPhMulDisp,2)      // Phobos multiplicity display class

};

#endif
