#include "TGraphErrors.h"
#include "TH2.h"
#include "TObject.h"
#include "TPhDispDef.h"
#include "TPhMapSet.h"
#include "TPhMul.h"
#include "TPhSensDef.h"
#include "TPhPadMap.h"
ClassImp(TPhMapSet)
////////////////////////////////////////////////////////////////////////////
//
// TPhMapSet manages an aggregate of TPhPadMaps.
//
// File: TPhMapSet.cxx
//
// Date     Vers  Description
// 09-Feb-99 1.0  Make destructor clear maps, redefine GetNMapsAct, add
//                SetFree (RV)
//
// $id$
//
////////////////////////////////////////////////////////////////////////////

// 05-Feb-99 1.0  Make print and report levels public; don't write null 
//                maps (RV)
// 24-Jan-99 1.0  Creation (Robin Verdier)

TPhMapSet::TPhMapSet
( Int_t  mapType     // 0 = physics displays, else sensor maps
  )
{ fMapType = mapType;

  fIPrint = 0;
  fNPrint = 50;
  fReportLevel = 1;

  fNStats = ENStats;
  for (Int_t nn = 0; nn < fNStats; nn++)
  { fStats[nn] = 0;
    fStatsEv[nn] = 0;
  }
  fVX = 0;
  fVY = 0;
  fVZ = 0;

  Int_t nVars;
  if (!fMapType)
  { // fDMD = new TPhDispDef(); // Does this require ()? Stack inst doesn't
    fDDD = new TPhDispDef();
    fDSD = (TPhSensDef *)0;
    fNMaps = kNDispMax;      // Defined in TPhMul.h
    nVars = EDNVmax;         // Defined in TPhMul.h
    fMaps = new TPhPadMap *[fNMaps];
    for (Int_t nn = 0; nn < fNMaps; nn++)
      fMaps[nn] = new TPhPadMap (nn, fDDD->GetName(nn), fDDD, nVars, 0);
  } 
  else
  { // fDMD = new TPhSensDef();
    fDDD = (TPhDispDef *)0;
    fDSD = new TPhSensDef();
    fNMaps = kNSensMax;      // Defined in TPhMul.h
    nVars = ESNVmax;         // Defined in TPhMul.h
    fMaps = new TPhPadMap *[fNMaps];
    for (Int_t nn = 0; nn < fNMaps; nn++)
      fMaps[nn] = new TPhPadMap (nn, fDSD->GetName(nn), fDSD, nVars, 0);
  } 
}

TPhMapSet::~TPhMapSet() { Clear(); }

Int_t TPhMapSet::Add (TPhPadMap *map)
// Add a map at the first unoccupied position; return iErr != 0 if failed.
{ Int_t iErr = (map)? 0 : 1;
  if (!iErr)
  { Int_t iMap;
    for (iMap = 0; iMap < fNMaps; iMap++) if (fMaps[iMap] == 0) break;
    if (iMap < fNMaps)
    { fMaps[iMap] = map;
      if (fReportLevel >= 3 && fIPrint < fNPrint)
      { printf ("TPhMapSet/Add: added map %d, %s.\n", iMap,
          fMaps[iMap]->GetName());
        fIPrint++;
      } 
    }
    else
    { iErr = 2;
      if (fReportLevel > 0 && fIPrint < fNPrint)
      { printf ("TPhMapSet/Add: all %d slots filled, unable to add map %s;"
          " try DeleteMap.\n", fNMaps, map->GetName());
        fIPrint++;
      }
    }
  }
  else
  if (fReportLevel > 0 && fIPrint < fNPrint)
  { printf ("TPhMapSet/Add: no map found at %p.\n", map);
    fIPrint++;
  }
  return iErr;
}

void TPhMapSet::Clear()
{ // if (fDMD) delete fDMD;
  if (fDDD) delete fDDD;
  if (fDSD) delete fDSD;
  Int_t nDel = Delete (0, fNMaps);
  if (nDel != fNMaps && fReportLevel >= 2 && fIPrint < fNPrint)
// Note this is a weak warning, as the undeleted maps may have been 
// freed to prevent deletion.
  { printf ("TPhMapSet/Delete deleted %d maps, %d requested.\n",
      nDel, fNMaps);
    fIPrint++;
  }
  if (fMaps) delete[] fMaps;
}

Int_t TPhMapSet::Delete (Int_t iFirst, Int_t nDelete)
{ // Take no action if anything is wrong with the range.
  Int_t nDo = nDelete;
  Int_t nDone = 0;
  if (iFirst < 0 || iFirst + nDo > fNMaps) nDo = 0;
  for (Int_t nn = iFirst; nn < iFirst + nDo; nn++)
    if (fMaps[nn] != 0)
    { delete fMaps[nn];
      fMaps[nn] = 0;
      nDone++;
    }
  return nDone;
}

Int_t TPhMapSet::Delete (Char_t *name, Int_t nDelete)
{ return Delete (GetMapIndex(name), nDelete);
}

TGraphErrors *TPhMapSet::GetGraph ( Char_t *rName, Char_t *selector,
  Int_t iVar, Int_t iVarErr, Float_t norm, Int_t nXBins, Int_t nYBins,
  Int_t nXMinSkip, Int_t nYMinSkip, Int_t nXMaxSkip, Int_t nYMaxSkip )
{ TPhPadMap *map = GetMap(selector);
  return (map)? map->GetGraph (fMaps, fNMaps, rName, selector, -1, iVar,
    iVarErr, norm, nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, 
    nYMaxSkip) : (TGraphErrors *)0;
}

TGraphErrors *TPhMapSet::GetGraph ( Char_t *rName, Int_t iMap,
  Int_t iVar, Int_t iVarErr, Float_t norm, Int_t nXBins, Int_t nYBins,
  Int_t nXMinSkip, Int_t nYMinSkip, Int_t nXMaxSkip, Int_t nYMaxSkip )
{ TPhPadMap *map = GetMap(iMap);
  return (map)? map->GetGraph (fMaps, fNMaps, rName, "", iMap, iVar,
    iVarErr, norm, nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, 
    nYMaxSkip) : (TGraphErrors *)0;
}

TH2F *TPhMapSet::GetHist ( Char_t *rName, Char_t *selector,
  Int_t iVar, Int_t iVarErr, Float_t norm, Int_t nXBins, Int_t nYBins,
  Int_t nXMinSkip, Int_t nYMinSkip, Int_t nXMaxSkip, Int_t nYMaxSkip )
{ TPhPadMap *map = GetMap(selector);
  return (map)? map->GetHist (fMaps, fNMaps, rName, selector, -1, iVar,
    iVarErr, norm, nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, 
    nYMaxSkip) : (TH2F *)0;
}

TH2F *TPhMapSet::GetHist ( Char_t *rName, Int_t iMap,
  Int_t iVar, Int_t iVarErr, Float_t norm, Int_t nXBins, Int_t nYBins,
  Int_t nXMinSkip, Int_t nYMinSkip, Int_t nXMaxSkip, Int_t nYMaxSkip )
{ TPhPadMap *map = GetMap(iMap);
  return (map)? map->GetHist (fMaps, fNMaps, rName, "", iMap, iVar,
    iVarErr, norm, nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, 
    nYMaxSkip) : (TH2F *)0;
}

TPhPadMap *TPhMapSet::GetMap (Char_t *name)
{ Int_t iMap = GetMapIndex (name);
  return (iMap >= 0)? fMaps[iMap] : (TPhPadMap *)0;
}

TPhPadMap *TPhMapSet::GetMap (Int_t iMap)
{ return (iMap >= 0 && iMap < fNMaps)? fMaps[iMap] : (TPhPadMap *)0;
}

Int_t TPhMapSet::GetMapIndex (Char_t *name)
{ Int_t iMap;
  for (iMap = 0; iMap < fNMaps; iMap++)
    if (fMaps[iMap] && !strcmp(name, fMaps[iMap]->GetName())) break;
  return (iMap < fNMaps)? iMap : -1;
}

Int_t TPhMapSet::GetNMapsAct ()
// Returns number of non-null TPhPadMaps with non-blank names.  Note this 
// may not be exactly what you want, if some maps are dynamically turned off.
{ Int_t iMap, nAct = 0;
  for (iMap = 0; iMap < fNMaps; iMap++)
    if (fMaps[iMap] != 0 && strcmp("", fMaps[iMap]->GetName())) nAct++;
  return nAct;
}

void TPhMapSet::List (Int_t iMapMin, Int_t iMapMax)
{ Int_t nPerLine = 6;
  Char_t *fmt = "%4d: %7s";
  if (fMapType)
  { nPerLine =  8;
    fmt = "%4d: %4s";
  }
  Int_t nn = 0;
  if (iMapMin < 0) iMapMin = 0;
  if (iMapMax >= fNMaps) iMapMax = fNMaps - 1;
  for (Int_t iMap = iMapMin; iMap <= iMapMax; iMap++)
  { TPhPadMap *map = fMaps[iMap];
    if (map)
    { printf (fmt, iMap, map->GetName());
      if (nn++ > 0 && nn % nPerLine == 0) printf ("\n");
    }
  }
  if (nn % nPerLine != 0) printf ("\n");
}

Int_t TPhMapSet::Replace (TPhPadMap *map)
// Replace a map by another of the same name; return iErr != 0 if 
// there is none.
// THIS DELETES THE ORIGINAL MAP, ASSUMING IT TO BE ON THE HEAP.
{ Int_t iErr = (map)? 0 : 1;
  if (!iErr)
  { Int_t iMap = GetMapIndex (map->GetName());
    if (iMap >= 0)
    { if (fReportLevel >= 3 && fIPrint < fNPrint)
      { printf ("TPhMapSet/Replace: deleting map %d, %s.\n", iMap,
          fMaps[iMap]->GetName());
        fIPrint++;
      } 
      delete fMaps[iMap];
      fMaps[iMap] = map;
    }
    else
    if (fReportLevel > 0 && fIPrint < fNPrint)
    { printf ("TPhMapSet/Replace: no map named %s; try AddMap.\n",
        map->GetName());
      fIPrint++;
    }
  }
  else
  if (fReportLevel > 0 && fIPrint < fNPrint)
  { printf ("TPhMapSet/Replace: no map found at %p.\n", map);
    fIPrint++;
  }
  return iErr;
}

Int_t TPhMapSet::Replace (Char_t *name, TPhPadMap *map)
// Replace a named map; return iErr != 0 if failed.
// THIS DELETES THE ORIGINAL MAP, ASSUMING IT TO BE ON THE HEAP.
{ Int_t iMap = GetMapIndex (name);
  Int_t iErr = (iMap >= 0)? 0 : 1;
  if (!iErr)
  { iErr = Replace (iMap, map);
  }
  else
  if (fReportLevel > 0 && fIPrint < fNPrint)
  { printf ("TPhMapSet/Replace: map %s not found.\n", name);
    fIPrint++;
  }
  return iErr;
}

Int_t TPhMapSet::Replace (Int_t iMap, TPhPadMap *map)
// Replace the map at index iMap; return iErr != 0 if failed.
// THIS DELETES THE ORIGINAL MAP IF IT EXISTS, ASSUMING IT TO BE ON THE HEAP.
{ Int_t iErr = (map)? 0 : 2;
  if (!iErr)
  { if (iMap >= 0 && iMap < fNMaps)
    { if (fMaps[iMap] != 0)
      { if (fReportLevel >= 3 && fIPrint < fNPrint)
        { printf ("TPhMapSet/Replace: replacing map %d, %s, by %s.\n",
            iMap, fMaps[iMap]->GetName(), map->GetName());
          fIPrint++;
        } 
        delete fMaps[iMap];
      }
      else
      { if (fReportLevel >= 3 && fIPrint < fNPrint)
        { printf ("TPhMapSet/Replace: replacing null map %d by %s.\n",
            iMap, map->GetName());
          fIPrint++;
        } 
      }
      fMaps[iMap] = map;
    }
    else
    { iErr = 1;
      if (fReportLevel > 0 && fIPrint < fNPrint)
      { printf ("TPhMapSet/Replace: invalid index %d, not in [0,%d].\n",
          iMap, fNMaps);
        fIPrint++;
      }
    }
  }
  else
  if (fReportLevel > 0 && fIPrint < fNPrint)
  { printf ("TPhMapSet/Replace: no map found at %p.\n", map);
    fIPrint++;
  }
  return iErr;
}

Int_t TPhMapSet::SetFree(Int_t iFirstMap, Int_t iLastMap)
// Free maps with indices from iFirstMap to iLastMap by zeroing their
// pointers so that they will not be deleted with the MapSet.  This is
// used when the pointers have been "borrowed" from another owner.
{ Int_t nFreed = 0;
  if (iFirstMap < 0 && iLastMap < 0 && fReportLevel && fIPrint < fNPrint)
  { printf ("MapSet/SetFree: Use: SetFree(iFirstMap, iLastMap) "
      "to PREVENT deletion of these maps.\n"); 
    fIPrint++;
  }
  if (iFirstMap < 0) iFirstMap = 0;
  if (iLastMap >= fNMaps) iLastMap = fNMaps-1;
  for (Int_t imap = iFirstMap; imap <= iLastMap; imap++)
  { fMaps[imap] = 0;
    nFreed++;
  }
  return nFreed;
}


void TPhMapSet::Streamer(TBuffer &R__b)
{
// Stream an object of class TPhMapSet.
// Hand-edited (RV & MPD)
// This will eventually use selectors to determine which maps to read
// and write.

  if (R__b.IsReading()) {
    Version_t R__v = R__b.ReadVersion(); if (R__v) { }
    TObject::Streamer(R__b);
    R__b >> fMapType;
    R__b >> fDDD;
    R__b >> fDSD;
    R__b >> fNMaps;
    //R__b.ReadArray(fMaps);
// This modification and the corresponding write are by Patrick Decowski
    fMaps = new TPhPadMap*[fNMaps];
    for (Int_t nn = 0; nn < fNMaps; nn++)
    { fMaps[nn] = new TPhPadMap();
      fMaps[nn]->Streamer(R__b);
    }
    R__b >> fNEvents;
    R__b >> fVX;
    R__b >> fVY;
    R__b >> fVZ;
    R__b >> fNStats;
    R__b.ReadStaticArray(fStatsEv);
    R__b.ReadStaticArray(fStats);
  } else {
    R__b.WriteVersion(TPhMapSet::IsA());
    TObject::Streamer(R__b);
    R__b << fMapType;
    R__b << fDDD;
    R__b << fDSD;
    R__b << fNMaps;
    //R__b.WriteArray(fMaps, __COUNTER__);
    for (Int_t nn = 0; nn < fNMaps; nn++)
// Don't write null maps.  This avoids a crash but may cause problems
// reading back.
      if (fMaps[nn]) fMaps[nn]->Streamer(R__b);
    R__b << fNEvents;
    R__b << fVX;
    R__b << fVY;
    R__b << fVZ;
    R__b << fNStats;
    R__b.WriteArray(fStatsEv, 20);
    R__b.WriteArray(fStats, 20);
  }
}
///////////////////// end ClassImp(TPhMapSet)
