#include <math.h>
#include <string.h>
#include "TGraphErrors.h"
#include "TH2.h"
#include "TObject.h"
#include "TPhDispDef.h"
#include "TPhMul.h"
#include "TPhPadMap.h"
#include "TPhSensDef.h"
#include "TPhSensGeo.h" // defines eNameSizEn
#include "TPhVertex.h"

ClassImp(TPhPadMap)

///////////////////////////////////////////////////////////////////////
//
// TPhPadMap is a data container class that stores arrays of infor-
// mation in cells associated either with physical sensor pads or
// bins in physics variables, and which has methods to map informa-
// tion from one representation to the other.  An arbitrary number
// of variables, such as raw ADC counts, reconstructed occupancy,
// and multiplicity, can be stored and mapped in parallel for effi-
// ciency.
//
// Date     Vers  Description
// 16-Feb-99 2.0  Correct loss of negative phi (RV)
// $Id: TPhPadMap.cxx,v 1.16 1999/02/17 02:02:15 verdier Exp $
//
///////////////////////////////////////////////////////////////////////

// 24-Jan-99 2.0  Add streamer method (RV)
// 22-Jan-99 2.0  Change SetNCells to delete and reallocate when 
//                changing # cells; add PrintStats method (RV)
// 18-Jan-99 2.0  Move is outside loop for OSF1 scoping (RV)
// 13-Jan-99 2.0  Remove GetHist.cxx defaults; add GetStats and 
//                GetIntegral methods; change SetNcells to SetNCells
//                for consistency (RV)
// 31-Dec-98 2.0  Add new GetGraph and GetHist methods (RV)
// 18-Dec-98 2.0  Minor changes to eliminate compiler warnings (RV)
//  NOTE PROBLEM WITH SETTING MAPS AT LINE 485
// 16-Nov-98 2.0  Add size of TPhSensGeo to allow streaming (RV)
// 30-Oct-98 2.0  Implement destructor (RV)
// 24-Sep-98 1.0  Correct Print() (RV)
// 27-Aug-98 1.0  Lower report level for GetHist (RV)
// 22-Aug-98 1.0  Weight combined histos by phase space (RV)
// 19-Aug-98 1.0  Correct vertex handling (RV)
// 11-Aug-98 1.0  Add SetVertex method (RV)
// 07-Aug-98 1.0  Change GetEtaPhi to use fast eta calculation (RV)
// 27-Jul-98 1.0  Correct Fill in GetHist (RV)
// 20-Jul-98 1.0  Use TArrays for TGraph temporaries for OSF1 (RV)
// 16-Jul-98 1.0  Generalized GetGraph and GetHist methods (RV)
// 09-Jul-98 1.0  Add Print method (RV)
// 08-Jul-98 1.0  Use ixy = 0, 1 for x, y in GetMin, GetNCells, GetSize (RV)
// 03-Jul-98 1.0  GetGraph and GetHist arguments changed (RV)
// 01-Jun-98 1.0  Add fNEvents, fMaxEvents, fSelect access (RV)
// 09-May-98 0.3  Add GetHist method (RV)
// 08-May-98 0.3  Add GetGraph method, GetNvars->GetNVars (RV)
// 12-Mar-98 0.3  Correct syntax errors for Abacus (RV)
// 23-Feb-98 0.3  Get enumerations from TPhMul.h (RV)
// 19-Feb-98 0.3  Removed method GetDispList; now done by TPhMul::Recon (RV)
// 09-Feb-98 0.3  Added method Select replacing Match (RV)
// 22-Jan-98 0.3  Added methods GetDispList, Match, ... (RV)
// 19-Jan-98 0.2  Added method GetVal, changed Loc to GetLoc (RV)
// 07-Nov-97 0.2  Added methods GetDLoc and SetData (RV)
// 03-Nov-97 0.2  Get Map method variables from vector, not PadMap (RV)
// 20-Oct-97 0.1  Added method GetEtaPhi (RV)
// 13-Oct-97 0.1  Creation (Robin Verdier)
//
///////////////////////////////////////////////////////////////////////

TPhVertex *fVertex;  //! Pointer to event vertex, for GetEtaPhi, NO STREAMER

TPhPadMap::TPhPadMap() {}

TPhPadMap::TPhPadMap(Int_t id, Char_t *name)
{ SetId(id);
  if (name)
    SetName(name);
  else
    SetName("Id=???"); // Use fprintf to get rep of Ident
}

TPhPadMap::TPhPadMap(Int_t id, Char_t *name, TPhDispDef* s, Int_t nvars,
    Float_t val)
{ s->Copy(s, id, &fS);
  if (name)
    SetName(name);
  else
    SetName("Id=???");
  fSSize = sizeof(fS);
  fNvars = nvars;
  SetNCells(fS.NXCells, fS.NYCells, val);
//  fSize = fNvars * fS.NXCells * fS.NYCells;
//  fData = new Float_t[fSize];
//  for (Int_t kk = 0; kk < fSize; fData[kk++] = val);
}

TPhPadMap::TPhPadMap(Int_t id, Char_t *name, TPhSensDef* s, Int_t nvars,
    Float_t val)
{ s->Copy(s, id, &fS); // id will eventually be removed
  if (name)
    SetName(name);
  else
    SetName("Id=???");
  fSSize = sizeof(fS);
  fMaxEvents = -1; // Default: accumulate all events
  fNEvents = 0;
  fNvars = nvars;
  fSelect[0] = 0; // Not yet implemented
  SetNCells(fS.NXCells, fS.NYCells, val);
//  fSize = fNvars * fS.NXCells * fS.NYCells;
//  fData = new Float_t[fSize];
//  for (Int_t kk = 0; kk < fSize; fData[kk++] = val);
}

TPhPadMap::TPhPadMap(Int_t id, Char_t *name, Int_t type,
    Int_t nxcells, Float_t xmin, Float_t xsize,
    Int_t nycells, Float_t ymin, Float_t ysize,
    Float_t xorig, Float_t yorig, Float_t zorig, 
    Float_t u11, Float_t u21, Float_t u31, 
    Float_t u12, Float_t u22, Float_t u32, 
    Float_t u13, Float_t u23, Float_t u33, 
    Float_t dist)
{ fS.Version = 0.0;
  fS.Id = id; 
  strncpy(fS.Name, name, EnameSize);
  fS.Type = type; 
  fS.NXCells = nxcells; fS.Xmin = xmin;      fS.Xsize = xsize;
  fS.NYCells = nycells; fS.Ymin = ymin;      fS.Ysize = ysize;
  fS.Orig[0] = xorig;   fS.Orig[1] = yorig;  fS.Orig[2] = zorig;
  fS.Axes[0][0] = u11;  fS.Axes[1][0] = u21; fS.Axes[2][0] = u31;
  fS.Axes[0][1] = u12;  fS.Axes[1][1] = u22; fS.Axes[2][1] = u32;
  fS.Axes[0][2] = u13;  fS.Axes[1][2] = u23; fS.Axes[2][2] = u33;
  fS.Dist = dist;
}

TPhPadMap::~TPhPadMap()
{
// Note that the destructor releases its heap storage but does not
// delete any plots or histograms created by methods like GetGraph.

  Release();
}

////////////////////////////////////////////////////////////////////////////

//  virtual void     TPhPadMap::Fill(TPhSensor *Sensor);
//  virtual void     TPhPadMap::Show(FILE *Dev);

// The get and set interfaces must all be protected against invalid 
// indices.  The stopgap below must be fixed:

// I don't know how to implement the following:
// F2_t::F2_t()
// { Float_t val[2]; }

////////////////////////////////////////////////////////////////////////////

Float_t TPhPadMap::GetDist() { return fS.Dist; }

////////////////////////////////////////////////////////////////////////////

Float_t *TPhPadMap::GetDLoc (Int_t iv, Int_t ix, Int_t iy, Int_t dl[])
{ dl[0] = fS.NXCells * fS.NYCells;
  dl[1] = fS.NYCells;
  dl[2] = 1;
  if (iv < fNvars && ix < fS.NXCells && iy < fS.NYCells)
    return &fData[(iv * fS.NXCells + ix) * fS.NYCells + iy];
  else
    return 0;
}

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::GetEtaPhi(Float_t ix, Float_t iy, Float_t etaPhi[3])
{
// GetEtaPhi calculates physics variables for pads of the sensor 
// described by this PadMap:
//  etaphi[0] = eta = - ln (tan (theta/2)), 
//  etaphi[1] = phi = the azimuthal angle, and
//  etaphi[2] = dtdr = (thickness of silicon detector) / (radius from 
//   vertex to the point X, Y)  
// at the lower edges of the pad with FLOATING indices ix, iy. 
// The lowest active point has indices 0.0, 0.0. 
//
// Limitations:
// We assume here that theta is a function only of the X-variable; 
// this will be generalized.
// Eventually etaphi should become a struct or a valarray.
// kBarrel is a static const.
// We might also want a version that takes X and Y, rather than 
// their indices.
// 23-Oct-97: Creation (Robin Verdier)

// The following should only be done once per event, so needs an 
// initialization flag.
  static Float_t Vertex[3] = {0., 0., 0.};
  if (fVertex) Vertex[2] = fVertex->GetZ();

  static Float_t cosTheta, dtdr, sinTheta, tanThOv2;
  static Double_t r, rho, z;

// The following calculation of eta should never overflow for 
// Phobos detectors since theta is never 0 or pi:
  
  if (fS.Type == kBarrel)
  { z = fS.Xmin + ix * fS.Xsize - Vertex[2];  
    rho = fS.Dist;
  }

  else

  { // Ring counters specify Size = dR/R at central Phi.
    z = fS.Dist - Vertex[2];
    rho = fS.Xmin * pow (1.0 + fS.Xsize, ix);
  }

  r = sqrt (rho * rho + z * z);
  sinTheta = rho / r;
  cosTheta = z / r;
// Normal incidence correction = sin(Theta) for barrel, |cos(Theta)| for ring
  dtdr = (fS.Type == kBarrel)? sinTheta : fabs(cosTheta);

  if (1)
    tanThOv2 = sinTheta / (1.0 + cosTheta);
  else
// For testing only
    tanThOv2 = tan (0.5 * atan2(rho, z));
    
  etaPhi[0] = -log(tanThOv2);
  etaPhi[1] = fS.Ymin + iy * fS.Ysize;
  etaPhi[2] = dtdr;
}

////////////////////////////////////////////////////////////////////////////

// General GetGraph returns a pointer to a TGraphErrors with data accumulated
// from one or more selected TPhPadMaps.
TGraphErrors *TPhPadMap::GetGraph
( TPhPadMap **maps, // Array of maps to be searched for selector
  Int_t nMaps,      // Number of maps to search; see GetHist for spec values
  Char_t *graphName,// Name of TGraphErrors for access via gROOT->GetObject
  Char_t *sel,      // Selector string such as "MPA3" or "MPA*"
  Int_t iMap,       // If >= 0, ignore selector and use maps[iMap]
  Int_t iVar,       // Index of plotted variable, defined in TPhMul.h
  Int_t iVarErr,    // Index of error variable, add 100 for linear combination
  Float_t norm,     // Normalizer, e.g. 360.0 to integrate on phi
// nXBins or nYBins should be 1; otherwise, we treat nYBins as 1.
  Int_t nXBins,     // Number of X-bins in histo
  Int_t nYBins,     // Number of Y-bins in histo
  Int_t nXMinSkip,  // Number of low X-bins to omit
  Int_t nYMinSkip,  // Number of low Y-bins to omit
  Int_t nXMaxSkip,  // Number of high X-bins to omit
  Int_t nYMaxSkip   // Number of high Y-bins to omit
)
{
// local variables:
  static Int_t iprint = 0;
  Int_t Nprint = 20;
  Int_t ReportLevel = 1;
//Default style parameters:
  Int_t mStyle  = 20;
  Int_t mColor  = 2;
  Float_t mSize = 0.8;

// Create two histos containing the plotted variable and error.
// We delete them before creating the TGraphErrors to avoid fragmenting
// the heap, but this may not work.
// We let the caller specify the normalization; typically should be 
// 360 for dN/dEta.

  Bool_t plotX = (nXBins != 1);
  Int_t nYBinsUse = (plotX)? 1 : nYBins;
// Temporary settings for weight, will later be an argument
  Int_t iVarWt = 0;
  Int_t iVarErrWt = 100;
  if (iVarErr < 0) iVarErrWt = -1;
  if (iVarErr >= 100) iVarErrWt = 0;
  TH2F *h1 = GetHist(maps, nMaps, "MulTemp1", sel, iMap, iVar, iVarWt, norm,
    nXBins, nYBinsUse, nXMinSkip, nYMinSkip, nXMaxSkip, nYMaxSkip); 
  TH2F *h2 = 0;
  if (iVarErr >= 0)
    h2 = GetHist(maps, nMaps, "MulTemp2", sel, iMap, iVarErr, iVarErrWt, norm,
    nXBins, nYBinsUse, nXMinSkip, nYMinSkip, nXMaxSkip, nYMaxSkip); 

  TGraphErrors *tg = 0;

// Create variable arrays as required by TGraph (this should be unnecessary)
  TAxis *ta = h1->GetXaxis();
  Int_t nCells = ta->GetNbins();
//  if (nCells > nCellsMax) nCells = nCellsMax;
  if (nCells > 0)
  { TArrayF x(nCells);
    x.Reset();
    TArrayF y = x;
    TArrayF ex = x;
    TArrayF ey = x;
//    Float_t y[nCellsMax];
    Float_t cellSize = (ta->GetXmax() - ta->GetXmin()) / nCells;
    for (int icell = 0; icell < nCells; icell++)
    { x[icell]  = ta->GetXmin() + (icell + 0.5) * cellSize;
      y[icell]  = h1->GetCellContent(icell+1,1);
      if (ReportLevel >= 4 && iprint < Nprint)
      { printf ("TPhPadMap/GetGraph: found %10.2e at %10.2e in %s.\n",
          y[icell], x[icell], h1->GetTitle());
        iprint++;
      }
      if (iVarErr >= 0)
        ey[icell] = h2->GetCellContent(icell+1,1);
    }

// Note: don't delete ta; this causes an "unreasonable size" abort
    if (iVarErr >= 0) h2->Delete();
    h1->Delete();


// Create the TGraphErrors

    tg = new TGraphErrors (nCells, x.GetArray(), y.GetArray(),
      ex.GetArray(), ey.GetArray());
    if (tg)
    { if (strcmp(graphName, "")) tg -> SetName(graphName);
      else tg -> SetName(strcat("g",sel));
      tg -> SetTitle(sel);

// Set default style parameters

      tg->SetMarkerColor(mColor);
      tg->SetMarkerSize(mSize);
      tg->SetMarkerStyle(mStyle);

// printf ("%s has x-title %s, y-title %s\n", h1->GetName(),
//  h1->GetXTitle(), h1->GetYTitle());

      TH1F *h3 = tg->GetHistogram();
      if (h3)
      { h3->SetXTitle(/*h1->GetXTitle*/"X");    
        h3->SetYTitle(/*h1->GetYTitle*/"Y");
      }
    }

    else

      if (ReportLevel >= 2 && iprint < Nprint)
      { printf("GetGraph could not create requested TGraphErrors\n");
        iprint++;
      }

  } // if (nCells

  return tg;
}

////////////////////////////////////////////////////////////////////////////

// Specialized GetGraph returns a pointer to a TGraphErrors with data from 
// the TPhPadMap whose GetGraph is called.
TGraphErrors *TPhPadMap::GetGraph
( Char_t *graphName,// Name of TGraphErrors for access via gROOT->GetObject
  Int_t iVar,       // Index of plotted variable, defined in TPhMul.h
  Int_t iVarErr,    // Index of error variable
  Float_t norm,     // Normalizer, e.g. 360.0 to integrate on phi
// nXBins or nYBins should be 1; otherwise, we treat nYBins as 1.
  Int_t nXBins,     // Number of X-bins in graph
  Int_t nYBins,     // Number of Y-bins in graph
  Int_t nXMinSkip,  // Number of low X-bins to omit
  Int_t nYMinSkip,  // Number of low Y-bins to omit
  Int_t nXMaxSkip,  // Number of high X-bins to omit
  Int_t nYMaxSkip   // Number of high Y-bins to omit
)
{ TPhPadMap *maps[] = { this };
  return GetGraph (maps, 1, graphName, "", 0, iVar, iVarErr, norm,
    nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, nYMaxSkip);
}

////////////////////////////////////////////////////////////////////////////

// Specialized GetGraph uses OBSOLETE call form for backwards compatibility
TGraphErrors *TPhPadMap::GetGraph
( TPhPadMap **maps, // Array of maps to be searched for selector
  Char_t *graphName,// Name of TGraphErrors for access via gROOT->GetObject
  Char_t *sel,      // Selector string such as "MPA3" or "MPA*"
  Int_t iVar,       // Index of plotted variable, defined in TPhMul.h
  Int_t iVarErr,    // Index of error variable, add 100 for linear combination
  Float_t norm,     // Normalizer, e.g. 360.0 to integrate on phi
// nXBins or nYBins should be 1; otherwise, we treat nYBins as 1.
  Int_t nXBins,     // Number of X-bins in histo
  Int_t nYBins,     // Number of Y-bins in histo
  Int_t nXMinSkip,  // Number of low X-bins to omit
  Int_t nYMinSkip,  // Number of low Y-bins to omit
  Int_t nXMaxSkip,  // Number of high X-bins to omit
  Int_t nYMaxSkip   // Number of high Y-bins to omit
)
{ return GetGraph (maps, 0, graphName, sel, -1, iVar, iVarErr, norm,
    nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, nYMaxSkip);
}

////////////////////////////////////////////////////////////////////////////

// General GetHist returns a pointer to a TH2F with data accumulated 
// from selected TPhPadMaps.  Alternate methods are more specialized.
TH2F *TPhPadMap::GetHist
( TPhPadMap **maps, // Array of maps to be searched for selector;
                    // if null, use this TPhPadMap
  Int_t nMaps,      // Number of maps to search;
                    // if 0, use nSensMax if histName = s..., else nDispMax
  Char_t *histName, // Name of TH2F for access via gROOT->GetObject
  Char_t *sel,      // Selector string such as "MPA3" or "MPA*"
  Int_t iMap,       // If >= 0, ignore selector and use maps[iMap]
  Int_t iVar,       // Index of histo'd variable, defined in TPhMul.h
  Int_t iVarWt,     // Index of weight variable, add 100 for square weight,
                    // -1 for weight 1.0
  Float_t norm,     // Normalizer, e.g. 360 to integrate on phi
  Int_t nXBins,     // Number of X-bins in histo
  Int_t nYBins,     // Number of Y-bins in histo
  Int_t nXMinSkip,  // Number of low X-bins to omit
  Int_t nYMinSkip,  // Number of low Y-bins to omit
  Int_t nXMaxSkip,  // Number of high X-bins to omit
  Int_t nYMaxSkip   // Number of high Y-bins to omit
)
{
// The Select method gives a match for name MOAB-1 and selector MOAB.
// There are additional problems.  Until these are fixed, use 
// specific selectors such as MOAB-*1 or specify iMap.

  Bool_t SumSq = iVarWt >= 100;
  if (SumSq) iVarWt = iVarWt % 100;
// printf ("GetHist: iVar = %d, iVarWt = %d, sumsq = %d\n", iVar, iVarWt,
//   SumSq);
// local variables:
  Float_t    cBin[2] = { 0., 0. };
  Size_t     cMax[2] = { 0., 0. };
  Size_t     cMin[2] = { 0., 0. };
 static
  Int_t      iprint = 0;
  TPhPadMap *map;
  Int_t      nBins[2] = { 0, 0 };
  Int_t      Nprint = 50; // 20;
  Int_t      nSel = 0;
  Int_t      ReportLevel = 2; // Set to 3 to show stats on selected maps
  TH2F *th = 0; // Created histogram
  Char_t    *title = sel;
// Replace selector and title if called by map index (overrides selector)
  if (iMap >= 0)
  { sel = "";
    if (maps[iMap]) title = maps[iMap]->GetName();
    else title = "NOMAP";
  }

// For backward compatibility, if nMaps <= 0, discriminate on the first 
// character of the object name and assume all maps are active

  if (nMaps <= 0) nMaps =
    (histName[0] == 's' || histName[0] == 'S')? kNSensMax : kNDispMax;

// Find the total x- and y- coverage of map[s] selected by sel or iMap.

  Int_t iFirstMap = -1;
  Int_t iMapExact = -1;
  Int_t mMin = (iMap >= 0)? iMap : 0;
  Int_t mMax = (iMap >= 0)? iMap : nMaps-1;
  for (int mm = mMin; mm <= mMax; mm++)
  { map = maps[mm];
// Look for an exact match to avoid double-counting a wild-card display
    if (map && (strcmp(map->GetName(),sel) == 0 || mm == iMap)) iMapExact = mm;
    if (map && (iMapExact >= 0 || Select(map->GetName(), sel)))
// Match found, set variable ranges.
    { for (int ixy = 0; ixy < 2; ixy++)
// TPhPadMaps use indices 0, 1 for x, y.
      { Float_t c1 = map->GetMin(ixy);
        Float_t c2 = c1 + map->GetSize(ixy) * map->GetNCells(ixy);
        if (c1 > c2) { Float_t tt = c1; c1 = c2; c2 = tt; }
        if (nSel == 0 || iMapExact >= 0)
        { cMin[ixy] = c1;
          cMax[ixy] = c2;
          iFirstMap = mm;
        }
        else
        { if (c1 < cMin[ixy]) cMin[ixy] = c1;
          if (c2 > cMax[ixy]) cMax[ixy] = c2;
        }
      } // for (ixy
// Stop if exact match was found
      if (iMapExact >= 0)
      { mMax = mm-1;
        nSel = 1;
      }
      else
        nSel++;
    } // if (iMapExact ...
  } // for (mm

// Take default # bins from the first found map.
  if (nSel)
  { map = maps[iFirstMap];
    for (int ixy = 0; ixy < 2; ixy++)
      nBins[ixy] = map->GetNCells(ixy);
  }
  else
  { if (ReportLevel >= 2 && iprint < Nprint)
    { printf ("TPhPadMap/GetHist: no maps found for selector %s.\n", 
        sel);
      iprint++;
    }
  }

// Warn if indices are invalid.

  if (nSel && (iVar < 0 || iVar >= map->GetNVars()))
  { if (ReportLevel >= 2 && iprint < Nprint)
    { printf ("TPhPadMap/GetHist: var index %d invalid for map"
        " %s.\n", iVar, map->GetName());
      iprint++;
    }
  }

  if (nSel && iVar >= 0 && iVar < map->GetNVars())

  { if (nXBins > 0) nBins[0] = nXBins;
    if (nYBins > 0) nBins[1] = nYBins;
// Skip bins as specified in arguments.  This should normally be done at a
// higher level, but is included here to allow developer scripts to skip
// edge bins with very low statistics.
    Int_t nMinSkip[2], nMaxSkip[2];
    nMinSkip[0] = nXMinSkip;
    nMaxSkip[0] = nXMaxSkip;
    nMinSkip[1] = nYMinSkip;
    nMaxSkip[1] = nYMaxSkip;
    for (int ixy = 0; ixy < 2; ixy++) 
    { cBin[ixy] = (cMax[ixy] - cMin[ixy]) / nBins[ixy];
      if (nMinSkip[ixy] > 0)
      { cMin[ixy] = cMin[ixy] + nMinSkip[ixy] * cBin[ixy];
        nBins[ixy] = nBins[ixy] - nMinSkip[ixy];
      }
      if (nMaxSkip[ixy] > 0)
      { cMax[ixy] = cMax[ixy] - nMaxSkip[ixy] * cBin[ixy];
        nBins[ixy] = nBins[ixy] - nMaxSkip[ixy];
      }
    }
    if (ReportLevel >= 3 && iprint < Nprint)
    { printf ("TPhPadMap/GetHist: %d / %d maps matched %s,\n x = [%f, %f],"
        " in %d bins,\n y = [%f, %f] in %d bins.\n", nSel, nMaps, sel, 
        cMin[0], cMax[0], nBins[0], cMin[1], cMax[1], nBins[1]);
      iprint++;
    }

// Create empty TH2F's for values and weights, covering the full range.

    th = new TH2F(histName, title, nBins[0], cMin[0], cMax[0],
                  nBins[1], cMin[1], cMax[1]);
    TH2F *thw = new TH2F("MulTemp99", "ww", nBins[0], cMin[0], cMax[0],
                  nBins[1], cMin[1], cMax[1]);

// Fill the TH2F's from all the sensors using specified weighting, normally
// dEtadPhi.  This allows us to combine sensors such as MO*, for which the 
// phi- coverage of different sensors is different.  However, the safest 
// way to get an integrated display is to create a new display like *+in 
// with a single phi bin and let TPhMul combine the errors correctly.

    Int_t mMin = (iMapExact >= 0)? iMapExact : 0;
    Int_t mMax = (iMapExact >= 0)? iMapExact : nMaps-1;
    for (int mm = mMin; mm <= mMax; mm++)
    { map = maps[mm];
      if (map && Select(map->GetName(), sel))
      { Axis_t cc[2];
	cc[0] = map->GetMin(0) + 0.5 * map->GetSize(0);
        for (int ix = 0; ix < map->GetNCells(0); ix++)
	{ cc[1] = map->GetMin(1) + 0.5 * map->GetSize(1);
          for (int iy = 0; iy < map->GetNCells(1); iy++)
	  { Float_t *ll = map->GetLoc(iVar, ix, iy);
            Float_t one = 1.0;
            Float_t *lw = (iVarWt >= 0)? map->GetLoc(iVarWt, ix, iy) : &one;
            if (ll && lw) 
// Only combine cells with non-zero weights.
	    { if (*lw)
              { Stat_t vv = (*ll) * (*lw);
                if (!SumSq) th->Fill (cc[0], cc[1], vv);
                else        th->Fill (cc[0], cc[1], vv * vv);
                thw->Fill (cc[0], cc[1], *lw);
                if ((ReportLevel >= 4 && *ll || ReportLevel >= 5)
                  && iprint < Nprint)
                { printf ("TPhPadMap/GetHist: %s added %10.2e to %s"
                   " at x = %8.3f, y = %8.2f.\n", map->GetName(), 
                   vv, histName, cc[0], cc[1]);
                  iprint++;
                }
              }
            }
            else
            if (!ll && ReportLevel >= 1 && iprint < Nprint)
            { printf ("TPhPadMap/GetHist: ix = %d, iy = %d invalid for %s.\n",
              ix, iy, map->GetName());
              iprint++;
            }
            cc[1] += map->GetSize(1);
          } // for (int iy
          cc[0] += map->GetSize(0);
        } // for (int ix
      } // if (map && Select
    } // for (int mm

// Normalize the combined histogram including the caller's norm
// Note the indexing ix+1, iy+1 skips over the underflow bins 

    for (Int_t ix = 0; ix < th->GetXaxis()->GetNbins(); ix++)
      for (Int_t iy = 0; iy < th->GetYaxis()->GetNbins(); iy++)
      { Stat_t ww = (iVarWt >= 0)? thw->GetCellContent (ix+1, iy+1) : 1.0;
        if (ww)
        { Stat_t vv = th->GetCellContent (ix+1, iy+1);
          if (!SumSq) th->SetCellContent (ix+1, iy+1, vv / ww * norm);
          else        th->SetCellContent (ix+1, iy+1, sqrt(vv) / ww * norm);
        }
      }
// Delete weight histogram
    if (thw) thw->Delete();

// Set default axis labels.
// Problem: the TPhPadMap type variable is not correctly implemented; 
// for now use norm to define the coordinate system.

    Char_t eta[] = "Eta";
    Char_t phi[] = "Phi";
//    Char_t dnde[] = "dN/dEta";
    Char_t z[] = "Z, cm.";
    Char_t ly[] = "Local Y, cm.";
    Char_t *xTitle;
    Char_t *yTitle;
    if (norm == 1.)
    { xTitle = z;
      yTitle = ly;
    }
    else
    { xTitle = eta;
      yTitle = phi;
    }
    th->SetXTitle(xTitle);    
    th->SetYTitle(yTitle);    

  } // else 

  return th;
}

////////////////////////////////////////////////////////////////////////////

// Specialized GetHist returns a pointer to a TH2F with data accumulated 
// from the TPhPadMap whose GetHist is called.
TH2F *TPhPadMap::GetHist
( Char_t *histName,    // Name of TH2F for access via gROOT->GetObject
  Int_t iVar,          // Index of histo'd variable, defined in TPhMul.h
  Int_t iVarWt,        // Index  of weight variable, add 100 for square weight,
                       // -1 for weight 1.0
  Float_t norm,        // Normalizer, e.g. 360 to integrate on phi
  Int_t nXBins,        // Number of X-bins in histo
  Int_t nYBins,        // Number of Y-bins in histo
  Int_t nXMinSkip,     // Number of low X-bins to omit
  Int_t nYMinSkip,     // Number of low Y-bins to omit
  Int_t nXMaxSkip,     // Number of high X-bins to omit
  Int_t nYMaxSkip      // Number of high Y-bins to omit
)
{ TPhPadMap *maps[] = { this };
  return GetHist (maps, 1, histName, "", 0, iVar, iVarWt, norm, 
    nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, nYMaxSkip);
}

////////////////////////////////////////////////////////////////////////////

// Specialized GetHist omits two arguments, for temporary compatibility
// with old invocations.
TH2F *TPhPadMap::GetHist
( TPhPadMap **maps,    // Array of maps to be searched for selector
  Char_t *histName,    // Name of TH2F for access via gROOT->GetObject
  Char_t *sel,         // Selector string such as "MPA3" or "MPA*"
  Int_t iVar,          // Index of histo'd variable, defined in TPhMul.h
  Int_t iVarWt,        // Index of weight variable, add 100 for square weight,
                       // -1 for weight 1.0
  Float_t norm,        // Normalizer, e.g. 360 to integrate on phi
  Int_t nXBins,        // Number of X-bins in histo
  Int_t nYBins,        // Number of Y-bins in histo
  Int_t nXMinSkip,     // Number of low X-bins to omit
  Int_t nYMinSkip,     // Number of low Y-bins to omit
  Int_t nXMaxSkip,     // Number of high X-bins to omit
  Int_t nYMaxSkip      // Number of high Y-bins to omit
)
{ // Use nMaps = 0 for backwards compatibility
  return GetHist (maps, 0, histName, sel, -1, iVar, iVarWt, norm, 
    nXBins, nYBins, nXMinSkip, nYMinSkip, nXMaxSkip, nYMaxSkip);
}

////////////////////////////////////////////////////////////////////////////

Int_t TPhPadMap::GetId()
{ return fS.Id;
}

////////////////////////////////////////////////////////////////////////////

Float_t TPhPadMap::GetIntegral
( Int_t   iVar,    // Variable #, e.g., 13 for d2N/dEta/dPhi rec
  Int_t   iVarRMS, // RMS error on var
  Float_t xMin,    // Range of integration
  Float_t yMin,
  Float_t xMax,
  Float_t yMax
)
{ Float_t stats[1];
  Int_t nRet =  GetStats (iVar, iVarRMS, 1, stats, xMin, yMin, xMax, yMax);
  if (nRet != 1) stats[0] = EunSpec;
  return stats[0]; 
}

////////////////////////////////////////////////////////////////////////////

Float_t *TPhPadMap::GetLoc (Int_t i1, Int_t i2, Int_t i3)
{ if (i1 < fNvars && i2 < fS.NXCells && i3 < fS.NYCells)
    return &fData[(i1 * fS.NXCells + i2) * fS.NYCells + i3];
  else
    return 0;
}

////////////////////////////////////////////////////////////////////////////

Int_t TPhPadMap::GetMaxEvents() { return fMaxEvents; }

////////////////////////////////////////////////////////////////////////////

Int_t TPhPadMap::GetNEvents() { return fNEvents; }

////////////////////////////////////////////////////////////////////////////

Float_t TPhPadMap::GetMin(Int_t ixy)
{ return (ixy == 0) ? fS.Xmin : fS.Ymin;
}

////////////////////////////////////////////////////////////////////////////

Char_t *TPhPadMap::GetName() { return fS.Name; }

////////////////////////////////////////////////////////////////////////////

Int_t TPhPadMap::GetNCells(Int_t ixy)
{ return (ixy == 0) ? fS.NXCells : fS.NYCells;
}

////////////////////////////////////////////////////////////////////////////

Int_t TPhPadMap::GetNVars()
{ return fNvars;
}

////////////////////////////////////////////////////////////////////////////

Float_t TPhPadMap::GetSize(Int_t ixy)
{ return (ixy == 0) ? fS.Xsize : fS.Ysize;
}

////////////////////////////////////////////////////////////////////////////

Int_t TPhPadMap::GetStats
( Int_t    iVar,    // Variable #, e.g., 13 for d2N/dEta/dPhi rec
  Int_t    iVarRMS, // Variable # for assigned RMS err, e.g. 14 for 13
  Int_t    nStats,  // number of statistics, <= 14, to store in stats 
  Float_t *stats,   // User buffer to receive statistics
  Float_t  xMin,    // Range of integration
  Float_t  yMin,
  Float_t  xMax,
  Float_t  yMax
)
{ 
// GetStats returns the number of words stored in stats, or its negative
// if any statistic was uncomputable, flagged by returning -99999.
  const Float_t unComp = -99999;
// Statistics in this version are: [0] integral(Var(x,y)); [1] <Var>;
// [2] <x>; [3] <y>; [4] xRms; [5] yRms; [6] x-y correlation coefficient;
// [7]-[8] rms assigned to [0]-[1], if iVarRMS >= 0; [9] # entries, 
// [10] <Var> unweighted, [11] <Var> unweighted RMS, [12-13] unused.
// These will change.

// These weightings use a broken straight line fit, which will change to 
// a quadratic fit later.

// A semantic problem with stats is that if GetStats returns a pointer to
// the internal buffer ss, it will be out of scope on return and the user
// will get garbage.  We could return a pointer to a buffer on the heap,
// but that loads the user with the responsibility of deleting it.  On
// the other hand, if the user calls with nStats > 0 but doesn't allocate
// space, he could get a crash.  I've tried to make this as safe as 
// possible without burdening the user to create a receiving structure.

  Float_t fSXmax = fS.Xmin + fS.Xsize * fS.NXCells;
  Float_t fSYmax = fS.Ymin + fS.Ysize * fS.NYCells;
  const Int_t maxStats = 14;
  Float_t ss[maxStats];
  Int_t is;
  for (is = 0; is < maxStats; is++) ss[is] = 0.; 
  Int_t iXMin = (xMin != EunSpec)? int((xMin - fS.Xmin) / fS.Xsize) : 0;
  Int_t iYMin = (yMax != EunSpec)? int((yMin - fS.Ymin) / fS.Ysize) : 0;
  Int_t iXMax = (xMin != EunSpec)? int((xMax - fSXmax) /fS.Xsize) : fS.NXCells;
  Int_t iYMax = (yMax != EunSpec)? int((yMax - fSYmax) /fS.Ysize) : fS.NYCells;
  Float_t wXY = 0;
  Float_t wXYSq = 0;
  Int_t nX = iXMax - iXMin + 1;
  Int_t nY = iYMax - iYMin + 1;
  for (Int_t iX = iXMin; iX < iXMax; iX++)
  { Float_t x = fS.Xmin + iX * fS.Xsize;
    Float_t xsq = x * x;
    for (Int_t iY = iYMin; iY < iYMax; iY++)
    { Float_t y = fS.Ymin + iY * fS.Ysize;
      Float_t ysq = y * y;
      Float_t xy = x * y;
      Float_t f = GetVal (iVar, iX, iY);
// Later, we'll choose dX and dY for occupied bins, and also perhaps choose 
// better weighting function.
      Float_t dXY = fS.Xsize * fS.Ysize;
      if ((iX == 0 || iX == iXMax - 1) && nX >= 3) dXY *= 0.5;
      if ((iY == 0 || iY == iYMax - 1) && nY >= 3) dXY *= 0.5;
      Float_t dN = f * dXY; 
      Float_t xN = x * dN;
      Float_t yN = y * dN;
      wXY += dXY;
      ss[0] += dN; // Integral
//      ss[1] += dN; // <Var>
// If Var is d2N/dEta/dPhi, multiply <Var> by 360 to get <dN/dEta>
      ss[2] += xN; // <X>
      ss[3] += yN; // <Y>
      ss[4] += xsq * dN; // Xrms
      ss[5] += ysq * dN; // Yrms
      ss[6] += xy * dN; // XY correlation
      Float_t ef = (iVarRMS >= 0)? GetVal (iVarRMS, iX, iY) : 0;
      Float_t dXYSq = dXY * dXY;
      wXYSq += dXYSq; 
      Float_t eNSq = ef * ef * dXYSq;
// RMS errors
      ss[7] += eNSq;
//      ss[8] += eNsq;
// Statistics for unweighted sums (not densities)
      ss[9]++;
      ss[10] += f;
      ss[11] += f * f;
// Rest are not yet useful
//      ss[12] += xsq * ysq * dN;
//      ss[13] += ysq * ysq * dN;
    }
  }
// Normalize, setting unnormalizable results to unComp.
  Float_t w = (ss[0])? 1.0 / ss[0] : unComp;
  Float_t sq;
  ss[1] = (wXY)? ss[0] / wXY : unComp; // * 360 if f = d2N/dEta/dPhi
  ss[2] = w * ss[2];
  ss[3] = w * ss[3];
  sq = w * ss[4] - ss[2] * ss[2]; 
  ss[4] = (sq >= 0)? sqrt(sq) : unComp;
  sq = w * ss[5] - ss[3] * ss[3]; 
  ss[5] = (sq >= 0)? sqrt(sq) : unComp;
  Float_t sigXsigY = ss[4] * ss[5];
  ss[6] = (sigXsigY)? (w * ss[6] - ss[2] * ss[3]) / sigXsigY : unComp;

//  Float_t w2 = (wXYSq >= 0)? 1.0 / sqrt(wXYSq) : unComp; // Not yet used
  ss[7] = (ss[7] >= 0)? sqrt(ss[7]) : unComp;
  ss[8] = w * ss[7];

  if (ss[9])
  { ss[10] /= ss[9];
    sq = ss[11] / ss[9] - ss[10] * ss[10];
    ss[11] = (sq >= 0)? sqrt(sq) : unComp;
  }
//  sq = ss[12] - ss[5] * ss[5];
//  ss[12] = (sq >= 0)? sqrt(sq) : unComp;
//  sq = ss[13] - ss[6] * ss[6];
//  ss[13] = (sq >= 0)? sqrt(sq) : unComp;

// Return requested stats

  if (nStats > maxStats) nStats = maxStats;
  Int_t nFail = 0;
  for (is = 0; is < nStats; is++)
  { if (ss[is] == unComp) nFail++;
    if (stats) stats[is] = ss[is];
  }
  return (nFail)? -nStats : nStats;
}

////////////////////////////////////////////////////////////////////////////

Int_t   TPhPadMap::GetType() { return fS.Type; }

////////////////////////////////////////////////////////////////////////////

Float_t TPhPadMap::GetVal(Int_t ivar, Int_t ix, Int_t iy)
{ Float_t *tt = GetLoc(ivar, ix, iy);
  return (tt)? *tt : 0.;
}

////////////////////////////////////////////////////////////////////////////

Float_t TPhPadMap::GetVersion() { return fS.Version; }

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::Map (Float_t *vars, Int_t isens, Int_t ix, Int_t iy,  
                     Float_t xy[2][2][2], Int_t *DisUp, TPhPadMap **Disps)
{
// Map a pad's variables from the array vars into display variable space.
// References to fS refer to private data members of the PadMap 
// whose method is invoked by the caller.
// In xy[i][j][k], i = 0 for Cartesian (not used), 1 for Eta/Phi coordinates;
// j = coordinate: 0 for X/Eta, 1 for Y/Phi;
// k = 0,1 for coordinate limits.
// This is still somewhat hybrid, since the pad's map variables are used
// to get the size.  We could also pass them in the vector, freeing us
// from dependence on the method.
//
// Date     Vers  Description
// 07-Nov-97 0.2  Added method SetData (RV)
// 28-Oct-97 0.2  Get input info from arrays, not a PadMap (RV)
// 13-Oct-97 0.1  Creation (Robin Verdier)

  Int_t Nprint = 50;
  Int_t ReportLevel = 2;

  static Int_t iprint = 0;
  Int_t dCell[2], ivar, ixy, nin, Nvinc;
  Float_t tt;

// Loop over displays whose indices are listed in array DisUp:

  for (Int_t iup = 0; iup < DisUp[0]; iup++)
  { Int_t idisp = DisUp[iup+1];
    TPhPadMap *disp = Disps[idisp];

// Check for a non-zero pointer (unfortunately we can't check for an
// invalid non-zero pointer
    if (disp)
    { if (ReportLevel >= 4 && iprint < Nprint)
      { printf ("PadMap/Map: iup = %d, map to idisp %d = %s.\n", iup, idisp,
        disp->GetName());
        iprint++;
      }
  
      Nvinc = disp->GetNVars();

// Only map the used variables.

      if (Nvinc > EDNVAcc) Nvinc = EDNVAcc;

// Find the number of physics coordinates in the range of this display
// The azimuthal angle phi poses a problem because it may not lie in
// the range 0, 360.  This is a generic problem made more difficult by
// the fact that there's not yet a "coordinate type" variable.
// The solution taken here is to add 360 to coordinate #1 if it is < 0
// and if the display minimum is <= 0.  This problem deserves a better
// solution.

      nin = 0;

      for (ixy = 0; ixy < 2; ixy++)

      { tt = 0.5 * (xy[1][ixy][0] + xy[1][ixy][1]);
        Float_t dmin = disp->GetMin(ixy);
        Float_t twoPi = 360.;
        if (ixy == 1 && tt < 0 && dmin >= 0) tt += twoPi;
        tt = (tt - dmin) / disp->GetSize(ixy);
        if (tt >= 0.)
        { dCell[ixy] = Int_t(tt); 
          if (dCell[ixy] < disp->GetNCells(ixy)) nin++;
        }
        if (ReportLevel >= 4 && iprint < Nprint)
        { printf ("PadMap/Map: ixy=%d, min xy=%10.2e, tt=%10.2e, sens=%d,"
            " disp=%d, nin=%d.\n", ixy, xy[1][ixy][0], tt, isens, idisp, nin);
          iprint++;
        }
      } // for (ixy

// If both coordinates were in range, increment the display bins.
// This may eventually be done according to a set of rules specified
// by a Reconstruction Control object; for now, we just accumulate.

      if (nin == 2)

      { for (ivar = 0; ivar < Nvinc; ivar++)
        { Float_t *ddo = disp->GetLoc(ivar, dCell[0], dCell[1]);
          if (ddo) *ddo += vars[ivar];
	  if (ddo != 0 && idisp == 26 && dCell[0] == 8 && dCell[1] == 0)
          { if (ReportLevel >= 5 && iprint < Nprint)
            { if (ivar == 0)
                printf ("PadMap/Map %s[%d][%d] vars for disp bin 0 8, ddo=%p:",
                  disp->GetName(), iy, ix, ddo);
              printf (" %8f", *ddo);
              if (ivar == Nvinc - 1)
              { printf ("\n");
                iprint++;
	      }
            }
	  } // if (ddo
        } // for (ivar

// Debug printout for every store into a selected set of display bins:
        if (ReportLevel >= 5 && iprint < Nprint
	    && idisp >= 64 && idisp <= 65
            && dCell[0] >= 24 && dCell[0] <= 24 /* && dCell[1] <= 1 */)
        { Int_t iph = dEtadPhi;
          Char_t iphname[] = "dEtadPhi";
          Int_t itr = nTracksPrimMC;
          Char_t itrname[] = "MC primary tracks";
          printf ("PadMap/Map %s[%d][%d], eta=%7.4f, phi=%7.3f", 
            fS.Name, iy, ix, 0.5*(xy[1][0][0]+xy[1][0][1]),
            0.5*(xy[1][1][0]+xy[1][1][1]));
          printf (" to disp %s[%d][%d] in [%d][%d]\n", 
            disp->GetName(), dCell[1], dCell[0], 
            disp->GetNCells(1), disp->GetNCells(0));
          printf ("%7.3f %s with %s = %7.4f,"
            " total %s = %7.4f.\n", vars[itr], itrname, iphname, vars[iph],
            iphname, *disp->GetLoc(iph, dCell[0], dCell[1]));
          iprint++;
        } // if (Report
      } // if nin
    }

    else

// Conditional warning for non-existent display
    { if (ReportLevel >= 2 && iprint < Nprint)
      { printf ("PadMap/Map: ignored request to map sensor %d to non-exis"
          "tent display %d.\n", isens, idisp);
        iprint++;
      }
    } // if (disp
  } // for (iup
}

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::Print (Int_t firstVar, Int_t lastVar)
{ Int_t nPerLine = 8;
  printf ("TPhPadMap %s has %d x %d bins and %d variables\n", 
    fS.Name, fS.NXCells, fS.NYCells, fNvars);
  if (firstVar < 0)
  { printf ("Use: Print(Int_t firstVar [, Int_t lastVar])\n");
    printf ("For display maps, variables 0...27 are\n");
    printf
      (" 0: dEtadPhi,   nPads,   Edep, E1, nTracks,\n"
       " 5: nTracksLim, FSec,    nTracksPrim, nTracksMC, nTracksPrimMC,\n");
    printf
      ("10: EdepPrimMC, Occ,     OccRMS,   d2NdEtadPhi,  d2NdEtadPhiRMS,\n"
       "15: EdepPTMC, EdepPTRMSMC, EdepPPTMC, EdepPPTRMSMC, FSecMC\n");
    printf
      ("20: FSecRMSMC,  OccMC,   OccRMSMC, d2NdEtadPhiMC, d2NdEtadPhiRMSMC,\n"
       "25: EdepPPad,   E1Avg,   FSecAvg\n");
    printf ("For sensor maps, variables 0...3 are\n");
    printf ("    ESEdep, ESNTracksMC, ESNTracksPrimMC, ESEdepPrimMC\n");
  }
  else
  { if (lastVar < 0) lastVar = firstVar;
    if (lastVar >= fNvars) lastVar = fNvars - 1;
    for (Int_t iVar = firstVar; iVar <= lastVar; iVar++)
    { Int_t iXMax = -1;
      printf ("Variable %d:\n", iVar);
      while (iXMax < fS.NXCells - 1)
      { Int_t iXMin = iXMax + 1;
        iXMax = iXMin + nPerLine - 2;
        if (iXMax > fS.NXCells - 1) iXMax = fS.NXCells - 1;
        printf ("  Y   X = ");
// Head rows and columns with BIN-CENTER values
        for (Int_t ix = iXMin; ix <= iXMax; ix++)
          printf ("%10.2e", fS.Xmin + (ix + 0.5) * fS.Xsize);
        printf ("\n");
        for (Int_t iy = 0; iy < fS.NYCells; iy++)
        { printf ("%10.2e", fS.Ymin + (iy + 0.5) * fS.Ysize);
          for (Int_t ix = iXMin; ix <= iXMax; ix++)
            printf ("%10.3e", GetVal(iVar, ix, iy));
          printf ("\n");
        }
        printf ("\n");
      } // while (iXMax
    } // for (Int_t iVar
  } // else
}

////////////////////////////////////////////////////////////////////////////

// Print a formatted dump of statistical variables to stdout.

void TPhPadMap::PrintStats
( Int_t iVar,   // Variable #, e.g., 13 for d2N/dEta/dPhi rec
  Int_t iVarRMS // Variable # for assigned RMS err
)
// Statistics in this version are: [0] integral(Var(x,y)); [1] <Var>;
// [2] <x>; [3] <y>; [4] xRms; [5] yRms; [6] x-y correlation coefficient;
// [7]-[6] rms assigned to [0]-[6], if iVarRMS >= 0.  These may change.
{ const Int_t maxStats = 14; // These ought to be global to TPhPadMap
  Float_t ss[maxStats];
  Char_t *vNames[14] =
  {"Int    ", "Avg    ", "<x>    ", "<y>    ", "xRms   ", "yRms   ",
   "xyCorr ", "IntRms ", "AvgRMS ", "nPts   ", "UnwAvg ", "Sigma  ", "", ""
  }; 
  Int_t is;
  if (iVar < 0)
  { printf ("Use: PrintStats (Int_t iVar [, Int_t iVarRMS])\n");
  }
  else
  { is = GetStats (iVar, iVarRMS, maxStats, ss);
    for (is = 0; is < maxStats; is++)
    { printf ("[%2d] %7s=%12.4e", is, vNames[is], ss[is]);
      if (!((is+1)%3) || is == maxStats-1) printf ("\n");
      else printf (", ");
    }
  }
}

////////////////////////////////////////////////////////////////////////////

Float_t *TPhPadMap::Release()
{ Float_t *res = fData;
  if (res != 0) 
  { delete[] fData;
    fData = 0;
  }
  return res;
}

////////////////////////////////////////////////////////////////////////////

Int_t TPhPadMap::Select(Char_t *name, Char_t *selectors)
{
// Select matches the character strings name and selectors and returns 
// the number of matches found, or 0 if at least one veto succeeded.
// selectors is of the form [<type>]<selector>[<type><selector>...]
// where type is '+' for a requirement or '-' for a veto.
// A selector field may contain the wild card character '*'.
// The initial type is assumed to be '+' if not explicitly present.
// Examples:
// Select("MOA1", "MO*"); returns 1.
// Select("MOA1", "T*+MOB*"); returns 0.
// Select("MOA1", "*"); returns 1.
// Select("MOA1", "*-MO*1"); returns 0.
// Strings can be of arbitrary length.
// Strings are explicitly assumed to be null-terminated.
// The matching is case-sensitive.  Case-insensitive matching can be
// achieved by using library functions to case the arguments.
//
// 08-Feb-98: Creation (Robin Verdier)

  Bool_t      exit = kFALSE;   // End of selectors
  Int_t       in = 0;          // Index in name
// iprint need not be declared static here, because the function is static.
  Int_t       iprint = 0;      // # of lines of debug printout
  Int_t       is = 0;          // Index in select
  const Int_t match = 0;       // Looking for next match
  Int_t       mode = match;    // search mode = match, skip, or wild
  Int_t       Nprint = 50;     // # debug printouts allowed
  Int_t       passes = 0;      // Number of matching fields in match mode
  Int_t       ReportLevel = 0; // 0 for no debug output
  const Int_t skip  = 1;       // Skip to next selector field
  Bool_t      veto = kFALSE;   // True if current selector is a veto
  Int_t       vetoes = 0;      // Number of fields producing a veto
  const Int_t wild  = 2;       // Skip to matching character in name

// Loop comparing characters in name with those in select

  while (!exit)
  { Char_t s = selectors[is++];
    Char_t n = name[in];

    if ( s == 0             // If selector character is End-of-selectors,
      || s == '+'           // start of required field, or
      || s == '-')          // start of veto-field:
      { if (veto)           // If the current selector is a veto,
        { if (mode != skip)
          { vetoes++;       // update # vetoes unless match failed
	    exit = kTRUE;   // and exit
          }
        }
        else
	  if (mode != skip) passes++; // else update # passes
        if (s == '+')       // If the current character in selectors is
        { veto = kFALSE;    // +, set to non-veto,
	  mode = match;     //    set mode to match,
          in = 0;           //    and return to start of name
        }
        else if (s == '-')  // -, set mode to veto and restart name
        { veto = kTRUE;
	  mode = match;     //    set mode to match,
          in = 0;
        }
        else
          exit = kTRUE;
      } // if (veto)

    else if (s == '*')   // wildcard:
    { if (mode != skip)  // change mode unless skipping to next field
        mode = wild;
    }

    else if (s == n)     // characters match:
    { in++;              // continue with next name character
    }

    else                 // characters don't match:
    { if (mode != wild)
        mode = skip;     // if not in wildcard mode, skip to next select field
      else               // if in wild card mode,
      { if (n)           //   and not at end of name,
        { is--;          //   stay on same selector character
          in++;          //   go to next name character
        }
        else             //   and at end of name,
        { mode = skip;   //   skip to next selector fiels
        }
      }
    } // if (s ==

    if (ReportLevel >= 2 && iprint < Nprint)
    { printf ("n = %c, s = %c, mode = %d, veto = %d.\n", char(n),
        s, mode, veto);
      iprint++;
    }
  } // while (!exit)
  return (vetoes)? 0 : passes;
}

// The set methods will all eventually return a status code.

////////////////////////////////////////////////////////////////////////////

Int_t TPhPadMap::SetData (Float_t val)
// Set contents to specified value; return total number of cells set
// 
{ Int_t jj; 
  for (jj = 0; jj < fSize; fData[jj++] = val);
  return jj;
}

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::SetDist(Float_t dist) { fS.Dist = dist; }

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::SetId(Int_t id) { fS.Id = id; }

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::SetMaxEvents(Int_t nEvs) { fMaxEvents = nEvs; }

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::SetMin(Int_t ixy, Float_t min)
{ if (ixy == 0) fS.Xmin = min; else if (ixy == 1) fS.Ymin = min; 
}

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::SetName(Char_t *t) { strncpy(fS.Name, t, EnameSize); }

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::SetNCells(Int_t nXCells, Int_t nYCells, Float_t val)
{ if (fData) delete[] fData; 
  fS.NXCells = nXCells;
  fS.NYCells = nYCells; 
  fSize = fNvars * fS.NXCells * fS.NYCells;
  // printf ("TPhPadMap: creating a data block of %i * %i * %i = %i words.\n", 
  //  fNvars, fS.NXCells, fS.NYCells, fSize);
  fData = new Float_t[fSize];
  for (Int_t kk = 0; kk < fSize; fData[kk++] = val);
}

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::SetNEvents(Int_t nev) { fNEvents = nev; }

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::SetSize(Int_t ixy, Float_t size)
{ if (ixy == 0) fS.Xsize = size; else if (ixy == 1) fS.Ysize = size; 
}

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::SetType(Int_t type) { fS.Type = type; }

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::SetVersion(Float_t vers)
{ fS.Version = vers;
}

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::SetVertex(TPhVertex *v)
{ fVertex = v;
}

////////////////////////////////////////////////////////////////////////////

void TPhPadMap::Streamer(TBuffer &R__b)
{
  // Stream an object of class TPhPadMap.
  // Hand-edited from MulCint.cxx

  if (R__b.IsReading()) {
    Version_t R__v = R__b.ReadVersion(); if (R__v) { }
    TObject::Streamer(R__b);
    R__b >> fSSize;
    //fS.Streamer(R__b);
      R__b.ReadFastArray((UChar_t *)&fS, fSSize);
      Int_t nBytesRead = fSSize;
      if (fSSize != nBytesRead) printf ("fS is %d bytes, read %d bytes.\n",
        fSSize, nBytesRead);
      R__b >> fMaxEvents;
      R__b >> fNEvents;
      R__b >> fNvars;
      R__b.ReadStaticArray(fSelect);
      R__b >> fSize;
      //R__b.ReadArray(fData);
      Int_t nFloatsRead = R__b.ReadArray(fData);
      if (fSize != nFloatsRead) printf ("fData is %d words, read %d words.\n",
        fSize, nFloatsRead);
  } else {
    R__b.WriteVersion(TPhPadMap::IsA());
    TObject::Streamer(R__b);
    R__b << fSSize;
    //fS.Streamer(R__b);
      R__b.WriteFastArray((UChar_t *)&fS, fSSize);
      R__b << fMaxEvents;
      R__b << fNEvents;
      R__b << fNvars;
      R__b.WriteArray(fSelect, 2);
      R__b << fSize;
      //R__b.WriteArray(fData, __COUNTER__);
      R__b.WriteArray(fData, fSize);
  }
}

////////////////////////////////////////////////////////////////////////////
