#include "TObject.h"
#include "TPhSensGeo.h"
#include "TPhSensDef.h"
#include "TPhSensDefV.h" // Default values

ClassImp(TPhSensDef)

////////////////////////////////////////////////////////////////////////////
//
// TPhSensDef is a sensor geometry definition class for Phobos multiplicity
// reconstruction.  Its data member is a struct, TPhSensGeo, defined in 
// TPhSensGeo.h, which can be copied into TPhPadMaps.
//
// Coordinates are defined as follows (cf. the discussion in TPhSensGeo.h):
// ixy = 0 or X is the eta-type coordinate; ixy = 1 or Y is the phi-type
// coordinate.  Thus for the barrel, ixy = 0 refers to world coordinate 
// Z and 1 to phi; for the rings, ixy = 0 is rho and 1 is phi.  Note that
// the barrel delta phi is not strictly constant for more than 2 Y-cells.
// The TPhSensor class's PxWidth is our Xsize and its Ncols is our NXcells;
// its PxHeight is the linear equivalent of our angle Ysize and its Nrows 
// is our NYcells; and its X0, X1, Y0, Y1, Z0, and Z1 are true world 
// coordinates.  Presently Z0 > Z1 and the Y-indexing is also reversed.
//
// Methods to set these values from the Sensor objects will be added.
//
// 13-Jan-99 v3  NXcells, NYcells -> NXCells, NYCells for consistency (RV)
// $id$
//
////////////////////////////////////////////////////////////////////////////

// 16-Dec-98 v3  Unused constructors commented out until needed (RV)
// 08-Dec-98 v3  Strengthen parallelism with TPhDispDef, work around apparent
//               SunOS compiler bug (RV)
// 15-Aug-98 1.0  Change default values filename from .v to V.h (RV)
// 08-Jul-98 1.0  ixy = 0, 1 for x, y (RV)
// 19-Jun-98 1.0  Change fSensors initializer for OSF1 (cf. TPhDispDef) (RV)
// 23-May-98 0.1  Added GetIndex and List methods (RV)
// 12-Mar-98 0.1  Corrected syntax errors for Abacus (RV)
// 10-Jan-98 0.1  Implement access methods (RV)
// 31-Oct-97 0.1  Creation (Robin Verdier)

// Enumerate return flags
  enum { OK = 0, badArg = -999991, unImplemented = -999992 };

TPhSensDef::TPhSensDef()
{ fSensors = gMulSensors;
  return;
}

/* Comment-out So-far unused special creators to avoid compiler warnings:

TPhSensDef::TPhSensDef(TPhSensDef *orig)
{ fSensors = gMulSensors;
  return;
}

TPhSensDef::TPhSensDef(Float_t Version, Char_t name[], Int_t id, Int_t type,
    Int_t nxcells, Float_t xmin, Float_t xsize,
    Int_t nycells, Float_t ymin, Float_t ysize,
    Float_t o1, Float_t o2, Float_t o3,
    Float_t a11, Float_t a12, Float_t a13,
    Float_t a21, Float_t a22, Float_t a23,
    Float_t a31, Float_t a32, Float_t a33 )
{ fSensors = gMulSensors;
  return;
}

*/  // End of commented-out constructors

TPhSensDef::~TPhSensDef() { return; }

Int_t TPhSensDef::Copy(TPhSensDef *source, Int_t isens, TPhSensGeo *copy)
{
// isens is the sensor index = 0, 1, 2, ..., kNSensMax-1.
// Returns 0 if successful, else badArg.
// This is a primitive implementation to be replaced by byte copy
// operations and, possibly, an overloaded = operator

  if (isens < 0 || isens >= kNSensMax || source == 0 || copy == 0)
    return badArg;
  else
  { copy->Version = source->fSensors[isens].Version;
    Int_t nn; // Avoids DEC / GNU disagreement about scope
// The following character-copy moves no bytes under SunOS:
//    for (nn=0; nn<5; copy->Name[nn]=source->fSensors[isens].Name[nn++]);
    strncpy (copy->Name, source->fSensors[isens].Name, EnameSize);
    copy->Id      = source->fSensors[isens].Id;
    copy->Type    = source->fSensors[isens].Type;
    copy->NXCells = source->fSensors[isens].NXCells;
    copy->Xmin    = source->fSensors[isens].Xmin;
    copy->Xsize   = source->fSensors[isens].Xsize;
    copy->NYCells = source->fSensors[isens].NYCells;
    copy->Ymin    = source->fSensors[isens].Ymin;
    copy->Ysize   = source->fSensors[isens].Ysize;
    for (nn=0; nn<3; copy->Orig[nn]=source->fSensors[isens].Orig[nn++]);
    for (nn=0; nn<3; nn++)
      for (int mm=0; mm<3;
        copy->Axes[nn][mm]=source->fSensors[isens].Axes[nn][mm++]);
    copy->Dist = source->fSensors[isens].Dist;
  }
  return OK;
}

Float_t TPhSensDef::GetAxis(Int_t ixyz, Int_t isens)
        { return (float) fSensors[isens].Axes[0][ixyz]; } 

Int_t   TPhSensDef::GetId(Int_t isens) { return fSensors[isens].Id; }

Int_t   TPhSensDef::GetIndex(Char_t *sensorName)
// Return sequential index of named active sensor, used in all Mul lists,
// or badArg (<< 0) if name is not in active list
{ Int_t isens;
  for (isens = kNSensMax-1; isens >= 0; isens--)
  { if ((fSensors[isens].Name[0]) 
    && !strcmp (sensorName, fSensors[isens].Name)) break;
  }
  return (isens >= 0)? isens : badArg;
}

Float_t TPhSensDef::GetMin(Int_t isens, Int_t ixy)
        { return (ixy == 0)? fSensors[isens].Xmin : fSensors[isens].Ymin; }

Char_t *TPhSensDef::GetName(Int_t isens) { return fSensors[isens].Name; }

Int_t TPhSensDef::GetNCells(Int_t isens, Int_t ixy)
      { return (ixy == 0)? fSensors[isens].NXCells : fSensors[isens].NYCells;
      }

Int_t TPhSensDef::GetNDefsAct()
// GetNDefsAct returns the number of active sensor definitions
// The # of definitions is determined by the contents of TPhSensDefV.h, but 
// we need to define the meaning of kNSensMax, say to allocate space so
// that users can add new definitions without recompiling.
{ Int_t geoSize = sizeof(TPhSensGeo);
  Int_t sensSize = sizeof(gMulSensors);
  Int_t nDefsAct = sensSize / geoSize;
  return nDefsAct; 
}

Float_t *TPhSensDef::GetOrig(Int_t isens)
         { return fSensors[isens].Orig; }

Float_t TPhSensDef::GetSize(Int_t isens, Int_t ixy)
        { return (ixy == 0)? fSensors[isens].Xsize : fSensors[isens].Ysize; }

Int_t  TPhSensDef::GetType(Int_t isens) { return fSensors[isens].Type; }

void  TPhSensDef::List(Int_t isMin = 0, Int_t isMax = kNSensMax-1)
       { for (Int_t is = isMin; is <= isMax; is++)
         { printf ("%3d:%5s ", is, fSensors[is].Name);
           if (!((is-isMin+1)%8) || is == isMax) printf ("\n");
         }
       }

void  TPhSensDef::SetAxis(Int_t isens, Int_t ixyz, Float_t val)
       { fSensors[isens].Axes[0][ixyz] = val; }

void  TPhSensDef::SetId(Int_t isens, Int_t id) {fSensors[isens].Id = id; }

void  TPhSensDef::SetMin(Int_t isens, Int_t ixy, Float_t min)
       { if (ixy == 0) fSensors[isens].Xmin = min;
         if (ixy == 1) fSensors[isens].Ymin = min;
       }

void  TPhSensDef::SetName(Int_t isens, Char_t *name)
       { strncpy(fSensors[isens].Name, name, EnameSize); }

void  TPhSensDef::SetNCells(Int_t isens, Int_t ixy, Int_t ncells)
       { if (ixy == 0) fSensors[isens].NXCells = ncells;
         if (ixy == 1) fSensors[isens].NYCells = ncells;
       }

void  TPhSensDef::SetOrig(Int_t isens, Float_t x, Float_t y, Float_t z)
       { fSensors[isens].Orig[0] = x;
         fSensors[isens].Orig[1] = y;
         fSensors[isens].Orig[2] = z; 
       }

void  TPhSensDef::SetSize(Int_t isens, Int_t ixy, Float_t size)
       { if (ixy == 0) fSensors[isens].Xsize = size;
         if (ixy == 1) fSensors[isens].Ysize = size;
       }

void  TPhSensDef::SetType(Int_t isens, Int_t type)
       { fSensors[isens].Type = type; }

void  TPhSensDef::Show(Int_t isens, Int_t ReportLevel)
        { if (ReportLevel == 1) printf ("SensDef: sensor%3d = %s\n",
          isens, fSensors[isens].Name);
          if (ReportLevel == 2) printf (" #  Name Type nXCells  Xmin  "
            "Xcell  nYCells   Ymin  Ycell    Dist\n");
          if (ReportLevel == 3) printf (
            "%3d %s %3d %6d%8.3f%7.3f%7d%8.3f%7.3f%9.3f\n",
          fSensors[isens].Id,      fSensors[isens].Name, fSensors[isens].Type,
          fSensors[isens].NXCells, fSensors[isens].Xmin, fSensors[isens].Xsize,
          fSensors[isens].NYCells, fSensors[isens].Ymin, fSensors[isens].Ysize,
          fSensors[isens].Dist);
        }

// End TPhSensDef implementation
