#include "TPhOnSensor.h"

// ROOT classes

#include "TH1.h"
#include "TCanvas.h"

// Phat classes

#include "TPhDetector.h"
#include "TPhSensor.h"
#include "TPhSiSensor.h"
#include "TSeqCollection.h"
#include "TPhSubDetector.h"
#include "TPhLayer.h"
#include "TPhHashContainer.h"
#include "TPhHitArray.h"


ClassImp(TPhOnSensor)
  ///////////////////////////////////////////////////////////////////////////
  //
  // This online class contains the set of histograms for a single sensor 
  // and member functions to operate them. It has the same name 
  // (with extension 'on') as the sensor to be monitored. So far, the 
  // histograms are: adc spectrum, #hits/pad and mean signal / noise.
  // To work with groups of sensors (e.g. monitor all sensors) use the 
  // class TPhOnSensorContainer.
  // 
  ///////////////////////////////////////////////////////////////////////////


TPhOnSensor::TPhOnSensor()
{
 // Default constructor. Does nothing use TPhOnSensor(const Char_t *name) 
 // instead.
  hadc      = NULL;
  hhit      = NULL;
  hnoise    = NULL;
  fDetector = TPhDetector::GetInstance();
  fNHistos  = 0;
}

//*********************************************************************
//*********************************************************************

TPhOnSensor::TPhOnSensor(const Char_t *name):TNamed(name,name)
{
  // constructor, set name, label etc. to initialize the histograms
  // use Init()
  fDetector = TPhDetector::GetInstance();
  if(!fDetector) {
    Warning("TPhOnSensor(name)","detector not connected   - return");
    return;
  }
  fNHistos  = 0;
  // shrink name to Phobos generic sensor name 
  strcpy(sname,name);
  sname[4] = '\0';
  //printf("sensor name %s\n",sname);
  TPhSensor *sensor = fDetector->GetSensor(sname);
  if(!sensor) Warning("TPhOnSensor(name)","sensor doesn't exits - return");
  fLabel = sensor->GetLabel();
  fNRow = sensor->GetRows();
  fNCol = sensor->GetColumns();
}

//*********************************************************************
//*********************************************************************

TPhOnSensor::~TPhOnSensor()
{
  // destructor; deletes histograms
 if(hadc) {
   delete hadc;
 }
 if(hhit) {
   delete hhit;
 }
 if(hnoise) {
   delete hnoise;
 }
};

void TPhOnSensor::Init()
{
  // Initialize adc, hit and noise histogram for this sensor. Sofar, one 
  // can't choose the histograms to be booked (e.g. only adc). If it turns 
  // out to be useful, the code can be changed (this means a lot of 
  // bookeeping).


  InitADC();
  InitHit();
  InitNoise();

}

void TPhOnSensor::InitADC(Int_t nadcchan)
{
  // initialize adc histogram - for internal use only
  if(hadc){
    Warning("InitADC","replace existing histogram");
    fNHistos--;
    delete hadc;
  }
  Char_t hname[256];
  Char_t htitle[256];
  sprintf(hname,"hadc%d",fLabel);
  sprintf(htitle,"hadc%s",sname);
  hadc = new TH1S(hname,htitle,nadcchan,0.,nadcchan-1); 
  fNHistos++;
}

void TPhOnSensor::InitHit()
{
  // initialize hit histogram - for internal use only
  if(hhit){
    Warning("InitHit","replace existing histogram");
    fNHistos--;
    delete hhit;
  }
  Int_t nchan = fNRow * fNCol;
  Char_t hname[256];
  Char_t htitle[256];
  sprintf(hname,"hhit%d",fLabel);
  sprintf(htitle,"hhit%s",sname);
  hhit = new TH1S(hname,htitle,nchan,0.,nchan-1); 
  fNHistos++;
}

void TPhOnSensor::InitNoise()
{
  // initialize noise histogram - for internal use only
  if(hnoise){
    Warning("InitNoise","replace existing histogram");
    fNHistos--;
    delete hnoise;
  }
  Int_t nchan = fNRow * fNCol;
  Char_t hname[256];
  Char_t htitle[256];
  sprintf(hname,"hnoise%d",fLabel);
  sprintf(htitle,"hnoise%s",sname);
  hnoise = new TH1S(hname,htitle,nchan,0.,nchan-1); 
  fNHistos++;
}

void TPhOnSensor::Draw(Option_t *option)
{
  // draw single histogram; the options are "adc","hit","noise"

  TString opt = option;
  opt.ToLower();
  if (opt.Contains("all")){
    Warning("Draw","only single histogram plotting is implemented use TPhOnSensorContainer::Draw(sensor) instead");
    return;
  }
  // ******** adc histo**************
  else if (opt.Contains("adc")){
    if(!hadc) {
      Warning("Draw","adc histogram doesn't exist - return");
      return;
    }
    hadc->Draw();
  }
  // ******** hit histo**************
  else if (opt.Contains("hit")){
    if(!hhit) {
      Warning("Draw","hit histogram doesn't exist - return");
      return;
    }
    hhit->Draw();
  }
  // ******** adc histo**************
  else if (opt.Contains("noise")){
    if(!hnoise) {
      Warning("Draw","noise histogram doesn't exist - return");
      return;
    }
    hnoise->Draw();
  }
  else {
    Warning("Draw","option doesn't exist");
  }
}

TH1S* TPhOnSensor::GetHistogram(Option_t *option)
{
  // returns single histogram; the options are "adc","hit","noise"
  TString opt = option;
  opt.ToLower();
  if (opt.Contains("adc")){
    return hadc;
  }
  else if (opt.Contains("hit")){
    return hhit;
  }
  else if (opt.Contains("noise")){
    return hnoise;
  }
  return NULL;
}

void TPhOnSensor::Reset()
{
  // reset all histograms 
  hadc->Reset();
  hhit->Reset();
  hnoise->Reset();
}

void TPhOnSensor::Fill()
{
  // fill all histograms; input is the hitarray stored in the 
  // corresponding sensor object; that means the event has to be added 
  // to the detector with TPhDetector::SetEvent() and 
  // TPhDetector::FillWithEventHitArrays("containername") first;

  if(!fDetector) {
    Warning("Fill","detector not connected  - return");
    return;
  }
  if(!hadc || !hhit || !hnoise){
    Warning("Fill","histograms not booked - return");
    return;
  }

  TPhSiSensor *sis = (TPhSiSensor*)fDetector->GetSensor(fLabel);
  if(!sis) {
    Warning("Fill","sensor doesn't exist  - return");
    return;
  }
  TPhHitArray *hitarray = sis->GetHitArray();
  if(!hitarray) {
    //    Warning("Fill","no hitarray in sensor  - return");
    return;
  }

  Int_t i;
  Int_t j;
  Short_t adc;
  Int_t ichan;      // pad nr linearised for easier display
  Int_t ncol = hitarray->GetColumns();
  for (i=0; i<hitarray->GetRows(); i++)
    {
      for(j=0; j<ncol; j++)
	{	
	  adc = hitarray->GetADC(i,j);
	  if(adc>0) 
	    {
	      hadc->Fill(adc);
	      ichan = i * ncol + j;
	      hhit->Fill(ichan);
	    }
	}
    }
  // sofar only the mean adc value is filled into the histogram (assuming 
  // noise = 1 for all channels). As soon as the noise data format will be 
  // defined the mean value should be devided by the noise of the channel 
  Stat_t peakadc;
  Stat_t peakadcmax=0;
  Int_t imax;
  for(i=0; i<hadc->GetNbinsX(); i++){
    peakadc = hadc->GetBinContent(i);
    if(peakadc>peakadcmax){
      peakadcmax = peakadc;
      imax = i;}	
  }
  for (i=0; i<fNRow*fNCol; i++)
    {
      hnoise->Fill(i,imax);
    }
}

Bool_t  TPhOnSensor::IsExistend(Option_t* option)
{
  // check whether a histogram has been booked; this is only needed 
  // if booking of a single histogram will be allowed in future.
  // The options are "all","adc","hit","noise".
  TString opt = option;
  opt.ToLower();
  if (opt.Contains("all"))
    {
      if(hadc && hhit && hnoise) return kTRUE;
    }
  else if(opt.Contains("adc"))
    {
      if(hadc)  return kTRUE;
    }
  else if(opt.Contains("hit"))
    {
      if(hhit)  return kTRUE;
    }
  else if(opt.Contains("noise"))
    {
      if(hnoise)  return  kTRUE;
    }
  return kFALSE;
}

Stat_t  TPhOnSensor::GetEntries(Option_t *option)
{
  // Get Entries of single histogram. The options are "adc","hit","noise".
  TString opt = option;
  opt.ToLower();
  if (opt.Contains("adc")){
    return hadc->GetEntries();
  }
  else if (opt.Contains("hit")){
    return hhit->GetEntries();
  }
  else if (opt.Contains("noise")){
    return hnoise->GetEntries();
  }
  else {
    Warning("GetEntries","option unknown - return 0");
    return 0;
  }
}





