#ifndef _TPhPID_INCLUDED_
#define _TPhPID_INCLUDED_
//#define DEBUG
//
///////////////////////////////////////////////////////////////////////
//                                                                   //
//    TPhPIDSpect                                                    //
//                                                                   //
//    Phobos PID class                                               //
//                                                                   //
//    TPhPIDSpect class is used to set probabilities for identyfing  //
//    a reconstructed track as a pion, kaon and proton using         //
//    the track energy depositions in silicon detectors in the       //
//    spectrometer, its momentum and trajectory. The PID algorithm   //
//    is based on MC precalculated PID bands.                        // 
//                                                                   //
//                                                                   //
///////////////////////////////////////////////////////////////////////

#include "TObject.h"

const Int_t   nhitsMax=20;                   // max number of hits allwed in one arm per track
const Int_t   nhitsMin=13;                  // min number of hits allwed in one arm per track
const Int_t   Ntrunb  = 1;                  // number of bottom truncations used 
const Int_t   Ntrunt  = 7;                  // number of top truncations used 
const Float_t ro_si   = 2.33;               // Si density

const Int_t   Nsigma=7;                     // distance from a mean in sigmas above it 
                                            //the weight is set = 0.
const Float_t r2 = 0.045, r3 = 0.035, r4 = 0.00;  // ratio k/pi,p/pi,e/pi
//const Float_t r2 = 1., r3 = 1., r4 = 1.;             // ratio k/pi, p/pi, e/pi


class TPhMCTrack;
class TPhTrack;
class TPhDetector;

class TPhPIDSpect : public TObject 
{ 
 public:
  TPhPIDSpect();
  virtual ~TPhPIDSpect();

  TPhPIDSpect(Char_t *FileName);           // the constructor calls ReadBandPara 
                                           // which reads in the PID band parameters
                                           // from the file PIDBands.dat
  virtual Bool_t  ReadBandPara(Char_t *FileName);
  virtual Int_t   DrawBands(Option_t *option); 
  virtual Int_t   DrawProfiles(Int_t k, Int_t gid); 

  virtual Float_t GetdEdxt(TPhMCTrack *t); // the function calculates the truncated dEdx 
                                           // under condition specified in Arm and GetOK
                                           // functions
  virtual Int_t   GetBand(TPhMCTrack *t, Int_t *iBand);
  virtual Int_t   GetProb(TPhMCTrack *t, Float_t *prob);
                                          // the function returns PID probabilities 
                                          // that the track is pi, K or p.
                                          // It calculates the truncated dEdx
                                          // - there is no need to call GetdEdxt 
                                          // first as the function does this.
  virtual Float_t ProfileFunction(Int_t pid, Float_t dedxtrd, Float_t p);
					  // The PID bands profile are described
                                          // by ProfileFunction
                                         
  virtual void GetOneHitBand(Float_t p, Float_t *Mean_RMS);
  virtual void    SetVerbose( Int_t n ) { fVerbose = n; }
  // some utility functions

  virtual Int_t   Arm(TPhMCTrack *t);
  virtual Int_t   GetOK(Int_t *t);
  virtual Float_t calc_truncmean(Float_t *g, Int_t n, Int_t start, Int_t stop);
  virtual void    Sort(Float_t *v, Int_t n); 
  virtual Float_t Interpol(Float_t p, Float_t a[], Int_t pid);

  // data members are the PID bands bondaries from p_(pi,k,p,e)-p_tot 
  // GeV/c in momenta  and parameters of the profiles in the same 
  // momenta range
 protected:
  Float_t p_tot;             // tot momentum interval  
   //data for pions
  Int_t   nbinsx_pi;         // number of  momentum bins for pi
  Float_t p_pi;              // lowest momentum for pi  
  Float_t fPiMomentum1[100]; // momentum in bins (lowedge)   
  Float_t fPiMomentum2[100]; // momentum in bins (centre)    
  Float_t fPiMomentum3[100]; // momentum in bins (upperedge) 
  Float_t fPiBandLo[100];    // lower PID band for pions 
  Float_t fPiBandHi[100];    // higher PID band for pions
  Float_t fPiBandMe[100];    // parameters of the pion PID band profile 
  Float_t fPiBandRM[100];    // parameters of the pion PID band profile
  //data for kaons
  Int_t   nbinsx_k;          // number of  momentum bins for k
  Float_t p_k;               // lowest momentum for k  
  Float_t fKMomentum1[100];  // momentum in bins (lowedge)   
  Float_t fKMomentum2[100];  // momentum in bins (centre)    
  Float_t fKMomentum3[100];  // momentum in bins (upperedge) 
  Float_t fKBandLo[100];     // lower PID band for kaons
  Float_t fKBandHi[100];     // higher PID band for kaons
  Float_t fKBandMea[100];    // parameters of the kaon PID band profile
  Float_t fKBandRMS[100];    // parameters of the kaon PID band profile
  //data for protons
  Int_t   nbinsx_p;          // number of  momentum bins for p
  Float_t p_p;               // lowest momentum for p  
  Float_t fPMomentum1[100];  // momentum in bins (lowedge)   
  Float_t fPMomentum2[100];  // momentum in bins (centre)    
  Float_t fPMomentum3[100];  // momentum in bins (upperedge) 
  Float_t fPBandLo[100];     // lower PID band for protons 
  Float_t fPBandHi[100];     // higher PID band for protons 
  Float_t fPBandMea[100];    // parameters of the proton PID band profile
  Float_t fPBandRMS[100];    // parameters of the proton PID band profile 
  // data for electrons
  Int_t   nbinsx_e;          // number of  momentum bins for e
  Float_t p_e;               // lowest momentum for e  
  Float_t fEMomentum1[100];  // momentum in bins (lowedge)   
  Float_t fEMomentum2[100];  // momentum in bins (centre)    
  Float_t fEMomentum3[100];  // momentum in bins (upperedge) 
  Float_t fEBandLo[100];     // lower PID band for protons 
  Float_t fEBandHi[100];     // higher PID band for protons 
  Float_t fEBandMea[100];    // parameters of the proton PID band profile
  Float_t fEBandRMS[100];    // parameters of the proton PID band profile 

   //data for pions to be used in reconstruction  program
  Float_t fPiBandLo_onehit[100];    // lower PID band for pions 
  Float_t fPiBandHi_onehit[100];    // higher PID band for pions
  Float_t fPiBandMe_onehit[100];    // parameters of the pion PID band profile 
  Float_t fPiBandRM_onehit[100];    // parameters of the pion PID band profile

  TPhDetector   *fDetector;  //!  Detector object for geometry access
  Bool_t          fVerbose;  //! Control debugging output

 public:

    ClassDef(TPhPIDSpect,2) // TPhPIDSpect class provides PID probabilities 

};

#endif
