      SUBROUTINE BANG(MODE, NTRGEN, NVERT)
*Documentation_Block
*****************************************************************************
*
*   DESCRIPTION:
*   Generates events with random particles, they may have different
*   momenta, eta or cos(theta), phi, id, vertex position.
*   NTRGEN  - Number of particles to be generated
*   NVERT   - Vertex number, if equal -1 new vertex is generated,
*             else previous vertex position is used
*   MODE = 1: The parameters describing particles are generated from flat
*             distributions (eta, P, phi)
*   MODE = 2: The emmision angle of the particles is selected to fit to
*             the spectrometer, other distributions are flat or fixed
*   MODE = 3: "Physical" distribution of particle momentum is used:
*             the distribution of tracks falls off exponentially
*             in transverse momentum as exp(-PT/REXP);
*             eta distribution has a trapezoidal shape.
*   MODE = 4,5:  temporarily undefined (for backward compatibility)
*   MODE = 6: flat distributions of: theta, 1/p, phi
*   MODE = 7: flat distributions of: eta, pt, phi
*   MODE = 8: flat distributions of: eta, 1/pt, phi
*   MODE = 9: not defined yet
*       In modes 1 and 2, the momentum may be fixed, distributed flat over
*   an interval, or in a series of spikes according to the choice of PLOW
*   and PHIGH (see details below).
*       The angle range in mode 2 is determined by RSPX,RSPZ and RSPR
*       the position and the radius of the circle describing spectrometer
*   The particle id can be constant or change randomly(see details below)
*
*   ARGUMENTS:   MODE, see above.
*
*   CALLED BY:   GUKINE, depending on flags set for various event generators
*
*   CALLS:   GSVERT, GSKINE
*
*   Implicit inputs, outputs:  Parameters come through common blocks:
*   Parameters for the event generator are passed in /RHBNG/ & /RHSYS/
*   as shown in include blocks BANGC.CMN and RHSYS.CMN
* Mnemonic: Variable in /RHBNG/:             Meaning:
*  RMOM      RMOM(3)       Range of flat momentum distribution; if PLOW
*    (PLOW, PHIGH, PSTEP)  is within 0.001 Gev/c of PHIGH, fixed momentum.
*                          If PSTEP < 0.0001 GeV/c flat distribution between
*                          PLOW and PHIGH is used
*                          If PSTEP > 0.0001 GeV/c the momentum is distributed
*                          in spikes at PLOW, PLOW+PSTEP, PLOW+2*PSTEP,....
*                          PSTEP is ignored in mode 3; REXP governs it instead.
*                          PSTEP is ignored in modes 6,7,8
*  RMUL      IRMUL         Multiplicity of each BANG event event.
*                          Used in GUKINE to calculate NTRGEN used in BANG call
*  RARM      IRARM         If IRARM = 2 the particles will be send in both
*                          spectrometer arms
*  RANG      RANG(3)       Angle range - for mode = 6
*                          the particles are generated with theta between
*                          RANG(1) and RANG(2)
*  RPHI      RPHI(2)       Azimuth angle range
*                          the flat distribution between RPHI(1) and RPHI(2)
*                          and (in case IRARM=2 only) between RPHI(1)+180.0 and
*                          RPHI(2)+180.0 is used
*  RROT      RROT          phi rotation (not used here)
*  RBXX      RBXX          Central position of the intersection of beams
*  RBYY      RBYY
*  RBZZ      RBZZ
*  RBXY      RBXY          Projected width of intersection region in x-y plane
*                          events are Gaussian distributed arround x=RBXX,
*                          y=RBYY in projection, and azimuthally symmetric.
*                          If RBXY > 0 then RBDZ defines triangular distribution
*                          of z arround RBZZ
*  RBDX      RBDX          Range of the vertex positions arround
*  RBDY      RBDY          RBXX, RBYY, RBZZ for flat distributions
*  RBDZ      RBDZ          (only in case RBXY=0)
*  RPID      IRPID(2)      Particle ID (Geant code) generation:
*                           - fixed (IRPID(1)=IRPID(2))
*                           - generated flat between IRPID(1) and IRPID(2)
*                           - follow "physical" distribution
*                             for IRPID(1)=IRPID(2)=0 :
*                             20:20:1:1:1:1  (pi+ : pi- : K+ : K- : P : anpi p)
*  REXP      REXP          Mean transv. momentum, mode 3
*                       definition of spectrometer position (in MODE = 2 only)
*  RSPX      RSPX          Center of the spectrometer X - position
*  RSPZ      RSPZ          Center of the spectrometer Z - position
*  RSPR      RSPR          Radius of the circle covering the spectrometer
*
*     Output is event setup initiated through  GSVERT, GSKINE
*
*  History:
*  DATE        NAME      DESCRIPTION
*  14-Jul-98   K.W.      Number of tracks to be generated NTRGEN given by
*                        by a parameter in the BANG call
*                        Vertex number given as a parameter, new vertex generated
*                        if NVERT < 0
*  11-FEB-98   K.W.      Correcred mode=8
*  17-Mar-97   K.W.      Forced STOP changed into a call to STOPER
*  20-SEP-96   K.W.      In MODE=2 new parameters independent of the magnetic
*                        field used: RSPX,RSPZ and RSPR (in place of MAGX,MAGZ)
*  16-Jul-96   K.W.      XOFF changed to MAGX
*  04-Apr-96   DMcL      RAN->RANF for OS independence
*  10-Jul-95   K.W.      New modes added
*  10-Apr-95   K.W.      Corrected error in the eta-theta transformation
*  10-Mar-95   K.W.      Changes in the modes of generation of the particles
*                        general changes in the code
*   1-Nov-93   R.V.      BANG.CMN -> BANGC.CMN to avoid MMS confusion
*  22-Nov-92   R.V.      Add pmc_inc: to includes; add UJKBUF.CMN; add space
*                          in GSKINE call for user buffer UBUF in JKINE
*  20-Nov-92   DMcL      Clean up, release
*  ~03/92      DMcL      Creation
****************************************************************************
*End_Documentation_Block

      IMPLICIT NONE

*KEEP,BANGC.
C Event generator parameters                 File: BANGC.CMN
C 20-Sep-96  K. Wozniak Added new parameters to control BANG mode = 2
C                       RSPX, RSPZ, RSPR (removed RMAG)
C 10-Mar-95  K. Wozniak Added new parameters to control track generation
c                       and selection (ETAMIN,ETAMAX, IDMIN,IDMAX ..)
c 22-Jun-94  R. V.      Remove NSEQN storage (see PH_HIT)
c 10-May-94  D. McLeod  Increased MXMUL to 13000 for HIJET events
c 17-Dec-92  R. Verdier Changed name from BANG to BANGC for CMS
c 22-Nov-92  D. Mcleod  Added EXF and NSEQN
      INTEGER MXMUL
      PARAMETER (MXMUL=13000)  ! maximum multiplicity for BANG etc.
      INTEGER IRMUL         !  multiplicity in random particle generation
      INTEGER IRARM         !  particles going to 1 or 2 arms selected
      INTEGER IRPID(2)      !  particle id range: min, max
      REAL RMOM(3)          !  momentum of particles:  min, max and step
                            !  (first step if 1/p is changed)
      REAL REXP             !  mean transv. mom. - for momentum generation
      REAL RETA(2)          !  ETA range: min, max
      REAL RANG(3)          !  starting theta angle and step
      REAL RPHI(2)          !  range of emission phi angle: min, max
      REAL RROT(2)          !  phi initial and additional rotation angle (GEN=0)
      REAL RBXX,RBYY,RBZZ   !  central position of the vertex
      REAL RBXY             !  width of the beam intersection in x-y
      REAL RBDX,RBDY,RBDZ   !  the range of positions of the vertex
C                           !
      REAL RSPX,RSPZ,RSPR   !  position and radius of the circle describing
C                           !  area covered by spectrometer
      REAL PHIR,CZ,ZC,RS    !  working variables for BANG
      COMMON/RHBNG/IRMUL, IRARM, IRPID, RMOM, REXP, RETA, RANG, RPHI,
     1             RROT,
     2             RBXX,RBYY,RBZZ, RBDX, RBDY, RBDZ, RBXY,
     3             RSPX,RSPZ,RSPR, PHIR,CZ,ZC,RS
*KEND.

*KEEP,FLAG.
C  PHOBOS flags, run control, operating modes, misc.   File FLAG.CMN
C   7-Jul-98  K. W.      Added  ITRIG flag (initial numer in TRIG command)
C  27-Nov-97  K. W.      Added IAOUF flag
C  31-Mar-95  R.V.       Change STRA to SECT to avoid GEANT 3.21 conflict
C  22-Nov-92  D. McLeod
C   These are set in input cards
      INTEGER ISASC,ISHBO,ISHDO,ISNTU, ISMEM,ISAOU,IAOUF, ISECT,IDRAW,
     1        IGEN,IANAL,ISCAT,IPIXL,ISEQ,NTRK,IZEB, ITRIG
      REAL WPIXL,WLOSS
      COMMON/RHFLG/ISASC,ISHBO,ISHDO,ISNTU,
     1  ISMEM,ISAOU,IAOUF, ISECT,IDRAW,
     2  IGEN,IANAL,ISCAT,IPIXL,ISEQ,NTRK,IZEB,WPIXL,WLOSS, ITRIG
*KEND.

*KEEP,RHSYS.
C  Common blocks for GPH PHOBOS simulation:  FILE RHSYS.CMN
C  General geometric layout of PHOBOS detector exclusive of detectors
C  These values are on input cards or data statements, except *MAX are
C  calculated to encompass all elements of the system.
C  20-Sep-96  K.W       Removed obsolete BMTBRAD, BERTHIK
C                       Removed FANGL - now local variable in BANG
C  12-Jul-96  K.W       Position of the magnet pole center MAGX, MAGZ
C                       (previously XOFF,ZOFF) separated from the magnet
C                       offset (now MOFF - 3-D vector)
C  14 Dec-92  R.V.      TMAG redefined and changed to integer
C  22-Nov-92 D. McLeod  Renamed from COMMON.SYS; TMAG, ZOFF, XOFF
C                       are added from the original GPH version
      REAL   BFLD         ! Magnetic field strength
     &      ,MAGX,MAGZ    ! Coordinates of field center
     &      ,MOFF(3)      ! offset of the field
     &      ,RFLD         ! Radius of field (perfect dipole only)
     1   ,XMAX,YMAX,ZMAX,ZMIN,XMMAX,ZMMAX
      INTEGER MAGT, TMAG  ! Field type: 0 Zebra bank, 1 = dipole,
                          ! 2 = ASCII map
      COMMON/RHSYS/ BFLD,MAGT,MAGX,MAGZ,MOFF
     1  ,RFLD,XMAX,YMAX,ZMAX,ZMIN,XMMAX,ZMMAX
      EQUIVALENCE (TMAG, MAGT)
*KEND.

*KEEP,UJKBUF.
C  PHOBOS JKINE user buffer definitions.          File: UJKBUF.INC
c 18-Jul-94  R. V.  Added NHitsAct
	integer Nwubuf, Nwubmax, NHitsAct
	parameter (Nwubmax = 11)
	common /UJKB/ Nwubuf, Ubuf(Nwubmax), NHitsAct
	real Ubuf
	integer IUBUF(Nwubmax)
	equivalence (IUBUF, UBUF)
c NHitsAct:   The number of hits on the track being propagated
c The user words at present are:
c UBUF(1):    Process in which particle was created (DCAY, BREM etc.)
c     (2):    Parent track number, negative is parent not stored
c             (possible in SECT=1 mode)
c     (3):    Volume number at origin, (4 characters in one integer word)
c             not definded for primaries and undefined for all tracks
c             in SECT=1 mode (integer value 0)
c     (4-6):  X, Y, Z at termination
c     (7-10): Px, Py, Pz, E at termination
c     (11):   Istop at track termination, positive when particle stopped
c             in a sensitive volume, negative otherwize
*KEND.

      INTEGER MODE, NTRGEN, NVERT

C Local:
      REAL VERT(3),PLAB(3),V,V1,V2,PP,PS,PHI
      REAL PLOW,PHIGH,PSTEP, CZ1,CZ2,Z, ETA, THETA
      REAL APLOW,APHIGH
      INTEGER ID,I,IR,NT,NSPK
      LOGICAL MOMFLG,SPKFLG
      SAVE IR     !random number seed
      DATA IR/12347/
      save VERT
      REAL RANF
      EXTERNAL RANF
*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

      PLOW = RMOM(1)
      PHIGH = RMOM(2)
      PSTEP = RMOM(3)
      IF(MODE .EQ. 6)   THEN
        APLOW = PLOW    !   APLOW,APHIGH - for  1/p generation
        IF(APLOW .LT. 0.01)  APLOW = 0.01
        APLOW = 1.0 / APLOW
        APHIGH = PHIGH
        IF(APHIGH .LT. 0.01)  APHIGH = 0.01
        APHIGH = 1.0 / APHIGH
      ENDIF
      IF(MODE .EQ. 8)   THEN
        APLOW = 0.01    !   APLOW,APHIGH - for  1/pt generation
        APLOW = 1.0 / APLOW
        APHIGH = PHIGH
        IF(APHIGH .LT. 0.01)  APHIGH = 0.01
        APHIGH = 1.0 / APHIGH
      ENDIF


************************************************************************
* Generate beam intersection production point and tell GEANT about it  *
************************************************************************
      IF(NVERT .LT. 0)  THEN
        IF(RBXY .LT. 0.0001)  THEN   ! flat distribution of vertex position
          IF(RBDX .GT. 0.0001)  THEN
             VERT(1) = RBXX + RBDX * (2.0*RANF(IR)-1.0)
          ELSE
             VERT(1) = RBXX
          ENDIF
          IF(RBDY .GT. 0.0001)  THEN
             VERT(2) = RBYY + RBDY * (2.0*RANF(IR)-1.0)
          ELSE
             VERT(2) = RBYY
          ENDIF   !  (RBDX .GT. 0.0001)
          IF(RBDZ .GT. 0.0001)  THEN
             VERT(3) = RBZZ + RBDZ * (2.0*RANF(IR)-1.0)
          ELSE
             VERT(3) = RBZZ
          ENDIF  !  (RBDZ .GT. 0.0001)

        ELSE    !  double Gaussian (x-y) and trianglular (z) distributions

    2     V1 = 2.0*RANF(IR) -1.0   !generate double Gaussian..
          V2 = 2.0*RANF(IR) -1.0   !see particle data booklet
          V = V1*V1 + V2*V2
          IF(V.GT.1.0) GOTO 2
          VERT(1) = RBXX + RBXY*V1*SQRT(-2.0*LOG(V)/V)
          VERT(2) = RBYY + RBXY*V2*SQRT(-2.0*LOG(V)/V)

C     Generate a triangular probability distribution for intersection z
          V1 = RANF(IR)
          V2 = RANF(IR)
          IF(V1.GT.V2) THEN
            VERT(3) = RBZZ + 0.5*RBDZ*(V1-1.0)
          ELSE
            VERT(3) = RBZZ + 0.5*RBDZ*V1
          ENDIF

        ENDIF   !   if(RBXY .GT. ...
      ENDIF   !  (NVERT .LT. 0)

      NVERT = -1   !  to force initialization of the vertex

************************************************************
*     Set up for various momentum distributions:
************************************************************
      MOMFLG = (ABS(PHIGH -PLOW).LT.0.001)
      SPKFLG = (PSTEP .GE. 0.0001)
      IF(MODE.EQ.3) THEN
        MOMFLG = .FALSE.
        SPKFLG = .FALSE.
      ENDIF
      IF(SPKFLG) NSPK = INT( (PHIGH-PLOW)/PSTEP ) + 1

***************************************************************
*     Set range (in cos theta) of track polar angles (MODE=2)
***************************************************************
      IF(MODE.EQ.2) THEN
         Z = VERT(3)
         ZC = RSPZ
         RS = SQRT((ZC-Z)**2 + RSPX**2)
         CZ = (ZC-Z)*SQRT(RS**2-RSPR**2)/RS**2
         CZ2 = RSPX * RSPR/RS**2
         CZ1 = CZ + CZ2
         IF(CZ1.GE.0.9999) CZ1 = 0.9999
         CZ2 = CZ - CZ2
      ENDIF

************************************************************************
*  START LOOP OVER MULTIPLICITY                                        *
************************************************************************

      DO I = 1,NTRGEN

********************************************
* Total momentum - MODE = 1, 2             *
********************************************
         IF(MOMFLG) THEN
            PS = PLOW
         ELSE IF(SPKFLG) THEN
            PS = PLOW + PSTEP*FLOAT( INT(FLOAT(NSPK)*RANF(IR)) )
            IF(PS.LT.0.010) PS = 0.010
         ELSE
            PS = PLOW + RANF(IR)*(PHIGH-PLOW)
         ENDIF

*********************************************************
*  Get random PLAB components:                          *
*    PS      = total momentum                           *
*    PLAB(3) = longitudinal component                   *
*    PLAB(1), PLAB(2) are calculated when phi is known  *
*********************************************************

         IF(MODE.EQ.1) THEN
           ETA = RETA(1) + RANF(IR)*(RETA(2)-RETA(1))
           PLAB(3) = PS*TANH(ETA)  !random pseudorapidity

         ELSE IF(MODE .EQ. 2)  THEN
           PLAB(3) = PS*(RANF(IR)*(CZ2-CZ1)+CZ1)   !random cosine theta

         ELSE IF(MODE .EQ. 3)   THEN

           IF(RETA(2)-RETA(1) .LT. 0.01)  THEN  ! fixed eta
              ETA = RETA(1)
           ELSE
             ETA = RETA(1) - 100.0
             DO  WHILE (ETA .LT. RETA(1) .OR. ETA .GT.RETA(2))
                IF(RANF(IR) .GE. 0.5)  THEN  !  flat central region
                   ETA = 4.0 * RANF(IR) - 2.0
                ELSE   !  triangular parts -6,-2  and  2,6
                   V1 = RANF(IR)
                   V2 = RANF(IR)
                   IF(V1 .GT. V2)  THEN   !  negative eta
                      ETA = -6.0 + 4.0 * V1
                   ELSE                   !  positive eta
                      ETA = 2.0 + 4.0 * V1
                   ENDIF
                ENDIF
              ENDDO
           ENDIF

           IF(PHIGH-PLOW .LT. 0.010)  THEN   !  fixed momentum
              PS = PLOW
           ELSE
C   Adding two values to make PP is supposed to distribute it as
C   PP*EXP(-PP/REXP) i.e. exponential falloff x phase space
              PS = PLOW - 1.0
              DO  WHILE (PS .LT. PLOW  .OR.  PS .GT. PHIGH)
                PP = -LOG(RANF(IR))*REXP
                PP = PP -LOG(RANF(IR))*REXP
                PP = PP / 2.0
                PS = PP / SQRT(1.0-TANH(ETA)**2)
                PLAB(3) = PS*TANH(ETA)  !random pseudorapidity
              ENDDO
           ENDIF

C---------------------         END  MODE = 3

         ELSE IF(MODE .EQ. 4  .OR.  MODE .EQ. 5)  THEN

            PRINT *, 'BANG:  GEN = 4  and  GEN = 5  have changed to',
     +               ' 11 and 0 respectively'
            CALL STOPER('BANG', 'ILLEGAL GEN MODE', 1)


         ELSE IF(MODE .EQ. 6)  THEN
            PS = 1.0 / ( APLOW + (APHIGH-APLOW)*RANF(IR) )
            THETA = RANG(1) + RANF(IR)*(RANG(2)-RANG(1))
            PLAB(3) = PS*COS(THETA/57.295779513)
C---------------------         END  MODE = 6

         ELSE IF(MODE .EQ. 7)   THEN
           ETA = RETA(1) + RANF(IR)*(RETA(2)-RETA(1))
           PS = PLOW - 1.0
           DO  WHILE (PS .LT. PLOW  .OR.  PS .GT. PHIGH)
                PP = RANF(IR) * PHIGH   !  flat Pt from 0.0 to PHIGH
                PS = PP / SQRT(1.0-TANH(ETA)**2)
                PLAB(3) = PS*TANH(ETA)  !random pseudorapidity
           ENDDO
C---------------------          END  MODE = 7

         ELSE IF(MODE .EQ. 8)   THEN
           ETA = RETA(1) + RANF(IR)*(RETA(2)-RETA(1))
           PS = PLOW - 1.0
           DO  WHILE (PS .LT. PLOW  .OR.  PS .GT. PHIGH)
              PP = 1.0 / ( APHIGH + RANF(IR)*(APLOW-APHIGH) )
C                                  !  1/Pt  from 0.01 to PHIGH
              PS = PP / SQRT(1.0-TANH(ETA)**2)
              PLAB(3) = PS*TANH(ETA)  !random pseudorapidity
           ENDDO
C---------------------          END  MODE = 8

         ELSE
           PRINT *, 'BANG: undefined GEN: ', MODE
           CALL STOPER('BANG', 'UNDEFINED GEN MODE', 2)
         ENDIF         !      all  MODE  values


C  Generate the value of phi angle
         IF(RPHI(2)-RPHI(1) .LT. 0.01)  THEN  !  fixed phi
            PHI = RPHI(1) / 57.295779513
         ELSE
            PHI = RPHI(1) + (RPHI(2)-RPHI(1)) * RANF(IR)
            PHI = PHI / 57.295779513
         ENDIF
         IF(IRARM .GT. 1)  THEN   !  possible change of the arm
            IF(RANF(IR) .GT. 0.5)  THEN
               PHI = PHI + 3.14159265359
               IF(PHI .GT. 3.14159265359)  PHI = PHI - 6.28318530718
            ENDIF
         ENDIF
         PP = SQRT(PS*PS-PLAB(3)*PLAB(3))
         PLAB(1) = PP*COS(PHI)
         PLAB(2) = PP*SIN(PHI)

*****************************************
*     Generate the particle id          *
*****************************************
         IF(IRPID(1) .EQ. IRPID(2))  THEN
            ID = IRPID(1)
            IF(ID .EQ. 0)  THEN   !  select pi, K, p with ratios 20:1:1
              V = RANF(IR)
              IF(V .GE. 0.55)  ID = 8
              IF(V .GE. 0.1    .AND.  V .LT. 0.55 )  ID = 9
              IF(V .GE. 0.075  .AND.  V .LT. 0.10 )  ID = 11
              IF(V .GE. 0.05   .AND.  V .LT. 0.075)  ID = 12
              IF(V .GE. 0.025  .AND.  V .LT. 0.05 )  ID = 14
              IF(V .LT. 0.025)  ID = 15
            ENDIF
         ELSE
            ID = IRPID(1) +
     +             INT(FLOAT(IRPID(2)-IRPID(1)+1)*RANF(IR)*0.99999)
         ENDIF

         IF(NVERT .LT. 0)  CALL GSVERT(VERT,0,0,0,0,NVERT)

         CALL UZERO(UBUF(1), 1, NWUBMAX)
         CALL GSKINE(PLAB,ID,NVERT,UBUF,NWUBMAX,NT)

      ENDDO
CCCC      NTRK = NTRGEN

      RETURN
      END
