      subroutine UGCIN (ludat, lurep, lstor, lline, line, ires)
*Documentation_Block
************************************************************************
*
*  Description:
*  Geometry and control information text file reader for Phobos
*  UGCIN reads the data file on logical unit "UGCDATA", parses
*  the lines, and calls UGCDEF to pass the specified values of the
*  variables to the geometry- and  reconstruction-defining subroutines,
*  or UGCSET to pass values to user commons, banks and variables.
*
*  Arguments:
*  ires*:  integer value returned to caller, <>0 if error(s) occurred
*  line*:  storage for text lines and stored definitions
*  *lline: the length of the line storage
*  *lstor: the total length reserved for line plus stored definitions
*  *ludat: Fortran logical input unit for reading GCDEFS:
*          if > 0, already open; if 0, quit; if < 0, open abs(ludat)
*  *lurep: Fortran logical unit for reporting errors
*
*  Called by:
*  UGINIT
*
*  Calls:
*  SERVER to open geometry text file gcdefs
*  UGCDEF to dispatch variables to geometry subroutine
*  UGCSET to set values in commons and banks
*  UGCGET to extract the next token from a character string
*  UGCVAL to classify and evaluate a token by content
*
*  Implicit inputs, outputs, side effects:
*  none
*
*  History:
*  Date       Name       Description
*  ----       ----       -----------
*  12-Dec-97  K.W.       Stripping of apostrophes moved from UGCGET to
*                        UGCIN, after a call to UGCVAL.
*                        previously volume names looking like numbers (1E00)
*                        could not be used even if surrounded by apostrophes
*  05-Nov-97  K.W.       Initialization of numeric (integer and real)
*                        command parameters to 0
*  05-Aug-96  R. V.      ")," -> ")" in writes for ANSI compliance;
*                        open -> call UFILE to handle IBM-specific open
*  05-Apr-96  Don McLeod Open GCDEFS via SERVER to interpret logical
*  22-Nov-93  R. V.      -maxargs -> mmaxargs, idebug -> ldebug for
*                        non-VMS Fortrans
*  24-Apr-93  R. V.      Implement vector handling: pass Ielem,
*                        ArgTypes and ArgLengths to UGCSET; get
*                        Index from UGCVAL.
*  02-Nov-92  R. Verdier Creation
************************************************************************
*End_Documentation_Block

      implicit none

      character*(*) line
      integer ires, lline, lstor, ludat, lurep

c Declare common with IDRUN for SERVER

*KEEP,GCFLAG.
      COMMON/GCFLAG/IDEBUG,IDEMIN,IDEMAX,ITEST,IDRUN,IDEVT,IEORUN
     +        ,IEOTRI,IEVENT,ISWIT(10),IFINIT(20),NEVENT,NRNDM(2)
      COMMON/GCFLAX/BATCH, NOLOG
      LOGICAL BATCH, NOLOG
C
      INTEGER       IDEBUG,IDEMIN,IDEMAX,ITEST,IDRUN,IDEVT,IEORUN
     +        ,IEOTRI,IEVENT,ISWIT,IFINIT,NEVENT,NRNDM
C
*KEND.

c Declare local variables:

c Debug control: 0 for no output, 1 for fatal errors only, ...
      integer ldebug /1/

c Keys and data for validation of input:

      integer maxargs, mmaxargs, nkeys, nsubkeys
      parameter (nkeys = 2, nsubkeys = 3, maxargs = 20, mmaxargs = -20)
c Note when changing keys, also check the lengths lkeys.
      character*8 keys(nkeys) /'GS', 'GC' /
      integer lkeys(nkeys) / 2, 2 /
      character*4 subkeys(nsubkeys) /'SET', 'DEF', 'END' /
      integer iarg, iiv, ikey, ils, irv, isubkey, kk
      integer ArgLengths(maxargs) /maxargs*0/,
     *        ArgTypes(maxargs)   /maxargs*0/,
     *        ivalues(maxargs)    /maxargs*0/,
     *        lsargs(maxargs)     /maxargs*0/,
     *        ssargs(maxargs)     /maxargs*0/
      real    rvalues(maxargs)    /maxargs*0/

c Define the number and types of arguments expected for each kind
c of input.
c Note that the number of tokens for UGC lines is two greater than
c the number of arguments seen by UGCDEF, since the input line
c includes the Id and the the name of the subroutine

      integer nargs(nkeys), vtypes(maxargs, nkeys)

      data nargs / mmaxargs, mmaxargs /

c Following must understand any variable type:
      data (vtypes (kk, 1), kk=1,1) / -1   /
      data (vtypes (kk, 2), kk=1,1) / -1   /

      integer ielem, index, ios, itoken, ivtype, ivalue, lent,
     *  lnext, loct, lui
      real rvalue
      logical eof, eol
      character*1 term /'!'/

c ArgLengths: the lengths in storage of the arguments, in bytes
c ArgTypes: the types of the arguments: 1=integer, 2=real, 3=string
c ielem: the vector index found in the current variable name
c index: ielem returned by UGCVAL for the current token
c ivalue: the value of the current token, if integer
c ivtype: the ArgType of the current token
c lent:   the length of the current token
c loct:   the index (1..lline) of the current token
c lnext:  the index of the whitespace preceding the next token
c rvalue: the value of the current token, if real
c term:   the line termination character

c For long form SERVER returns:
      CHARACTER*80 FILENAME

c End of declarations

      ires = 0

c Open dataset only if not already done by caller
      lui = iabs(ludat)
      if (ludat .lt. 0) then
         call UFILE (0, 'GCDEFS', lui, FileName)
      endif
c Bypass reading if ludat is 0
      eof = lui .eq. 0

c Loop over input lines until end of file is reached

      do while (.not. eof)

        read (lui, '(a)', iostat=ios) line
        eof = ios .ne. 0
        itoken = 0
        if (ldebug .ge. 2) write (lurep, '(a,i3/1x,a)')
     *    ' UGCIN line read, length = ', lline, line(1:lline)
        loct = 1
        if (eof) loct = lline + 1
        do  irv=1,maxargs  ! initialize parameters to 0
           rvalues(irv) = 0.0
           ivalues(irv) = 0
        enddo
        ikey = 0
        isubkey = 0
        iarg = 0
        irv = 0
        iiv = 0
        ils = 0
        eol = loct .gt. lline .or. line(loct:loct) .eq. term

c Unless read failed, loop extracting tokens from the current line

        do while (.not. eol)

c Set loct and lent to the location and length of the next token
          call UGCGET (line, lline, itoken, loct, lent, lnext)
          eol = loct .gt. lline .or. line(loct:loct) .eq. term
          if (.not. eol) then
            if (ldebug .ge. 4) write (lurep, '(3(a,i3),2a)')
     *        ' UGCGET returned loct = ', loct, ', lent = ', lent,
     *        ', token # ', itoken, ' = ', line(loct:loct+lent-1)

c Determine the token type by its content
            call UGCVAL (line(loct:loct+lent-1), lent, ivtype,
     *        rvalue, ivalue, index)

C Store the token's type and value, or its location if a string
            if (iarg .lt. maxargs) iarg = iarg + 1
            ArgTypes(iarg) = ivtype
            if (ivtype .eq. 1) then
              iiv = iiv + 1
              ivalues(iiv) = ivalue
              ArgLengths(iarg) = 4
            elseif (ivtype .eq. 2) then
              irv = irv + 1
              rvalues(irv) = rvalue
              ArgLengths(iarg) = 4
            elseif (ivtype .eq. 3) then
              ils = ils + 1
              if(line(loct:loct) .ne.  '''')  then
                 lsargs(ils) = loct
                 ssargs(ils) = lent
              else   !  stripping of apostrophes
                 lsargs(ils) = loct + 1
                 ssargs(ils) = lent - 2
                 if(line(loct:loct) .ne.  '''')
     +                             ssargs(ils) = lent - 1
              endif
              ArgLengths(iarg) = lent
            endif
c If this is the first token, compare it with the list of keywords.
c The match is case-sensitive and the exact length must be matched.
c Ikey will be the index of the found keyword; <=0 if none matched.
            if (itoken .eq. 1) then
              kk = 1
              ikey = 0
              do while (ikey .eq. 0 .and. kk .le. nkeys)
                if (lent .eq. lkeys(kk) .and.
     *            line(loct:loct+lent-1) .eq. keys(kk)(1:lkeys(kk)))
     *            ikey = kk
                kk = kk + 1
              enddo
              if (ikey .gt. 0 .and. ldebug .ge. 3)
     *          write (lurep, '(3a,i3)') ' UGCIN found ',
     *          line(loct:loct+lent-1), ' = key #', ikey

            elseif (itoken .le. 3 .and. ivtype .eq. 3
     *              .and. ikey .gt. 0) then
              if (itoken .eq. 2) then
                kk = 1
                isubkey = 0
                do while (isubkey .eq. 0 .and. kk .le. nsubkeys)
                  if (line(loct:loct+lent-1) .eq. subkeys(kk))
     *              isubkey = kk
                  kk = kk + 1
                enddo
                if (isubkey .gt. 0) then
                  if (ldebug .ge. 3)
     *              write (lurep, '(3a,i3)') ' UGCIN found ',
     *                line(loct:loct+lent-1), ' = subkey #', isubkey
                endif
              elseif (isubkey .gt. 0) then
c If third token and subkey found, save vector index
                ielem = index
              endif
            endif
c Update the starting location for the next token
            loct = lnext

          endif

c Stop parsing if end of line or if first token was unrecognized
          eol = eol .or. (itoken .eq. 1 .and. ikey .le. 0)

        enddo ! while .not. eol

c Verify that the number of tokens found is correct

        if (ikey .gt. 0) then
          if (nargs(ikey) .gt. 0 .and. iarg .ne. nargs(ikey)) then
            if (ldebug .gt. 0) write (lurep, '(3(a,i3),a)')
     *        ' UGCIN:', iarg, ' arguments found for keyword ',
     *        ikey, '; ', nargs(ikey), ' expected.'
            ikey = -1
          endif
        endif

c Call the dispatch subroutine

        if (ikey .gt. 0) then
          if (isubkey .eq. 1) then
            call UGCSET (line(lsargs(3):lsargs(3)+ssargs(3)-1),
     *        ielem, iarg-3, ArgTypes(4), ArgLengths(4), rvalues,
     *        ivalues, line(lsargs(4):lsargs(4)+ssargs(4)-1), ires)
c Note the string argument of UGCSET is a single value for now.
c Map ires into UGCIN definitions:
            if (ires .gt. 0) then
              ires = 0
            elseif (ires .lt. 0) then
              if (ldebug .gt. 0) write (lurep, '(a,i4,2a)')
     *          ' UGCIN: UGCSET returned error code', ires,
     *          ' while processing values of ',
     *          line(lsargs(3):lsargs(3)+ssargs(3)-1)
            endif

          elseif (isubkey .eq. 2) then
c DEF (local definition): not yet implemented

          elseif (isubkey .eq. 3) then
c END: quit reading, return the argument
            ires = ivalues(1)
            eof = .true.

          else
            call UGCDEF (line(lsargs(2):lsargs(2)+ssargs(2)-1),
     *        iarg-2, ArgTypes(3),
     *        ivalues, rvalues, lsargs(3), ssargs(3), line, ires)

          endif ! isubkey
        endif ! ikey

      enddo ! while not eof

c Close dataset only if opened here
      if (ludat .lt. 0) close (lui)

      return
      end
