#ifndef _PhMCTrackTemplate_INCLUDED_
#define _PhMCTrackTemplate_INCLUDED_

//////////////////////////////////////////////////////////////////////////
//                                                                      //
//    TPhMCTrackTemplate                                               //
//                                                                      //
//    Phobos track template class                                       //
//                                                                      //
//    TPhMCTrackTemplate contains a template track used in the PHOBOS  //
//    track finding                                                     // 
//                                                                      //
//    Author  : Krzysztof Wozniak                                       //
//              (based on TPhTrack Template created by Gunther Roland)  //
//    Created : 01/12/98                                                //
//    Version : 0.0  (before first release !)                           //
//    Changed :                                                         //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

// Root Classes

#include "TObjArray.h"

// Phobos Classes


class TPhHit;
class TPhMCHit;
class TPhTrack;
class TPhMCTrack;
class TPhDetector;

#include "TPhMomentumVector.h"

#include "../../matrix++/matrix.h"


const Int_t   kNMaxRows  = 17;

class TPhMCTrackTemplate : public TPhMomentumVector {

public:
    TPhMCTrackTemplate();
    TPhMCTrackTemplate(Int_t NLayers);
    virtual ~TPhMCTrackTemplate();

    virtual void     CalcChi2(TPhTrack *track, Float_t *chi2, Float_t *ndof);
    virtual void     CalcChi2(TPhTrack *track, Float_t *chi2, Float_t *ndof, Bool_t UseCovar);

    virtual void    FillTrack(TPhMCTrack *track);
    virtual void    AddVector(Double_t *vec);
    virtual void    AddTrack(TPhMCTrack *track, Bool_t AddToCovar = kFALSE);
    virtual void    CalcChi2(Double_t *d, Float_t *chi2, Float_t *ndof);
    virtual void    NormEntries();
    virtual void    NormCovar();
    virtual void    AddPadWidth(Float_t factor = 0.2887);
    virtual void    MakeStamp();

    virtual void    SetMass(Double_t x)         { fMass          = x; }
    virtual void    SetMinField(Double_t x)     { fMinField      = x; }
    virtual void    SetVerbose(Bool_t x)        { fVerbose       = x; }
    virtual void    SetMaxCompDist(Double_t x)  { fMaxCompDist   = x; }
    virtual void    SetMinCompEnergy(Double_t x){ fMinCompEnergy = x; }
    virtual void    SetCovar(matrix *m)         { fCovar         = m; }
    virtual void    SetNRows(Int_t n)           { fNRows         = n; }

    virtual Double_t GetMeanDist(Int_t layer);
    virtual Double_t GetMeanDe(Int_t layer);
    virtual Double_t GetMeanDLength(Int_t layer);
    virtual Double_t GetMeanLength(Int_t layer);
    virtual Double_t GetMeanR(Int_t layer);
    virtual Double_t GetMeanX0(Int_t layer);
    virtual Double_t GetMeanZ0(Int_t layer);
    virtual Float_t  GetPadHeight(Int_t layer);
    virtual Float_t  GetPadWidth(Int_t layer);
    virtual Int_t    GetEntries(Int_t layer);

    virtual Bool_t   CloneTemplate(TPhMCTrackTemplate *tmpl);
    virtual Bool_t   CompatibleTemplate(TPhMCTrackTemplate *tmpl);
    virtual Bool_t   CompatibleTrack(TPhTrack *track);
 
    virtual Int_t    GetRow(Int_t layer);
    virtual void     InvertCovar(); 
    virtual Bool_t   ReadCovar(Char_t *FileName); 
    virtual Bool_t   WriteCovar(Char_t *FileName); 
    virtual void     PrintCovar();
    virtual matrix  *GetCovar()                  { return fCovar;         }

    virtual Int_t     GetStamp()                 { return fStamp;         }
    virtual Bool_t    GetUseMeanDist()           { return fUseMeanDist;   }
    virtual void      SetUseMeanDist(Bool_t usemeandist = kTRUE);
    virtual Bool_t    GetAddPadWidthFlag()       { return fAddPadWidthFlag;  }
    virtual Bool_t    GetNormEntriesFlag()       { return fNormEntriesFlag;  }
    virtual Bool_t    GetNormCovarFlag()         { return fNormCovarFlag;    } 
    virtual Bool_t    GetInvertFlag()            { return fInvertFlag;       } 
    virtual Float_t   GetPadWidthFactor()        { return fPadWidthFactor;}
    virtual Double_t  GetMass()                  { return fMass;          }
    virtual Double_t  GetMinField()              { return fMinField;      }
    virtual Int_t     GetNRows()                 { return fNRows;         }
    virtual Int_t     GetNLayers()               { return fNLayers;       }
    virtual Int_t     GetLayerToRow(Int_t i)     { return fLayerToRow[i]; }
    virtual Double_t  GetMeanDistXZ(Int_t plane);
    virtual Float_t   GetMeanDist45(Int_t plane);
    virtual Float_t   GetMinDist45(Int_t plane);
    virtual Float_t   GetMaxDist45(Int_t plane);
    virtual Float_t   GetMeanNPads(Int_t plane);

    virtual Double_t  GetMaxCompDist()           { return fMaxCompDist;   }
    virtual Double_t  GetMinCompEnergy()         { return fMinCompEnergy; }
    virtual Double_t  GetDistXZ(TPhHit *hit);
    virtual Double_t  GetDistXZ(TPhHit *hit,Int_t row);
    virtual Double_t  GetDistLast(Float_t x, Float_t z, Int_t layer);
    virtual void      GetLast(Float_t *x, Float_t *z, Int_t *layer);
    virtual Bool_t    IsCompatible(TPhHit *hit);
    virtual Bool_t    IsCompatible(TPhHit *hit, Double_t slopey, Int_t slopeylayer);

    virtual void      PrintParameters(Int_t plane);
  
private:

    Int_t     fLabel;                 // Template label
    Int_t     fStamp;                 //!Identifies template layers
    Bool_t    fUseMeanDist;           // Flag for covariance matrix normalization type
    Bool_t    fAddPadWidthFlag;       // Indicator of added pad width to cov. matrix
    Float_t   fPadWidthFactor;        // factor multiplying deviation due to pad width
    Bool_t    fNormEntriesFlag;       // Indicator of distance normalization
    Bool_t    fNormCovarFlag;         // Indicator of cov. matrix normalization
    Bool_t    fInvertFlag;            // Status of cov. matrix invertion
    Double_t  fMinField;              // Min field for using circle approximation
    Double_t  fMass;                  // Mass for template track
    Int_t     fNLayers;               // Max number of planes for a valid track
    Int_t     fNRows;                 // Dimension of error vector
    Int_t     fEntries[kNMaxRows];    // Entries per plane, for normalization  
    Int_t     fLayerToRow[kNMaxRows]; // Translation from detector layer to covar matrix row
    Bool_t    fVerbose;               //!Controls debugging output during filling
    Double_t  fMinCompEnergy;         //!Min. energy for hits compatible with template
    Double_t  fMaxCompDist;           //!Max. distance for hits compatible with template
    Double_t  fLastX;                 // X of hit in last layer
    Double_t  fLastZ;                 // Z of hit in last layer
    Int_t     fLastLayer;             // Number of the last layer 
    Double_t  fMeanDe[kNMaxRows];     // Mean De in a plane for a valid track
    Double_t  fMeanDLength[kNMaxRows];// Mean length of the track inside sensor
    Double_t  fMeanLength[kNMaxRows]; // Mean length of trajectory to plane
    Float_t   fMeanDist45[kNMaxRows]; // dist45 mean value
    Float_t   fMinDist45[kNMaxRows];  // minimal value of dist45 
    Float_t   fMaxDist45[kNMaxRows];  // maximal value of dist45
    Float_t   fMeanNPads[kNMaxRows];  // mean number of pads hit in the plane
    Double_t  fMeanDist[kNMaxRows];   // Spread of hit distance to template 
    Double_t  fMeanR[kNMaxRows];      // Mean radius of track at plane
    Double_t  fMeanX0[kNMaxRows];     // Mean X center of circle approximation to trajectory
    Double_t  fMeanZ0[kNMaxRows];     // Mean Z center of circle approximation to trajectory
    Float_t   fPadHeight[kNMaxRows];  // height of the pad in the sensor
    Float_t   fPadWidth[kNMaxRows];   // width of the pad in the sensor
    matrix   *fCovar;                 // Inverted covariance matrix 
    Double_t *fCovarEntries;          //! Entries for covariance matrix
    TPhDetector *fDetector;           //! Detector object 

public:
    ClassDef(TPhMCTrackTemplate,1) // PHOBOS track template class

};

#endif
