#ifndef _PhTrackTemplate_INCLUDED_
#define _PhTrackTemplate_INCLUDED_

///////////////////////////////////////////////////////////////////////
//                                                                   //
//    TPhTrackTemplate                                               //
//                                                                   //
//    Phobos track template class                                    //
//                                                                   //
//    TPhTrackTemplate contains a template track used in the PHOBOS  //
//    track finding                                                  // 
//                                                                   //
//    Author  : Gunther Roland (rolandg@vxcern.cern.ch)              //
//    Created : 01/09//96                                            //
//    Version : 1.0                                                  //
//    Changed :                                                      //
//                                                                   //
///////////////////////////////////////////////////////////////////////

// Root Classes

#include "TObjArray.h"

// Phobos Classes


class TPhHit;
class TPhTrack;
class TPhMCTrack;

#include "TPhMomentumVector.h"

#include "../../matrix++/matrix.h"


const Int_t   kMaxRows  = 16;

class TPhTrackTemplate : public TPhMomentumVector {

public:
    TPhTrackTemplate();
    TPhTrackTemplate(Int_t NLayers);
    virtual ~TPhTrackTemplate();

    virtual void     CalcChi2(TPhTrack *track, Float_t *chi2, Float_t *ndof);
    virtual void     CalcChi2(TPhTrack *track, Float_t *chi2, Float_t *ndof, Bool_t UseCovar);

    virtual void    FillTrack(TPhMCTrack *track, Bool_t HasMS);
    virtual void    AddVector(Double_t *vec);
    virtual void    AddTrack(TPhMCTrack *track, Bool_t AddToCovar = kFALSE);
    virtual void    CalcChi2(Double_t *vec, Float_t *chi2, Float_t *ndof);
    virtual void    NormEntries(Bool_t UseMeanDist = kFALSE);
    virtual void    NormCovar();

    virtual void    MakeStamp();

    virtual void    SetMass(Double_t x)         { fMass          = x; }
    virtual void    SetMinField(Double_t x)     { fMinField      = x; }
    virtual void    SetVerbose(Bool_t x)        { fVerbose       = x; }
    virtual void    SetMaxCompDist(Double_t x)  { fMaxCompDist   = x; }
    virtual void    SetMinCompEnergy(Double_t x){ fMinCompEnergy = x; }
    virtual void    SetCovar(matrix *m)         { fCovar         = m; }
    virtual void    SetNRows(Int_t n)           { fNRows         = n; }

    virtual Double_t GetMeanDist(Int_t layer);
    virtual Double_t GetMeanDedx(Int_t layer);
    virtual Double_t GetMeanLength(Int_t layer);
    virtual Double_t GetMeanR(Int_t layer);
    virtual Double_t GetMeanX0(Int_t layer);
    virtual Double_t GetMeanZ0(Int_t layer);
    virtual Double_t GetMeanY(Int_t layer);

    virtual Bool_t   CloneTemplate(TPhTrackTemplate *tmpl);
    virtual Bool_t   CompatibleTemplate(TPhTrackTemplate *tmpl);
    virtual Bool_t   CompatibleTrack(TPhTrack *track);
    virtual Bool_t   NeighbourPoint(Int_t i, Int_t j, Int_t d, Int_t n, Int_t *i1, Int_t *j1);
 
    virtual void     InvertCovar(); 
    virtual Bool_t   ReadCovar(Char_t *FileName); 
    virtual Bool_t   WriteCovar(Char_t *FileName); 
    virtual void     PrintCovar();
    virtual matrix  *GetCovar()                  { return fCovar;     }

    virtual Int_t     GetStamp()                 { return fStamp;     }
    virtual Double_t  GetMass()                  { return fMass;      }
    virtual Double_t  GetMinField()              { return fMinField;  }
    virtual Int_t     GetNRows()                 { return fNRows;     }
    virtual Int_t     GetNLayers()               { return fNLayers;   }
    virtual Int_t     GetLayerToRow(Int_t i)     { return fLayerToRow[i]; }
    virtual Double_t  GetMeanDistXZ(Int_t plane);

    virtual Double_t  GetMaxCompDist()           { return fMaxCompDist;   }
    virtual Double_t  GetMinCompEnergy()         { return fMinCompEnergy; }
    virtual Double_t  GetDistXZ(TPhHit *hit);
    virtual Double_t  GetDistXZ(TPhHit *hit,Int_t row);
    virtual Double_t  GetDistY(TPhHit *hit);
    virtual Bool_t   IsCompatible(TPhHit *hit);

    virtual void     PrintParameters(Int_t plane);
  
private:

    Int_t     fStamp;                //!Identifies template layers
    Double_t  fMinField;             // Min field for using circle approximation
    Double_t  fMass;                 // Mass for template track
    Int_t     fNLayers;              // Max number of planes for a valid track
    Int_t     fNRows;                // Dimension of error vector
    Int_t     fEntries[kMaxRows];    // Entries per plane, for normalization  
    Int_t     fLayerToRow[kMaxRows]; // Translation from detector layer to covar matrix row
    Bool_t    fVerbose;              //!Controls debugging output during filling
    Double_t  fMinCompEnergy;        //!Min. energy for hits compatible with template
    Double_t  fMaxCompDist;          //!Max. distance for hits compatible with template
    Double_t  fMeanDedx[kMaxRows];   // Mean dedx per plane for a valid track
    Double_t  fMeanLength[kMaxRows]; // Mean length of trajectory to plane
    Double_t  fMeanDist[kMaxRows];   // Spread of hit distance to template 
    Double_t  fMeanR[kMaxRows];      // Mean radius of track at plane
    Double_t  fMeanX0[kMaxRows];     // Mean X center of circle approximation to trajectory
    Double_t  fMeanZ0[kMaxRows];     // Mean Z center of circle approximation to trajectory
    Double_t  fMeanY[kMaxRows];      // Mean vertical Y of track at plane
    matrix   *fCovar;                // Inverted covariance matrix                  
    Double_t *fCovarEntries;         //! Entries for covariance matrix                  

public:
    ClassDef(TPhTrackTemplate,1) // PHOBOS track template class

};

#endif


