//
//
// TPhTrackFinder
//
//
//
//
// $Id: TPhTrackFinder.cxx,v 1.10 1999/03/16 16:20:00 icpark Exp $

#include "math.h"

#include "TFile.h"
#include "TKey.h"
#include "TRandom.h"

#include "TPhTrackTemplate.h"
#include "TPhDetector.h"
#include "TPhSubDetector.h"
#include "TPhSiSensor.h"
#include "TPhLayer.h"
#include "TPhHit.h"
#include "TPhRecHit.h"
#include "TPhTrack.h"
#include "TPhRecTrack.h"
#include "TPhEvent.h"
#include "TPhStuff.h"
#include "TPhObjectContainer.h"
#include "TPhMulguisin.h"
#include "TPhTrackSeedFinder.h"
#include "TPhHough.h"
#include "TPhHoughHit.h"
#include "TPhHoughRoad.h"
#include "TPhMerging.h"
#include "TPhTrackFinder.h"

ClassImp(TPhTrackFinder)

TPhTrackFinder::TPhTrackFinder()
{
  fVerbose	  = 0;
  fOuterLayer     = 8;
  fMaxCompDist    = 2.0;
  fMinCompEnergy  = 0.5;
  fMaxSharedHits  = 2;
  fMinHits        = 11;
  fMinProb        = 0.01;
  fMerge          = new TPhMerging();
  fMerge->SetMergingType(1);
  fMerge->SetMaxMergedPads(3);
  fStuff          = new TPhStuff();
  fDetector       = TPhDetector::GetInstance();
  fTemplateList   = NULL; 
  fTemplateList   = new TObjArray();
  fCandidateList  = new TObjArray();
  fTrackList      = new TObjArray();
  if (!fDetector) fprintf(stderr,"TPhTrackFinder: Couldn't get detector object\n");
  //ReadHoughTable();         
  for (int l=0; l<kNLayer; l++) fHoughHitA[l]=new TObjArray();
  for (int i=0; i<kNPairs; i++)
  {
    fHoughRoadA[i] = new TObjArray();
    //fHoughRoadA[i]=NULL;
  }
  fCurveList      = new TObjArray();
  for (int il=0; il<16; il++) fTrack[il]=new TObjArray();
  fSeedList = NULL;
}

TPhTrackFinder::~TPhTrackFinder()
{
  if (fVerbose) printf("---- ~TPhTrackFinder ----\n");
  delete fStuff;
  delete fDetector;
  delete fTemplateList;
  delete fCandidateList;
  delete fTrackList;
}

Bool_t TPhTrackFinder::ReadTemplateList(Char_t *TemplName)
{
  
  //
  // open a template file
  //
  
  TFile f(TemplName);
  if (!f.IsOpen()) 
  { 
    fprintf(stderr,"Can't open file %s\n",TemplName);
    return kFALSE;
  } 
  
  //
  //
  //
  
  TList *l = f.GetListOfKeys();
  TIter NextKey(l);
  if (fVerbose) printf("\nFound a total of %d templates",l->GetSize());
  TKey *key;
  while ( (key=(TKey *)NextKey()) )
  {
    TPhTrackTemplate *t = (TPhTrackTemplate *) key->ReadObj();
    t->SetMaxCompDist(fMaxCompDist);
    t->SetMinCompEnergy(fMinCompEnergy);
    t->MakeStamp();
    fTemplateList->Add(t);
  }
  return kTRUE;
}

Int_t TPhTrackFinder::GetCompatibleHits(TPhTrackTemplate *t)
{
  //  
  // initialization
  //
  
  for (int i=0; i<kNLayer; i++) 
  {
    fHitCnt[i]=0;
    for (int j=0; j<kNumHits; j++)
    {
      fHitArray[i][j]=NULL;
    }
  }
    
  //
  // looping on hits
  //
				    
  Int_t arm,cnt,n_missed=0,flag,nhits,label;
  TPhLayer  *layer; 
  TPhSensor *sensor;
  TPhHit    *hit;
  TPhSubDetector *SpectrometerArm[2];
  SpectrometerArm[0]=fDetector->GetSubDetector("SpectP");
  SpectrometerArm[1]=fDetector->GetSubDetector("SpectN");

  nhits=0;

  for (arm=0; arm<1; arm++) 
  {// loop on spectrometers

    TIter NextLayer(SpectrometerArm[arm]->GetLayerList());
    while ((layer=(TPhLayer*)NextLayer())) // && (n_missed<3))
    {// loop on layers
      label=layer->GetLabel();
      if (label<kNLayer)
      {
        if (fVerbose>=4) printf("Check hits in layer %d (%d sensors)\n",
		       layer->GetLabel(),layer->GetSensorList()->GetEntries());
        cnt=fHitCnt[label];
        if (t->GetLayerToRow(label)>=0)
        {
          flag = 1;
          TIter NextSensor(layer->GetSensorList());
          while ( (sensor=(TPhSensor *)NextSensor()) )
          {
            if (fVerbose>=4)
            if (sensor->GetHitList())
              printf("Check %d hits in sensor %s (%d missed)\n",
			 sensor->GetHitList()->GetEntries(),
			 sensor->GetChLabel(),n_missed);
            TIter NextHit((sensor->GetHitList()));
            while ((hit=(TPhHit *)NextHit()))
            {
	      if (t->IsCompatible(hit))
	      {
                nhits++; flag = 0;
	        fHitArray[label][cnt] = hit;
                if (cnt<(kNumHits-1)) cnt++;
	      }
            }
          }
        }
        fHitCnt[label] = cnt;
        n_missed += flag;
      }
    }
  }
  return nhits;
}

void TPhTrackFinder::ScanTemplates(TPhTrackSeedFinder *tsf)
{
  //
  //  initialization
  //
  
  if (fVerbose) printf("\n---- TPhTrackFinder::ScanTemplates 0 ----");
  fCandidateList->Delete();
  //fCandidateList->Clear();

  //
  // scan 
  //
  
  TPhTrackTemplate *templ;    
  TIter NextTemplate(fTemplateList);
  while( (templ=(TPhTrackTemplate *) NextTemplate()) )
  {// ---- loop over templates 0 ----

    //
    if (!templ->GetCovar()) continue; // this template has no Covar
    //                      ======== 
   
    if (tsf)
    {
    TObjArray *seeds=tsf->GetCompatibleSeedList(templ);
    int n_seeds = seeds->GetEntries();   
    //
    if (n_seeds == 0) continue; // this template is far from the pre-jet
    //                ======== 
    if (fVerbose>=2) printf("\n\t(%5.3f-%5.3f) ",
	templ->GetP(),-log(atan(templ->GetPx()/templ->GetPz())/2));
    }
    //
    // make compatible hit array and check whether it can be a good track
    //
    
    int nhits = GetCompatibleHits(templ);
    if (fVerbose>=3) printf("\n\t number of compatible hits %d",nhits);
  
    Bool_t good_track =
    (fHitCnt[0] >0 || fHitCnt[1] >0) &&
    (fHitCnt[2] >0 || fHitCnt[3] >0) &&
    (fHitCnt[4] >0 || fHitCnt[5] >0) &&
    (fHitCnt[6] >0 || fHitCnt[7] >0) &&
    (fHitCnt[8] >0 || fHitCnt[9] >0) &&
    (fHitCnt[10]>0 || fHitCnt[11]>0) &&
    (fHitCnt[12]>0 || fHitCnt[13]>0) &&
    (fHitCnt[14]>0 || fHitCnt[15]>0) ;
        
    //
    if (!good_track) continue; // this hit array can not make a good track
    //               ========
        
    //
    // recount number of compatible hits 
    //
      
    int n_hits=0;
    int i;
    for (i=0; i<kNLayer; i++)
    {
      int min_hit=-1;
      float min_dist=1000000;
      for (int j=0; j<fHitCnt[i]; j++)
      {
	TPhHit *hit = (TPhHit *) fHitArray[i][j];
	float dist = fabs(templ->GetDistXZ(hit));
	if (dist<min_dist)
      	{
          min_dist=dist;
	  min_hit=j;
	}
      }
      if (min_hit!=-1)
      {
	n_hits++;
	fHitArray[i][0]=fHitArray[i][min_hit]; 
	fHitCnt[i]=1;
      }
    }
      
    //  
    if (n_hits<fMinHits) continue; // do the next template  
    //                   ========
    
    int n_hits_array=0;
    int n_combinations=1;
    if (fVerbose>=3) printf("\n---temp %f",templ->GetP());
    for (int k=0;k<kNLayer;k++)
    {
      if (fVerbose>=3) printf("%d ",fHitCnt[k]);
      if (fHitCnt[k] >0) n_hits_array++;
      if (fHitCnt[k]==0) fHitCnt[k]=1; // dummy number for the combination
      n_combinations*=fHitCnt[k];
    }
  
    //
    // create track candidates by using of combinations of the compatible hits
    //
	      
    int il[16],n_candidates=0;
    float max_prob=0;
    TPhRecTrack *trk_prob;
  
    for (il[0] =0; il[0] <fHitCnt[0] ; il[0]++ ){
    for (il[1] =0; il[1] <fHitCnt[1] ; il[1]++ ){
    for (il[2] =0; il[2] <fHitCnt[2] ; il[2]++ ){
    for (il[3] =0; il[3] <fHitCnt[3] ; il[3]++ ){
    for (il[4] =0; il[4] <fHitCnt[4] ; il[4]++ ){
    for (il[5] =0; il[5] <fHitCnt[5] ; il[5]++ ){
    for (il[6] =0; il[6] <fHitCnt[6] ; il[6]++ ){
    for (il[7] =0; il[7] <fHitCnt[7] ; il[7]++ ){
    for (il[8] =0; il[8] <fHitCnt[8] ; il[8]++ ){
    for (il[9] =0; il[9] <fHitCnt[9] ; il[9]++ ){
    for (il[10]=0; il[10]<fHitCnt[10]; il[10]++){
    for (il[11]=0; il[11]<fHitCnt[11]; il[11]++){
    for (il[12]=0; il[12]<fHitCnt[12]; il[12]++){
    for (il[13]=0; il[13]<fHitCnt[13]; il[13]++){
    for (il[14]=0; il[14]<fHitCnt[14]; il[14]++){
    for (il[15]=0; il[15]<fHitCnt[15]; il[15]++){
      n_candidates++;
      TPhRecTrack *trk = new TPhRecTrack();
      float chi2, ndof;
      for (int l=0; l<kNLayer; l++)
      {
    	TPhHit *hit = (TPhHit *) fHitArray[l][il[l]];
        if (hit) trk->AddHit(hit);
      }
      templ->CalcChi2(trk,&chi2,&ndof);
      trk->SetChi2(chi2);
      trk->SetNDof(ndof);
      float prob=trk->CountProb(fStuff);
      if (prob>max_prob) 
      {
        max_prob=prob;
        //if (!trk_prob) delete trk_prob;
        trk_prob=trk;
      }
      else
      {
        delete trk;
      }
    }}}}}}}}}}}}}}}}
                  
    // printf("\n %d combinations: %d candidates",ncomb,ncandi);    
    if (max_prob > fMinProb)
    {
      if (fVerbose>=3) printf("K(%f)",max_prob);
      trk_prob->SetProb(max_prob);
      trk_prob->SetMass(templ->GetMass());
      trk_prob->SetCharge(templ->GetCharge());
      trk_prob->SetPx(templ->GetPx());
      trk_prob->SetPy(templ->GetPy());
      trk_prob->SetPz(templ->GetPz());
      trk_prob->SetStatus(1);
      fCandidateList->Add(trk_prob);
    }
    else
    {
      if (fVerbose>=3) printf("D(%f)",max_prob);
      //delete trk;
    }
    
  }// ---- loop over templates 1 ----

  if (fVerbose>=2) printf("\nFound %d track candidates",fCandidateList->GetEntries());
  if (fVerbose) printf("\n---- TPhTrackFinder::ScanTemplates 1 ----");
  return;
}

void TPhTrackFinder::SelectTracks(TObjArray *list)
{
  TPhRecTrack *t1;
  TPhRecTrack *t2;
  int max = list->GetEntries();
  if (fVerbose) printf("Found %d track candidates\n",max);
  int i;
  for (i=0;i<max-1;i++)
  {
    t1 = (TPhRecTrack *) list->At(i);
    for (int j=i+1;j<max;j++)
    {
      t2 = (TPhRecTrack *) list->At(j);
      if (t1->GetStatus() && t2->GetStatus())
      {
        if (DoCompete(t1,t2))
	{
          if(t1->GetProb() > t2->GetProb())
          {
            t1->SetStatus(kTRUE);
            t2->SetStatus(kFALSE);
          }
          else
          {
            t1->SetStatus(kFALSE);
            t2->SetStatus(kTRUE);
          }
	}
      }
    }
  }
  if (fVerbose) printf("Create selected track list\n");
  if (!fTrackList) fTrackList = new TObjArray;
  else fTrackList->Clear();
  
  //
  // copy only tracks with status TRUE into TrackList 
  //
  
  for (int k=0; k<max; k++)
  {
    t1 = (TPhRecTrack *) list->At(k);
    if (t1->GetStatus()) fTrackList->Add(t1);
  }
}

Bool_t TPhTrackFinder::DoCompete(TPhRecTrack *t1, TPhRecTrack *t2)
{
  TPhHit *hits1[kNLayer];
  TPhHit *hits2[kNLayer];
  TPhHit *hit;

  for (int i; i<kNLayer; i++)
  {
    hits1[i]=NULL;
    hits2[i]=NULL;
  }
  
  TIter NextHit1(t1->GetHitList());
  while ((hit=(TPhHit *)NextHit1())) { hits1[hit->GetLayer()]=hit; }

  TIter NextHit2(t2->GetHitList());
  while ((hit=(TPhHit *)NextHit2())) { hits2[hit->GetLayer()]=hit; }
  

  int cnt = 0;
  for(int j=fOuterLayer; j<kNLayer; j++)
  {
    if(hits1[j]==hits2[j] && hits1[j]!=NULL) cnt++;
  }
  
  if (cnt > fMaxSharedHits) return kTRUE;
  else return kFALSE;
}

void TPhTrackFinder::DrawCurves()
{
  int n_curves=fCurveList->GetEntries();
  for (int icurve=0;icurve<n_curves;icurve++) 
  {
    TPhTrack *curve=(TPhTrack *)fCurveList->At(icurve);
    TObjArray *points=curve->GetHitList(); 
    int n_points=points->GetEntries();
    TPolyLine3D *line=new TPolyLine3D(n_points);
    for (int ip=0;ip<n_points;ip++)
    {
      TPhHit *point=(TPhHit *)points->At(ip);
      line->SetPoint(ip,point->GetX(),point->GetY(),point->GetZ());
    }
    line->SetLineWidth(2);
    line->SetLineColor(2);
    line->SetLineStyle(2);
    line->Draw();	    
  }  
}



//
//
// H O U G H   T R A N S F O R M A T I O N 
//
//
//
//
//
//
//
//

void TPhTrackFinder::FillHit()
{
  //
  // initialization
  //
  
  //
  // call the actual hit merging algorithm
  //
  fMerge->Reset('P'); 
  fMerge->MergeSpecHits('P');
  
  //
  // copy into hit_buf
  //
  
  for (int i=0; i<kNLayer; i++)
  {
    //
    // initialization
    //
  
    fHoughHitA[i]->Clear();
    
    int index0, index1;
    int nhits = fMerge->GetSpecHitIndexes(i,&index0,&index1);
    if (nhits != 0) 
    {  
      for (int idx=index0; idx<=index1; idx++)
      {
	TPhRecHit *h=(TPhRecHit *)fMerge->GetHit(idx);
  	TPhHoughHit *hit=new TPhHoughHit(h);
	fHoughHitA[i]->Add(hit);
      }
    }
  }

  if (fVerbose)
  { 
    printf("Number of hits:"); 
    for (int j=0; j<kNLayer; j++) printf("%d-",fHoughHitA[j]->GetEntries());
    printf("\n");
  }

}

/*

  Bool_t Lseed[200];
  if (fSeedList)
  {  
    for (int ib=0;ib<200;ib++) Lseed[ib]=0;
    int n_seeds = fSeedList->GetEntries();
    for (int iseed=0;iseed<n_seeds;iseed++)
    { 
      TPhTrack *seed=(TPhTrack *)fSeedList->At(iseed);
      float x=seed->GetPx();
      float y=seed->GetPy();
      float z=seed->GetPz();
      float r=sqrt(x*x+y*y+z*z);
      float theta_s=acos(z/r);
      float eta_s=-log(tan(theta_s/2));
      float phi_s=atan2(y,x);
      int ieta0 = (int) ((eta_s-fSeedResol)*100);
      int ieta1 = (int) ((eta_s+fSeedResol)*100);
      for (int ieta=ieta0;ieta<=ieta1;ieta++)
      {
        Lseed[ieta]=1; 
      }
    }
  }

        if (fSeedList && !Lseed[(int)(eta*100)]) continue;



*/


void TPhTrackFinder::DoHough(int charge, float avg_eta, float sgm_eta, float avg_phi, float sgm_phi)
{
  
  // This routine takes so much CPU time.... ??
  
  //
  // reset hough hit
  //
  for (int ilayer=0;ilayer<kNLayer;ilayer++)
  {
    int nhit = fHoughHitA[ilayer]->GetEntries();
    for (int ihit=0;ihit<nhit;ihit++)
    {
      TPhHoughHit *h = (TPhHoughHit *)fHoughHitA[ilayer]->At(ihit);
      h->Reset(); 
    }
  }
  
  double theta=0, oop=0;
  double alpha,curve;

  for (int ipair=0; ipair<kNPairs; ipair++)  
  {// ---- loop over all layer combinations 0 ----
    
    //
    // initialization
    //
    
    //fHoughRoadA[ipair]->Clear();
    fHoughRoadA[ipair]->Delete();
    
    int ip0=k0Pairs[ipair];
    int ip1=k1Pairs[ipair];
    if (fVerbose) 
      printf("%d,%d: %d x %d hits\n",
	ip0,ip1,fHoughHitA[ip0]->GetEntries(),fHoughHitA[ip1]->GetEntries());  
  
    int nh0 = fHoughHitA[ip0]->GetEntries();
    int nh1 = fHoughHitA[ip1]->GetEntries();
    
    for (int jp0=0; jp0<nh0; jp0++)
    {
      TPhHoughHit *h0 = (TPhHoughHit *)fHoughHitA[ip0]->At(jp0);
      float x0 = h0->GetX(); 
      float y0 = h0->GetY(); float sy0=h0->GetPadSizeY()/2;
      float z0 = h0->GetZ();

      // horizontal cut (charge related)
      
      float eta0 = h0->GetEta();
      if ((eta0-avg_eta)*charge<0) continue;
      
      // vertical cut
      
      float ys0 = sqrt(x0*x0+z0*z0)*tan(avg_phi);
      float es0 = sqrt(x0*x0+z0*z0)*sgm_phi;
      if (fabs(ys0-y0)>sqrt(es0*es0+sy0*sy0)) continue;
      
      for (int jp1=0; jp1<nh1; jp1++)
      {
        TPhHoughHit *h1 = (TPhHoughHit *)fHoughHitA[ip1]->At(jp1);
        float x1 = h1->GetX(); 
        float y1 = h1->GetY(); float sy1=h1->GetPadSizeY()/2;
        float z1 = h1->GetZ(); 
     
        // horizontal cut
        float eta1 = h1->GetEta();
        if ((eta1-avg_eta)*charge<0) continue;

        // vertical cut	
	float ys1 = sqrt(x1*x1+z1*z1)*tan(avg_phi);
        float es1 = sqrt(x1*x1+z1*z1)*sgm_phi;	
	if (fabs(ys1-y1)>sqrt(es1*es1+sy1*sy1)) continue;
	
        //
	// we have two hits now... 
	//
	
        fHough->Transform(charge,ipair,z0,x0,z1,x1,&theta,&oop);      
     
        if (!theta || !oop) continue;
      
        float eta = -log(tan(theta/2));
	if (fabs(eta-avg_eta)>sgm_eta) continue;
	

	if (fVerbose>1)
	{
	  fHough->GetAngleCurve(z0,x0,z1,x1,&alpha,&curve);
	  printf("hit %d-%d (%5.3f,%5.3f) (%5.3f,%5.3f) : %5.3f %5.3f : %5.3f %5.3f\n",
	    jp0,jp1,z0,x0,z1,x1,alpha,curve,theta,oop);
        }
	
	TPhHoughRoad *hr=new TPhHoughRoad();
	hr->SetTH(theta);
	hr->SetQP(oop);
	hr->SetHit(0,h0);
	hr->SetHit(1,h1);
	fHoughRoadA[ipair]->Add(hr);
          
  	h0->AddHit(1,h1);
    	h0->AddRoad(1,hr);
        h1->AddHit(0,h0);
        h1->AddRoad(0,hr);
      }
    }
     
    if (fVerbose) 
      printf("pair_%d has %d Hough Roads\n",
	ipair,fHoughRoadA[ipair]->GetEntries());	
  }
  
}

void TPhTrackFinder::FindTrackCandidate(TPhTrack *seed,int direction)
{
  const int LayerList[5]={8,9,10,12,14};
  
  for (int i=0; i<16;i++) fTrack[i]->Clear(); 

  //
  // create a track 
  //
  
  TObjArray *shs = seed->GetHitList();
  int nsh = shs->GetEntries();
  
  int I0 =8 ; 
  int nI0= fHoughHitA[I0]->GetEntries();
  for (int ih0=0; ih0<nI0; ih0++)
  {
    TPhHoughHit *h=(TPhHoughHit *)fHoughHitA[I0]->At(ih0);
    TPhTrack *t0 = new TPhTrack();
    t0->SetPx(seed->GetPx());
    t0->SetPy(seed->GetPy());
    t0->SetPz(seed->GetPz());
    for (int ish=0;ish<nsh;ish++)
    {
      TPhHit *sh = (TPhHit *)shs->At(ish);
      t0->AddHit(sh);
    }
    t0->AddHit(h);
    fTrack[I0]->Add(t0);
  } 
  if (fVerbose) printf("How many initial tracks? %d\n",fTrack[I0]->GetEntries());
   
  //
  //
  //
  
  for (int idx=0; idx<4; idx++)
  {
    int il = LayerList[idx];
    if (fVerbose>1) printf("Running layer %d:\n",il);

    int nt = fTrack[il]->GetEntries();
    for (int it=0;it<nt;it++)
    { 
      TPhTrack *ht = (TPhTrack *)fTrack[il]->At(it);
      float Pht = ht->GetP(); 
      float Tht = acos(ht->GetPz()/Pht);
      TObjArray *ths = ht->GetHitList(); 
      int nth = ths->GetEntries();

      if (fVerbose>1) printf("\ttrk %d (%f %f)\n",it,Pht,Tht);
      
      TPhHoughHit *hh=(TPhHoughHit *)ths->At(nth-1);      
      
      //int nh = fHoughHitA[il]->GetEntries();
      //for (int ih=0; ih<nh; ih++)
      //{
      //  TPhHoughHit *hh = (TPhHoughHit *)fHoughHitA[il]->At(ih);
        TObjArray *hrs = hh->GetRoadList(1);
        int nr = hrs->GetEntries();
        for (int ir=0;ir<nr;ir++)
        {
          TPhHoughRoad *hr = (TPhHoughRoad *)hrs->At(ir);
          float Phr = 1/hr->GetQP();
          float Thr = hr->GetTH();
        
          if (fVerbose>1) printf("\t %f %f -",Phr,Thr);	  
          if (
	       (idx==0 || fabs(Phr-Pht)/Pht<fErrorQP[idx]) && 	  
                fabs(Thr-Tht)<fErrorTH[idx]
	     )
          {
	    if (fVerbose>1) printf("*");
	    TPhTrack *t = new TPhTrack();
	    t->SetPx(Phr*sin(Thr));
	    t->SetPy(0);
	    t->SetPz(Phr*cos(Thr));
	    for (int ith=0; ith<nth; ith++)
	    {
	      TPhHoughHit *h=(TPhHoughHit *)ths->At(ith);
	      t->AddHit(h);   
	    }
	    t->AddHit(hr->GetHit(1));
	    fTrack[LayerList[idx+1]]->Add(t);
	  }
	}
      //}printf("\n");
    }if (fVerbose>1) printf("\n");
  }
  
  //
  //
  //
  
  int ncurve=fTrack[14]->GetEntries();
  for (int icurve=0; icurve<ncurve; icurve++)
  {
    TPhTrack *curve=(TPhTrack *)fTrack[14]->At(icurve);
    fCurveList->Add(curve);
  }  
}
  

void TPhTrackFinder::FindCurve(int charge,int direction)
{
  TObjArray *hrs;
 
  for (int il=0; il<16;il++) fTrack[il]->Clear();

  int I0,I1,IP0,IP1,IPD,IS;
  if (direction>0)
  {
    I0= 8; I1=14; IP0=0; IP1=3; IPD= 1; IS=1; 
  }
  else
  {  
    I0=14; I1= 8; IP0=3; IP1=0; IPD=-1; IS=0;
  }
  
  int nI0= fHoughHitA[I0]->GetEntries();
  for (int i=0; i<nI0; i++)
  {
    TPhRecTrack *t = new TPhRecTrack();
    TPhHoughHit *h=(TPhHoughHit *)fHoughHitA[I0]->At(i);
    float x=h->GetX(), y=h->GetY(), z=h->GetZ();
    float r=sqrt(x*x+y*y+z*z); if(r==0) printf("?");
    t->SetPx(x/r);
    t->SetPy(y/r);
    t->SetPz(z/r);
    t->SetXProd(0);
    t->SetYProd(0);
    t->SetZProd(0);
    t->AddHit(h);
    fTrack[I0]->Add(t);
  }
  int nt=fTrack[I0]->GetEntries(); 
  if (fVerbose) printf("How many initial tracks? %d\n",nt);
    
  for (int ipair=IP0; ipair>=0 && ipair<kNPairs; ipair+=IPD)
  {
    int ip0,ip1;
    if (direction>0)
    {
      ip0 = k0Pairs[ipair];
      ip1 = k1Pairs[ipair];
    }
    else
    {
      ip0 = k1Pairs[ipair];
      ip1 = k0Pairs[ipair];
    }
    for (int t0=0; t0<fTrack[ip0]->GetEntries(); t0++)
    {
      TPhTrack *t= (TPhTrack *) fTrack[ip0]->At(t0); 
      float t_p=t->GetP();
      float t_t=atan(t->GetPx()/t->GetPz());
      
      TPhHoughHit *lh=(TPhHoughHit *)t->GetHitList()->At(t->GetNHits()-1);
      hrs=lh->GetRoadList(IS);
      if (hrs)
      {	
        for (int ir=0; ir<hrs->GetEntries(); ir++)
        {
          TPhHoughRoad *r = (TPhHoughRoad *)hrs->At(ir);
          float r_p = 1/r->GetQP();
	  float r_t = r->GetTH();
	  
	  if (ipair==IP0 ||
	      (fabs(r_p-t_p)/t_p<fErrorQP[ipair] &&
	       fabs(r_t-t_t)<fErrorTH[ipair]))
	  { 
	    float px = r_p*sin(r_t);
	    float py = 0;
	    float pz = r_p*cos(r_t);
	    TPhHoughHit *oh = (TPhHoughHit *)r->GetHit(IS);
	    TPhTrack *t1 = new TPhTrack();
            t1->SetPx(px);
            t1->SetPy(py);
            t1->SetPz(pz);
	    for (int ihb=0; ihb<t->GetNHits(); ihb++)
	    {
	      TPhHoughHit *h0=(TPhHoughHit *)t->GetHitList()->At(ihb);
	      t1->AddHit(h0);
	    }  
	    t1->AddHit(oh);
            fTrack[ip1]->Add(t1);
          }	    
	}
      } 
    }
    int nt1=fTrack[ip1]->GetEntries(); 
    if (fVerbose) printf("number of tracks after run layer %d = %d\n",ip1,nt1);
  }
  
  int ncurve=fTrack[I1]->GetEntries();
  for (int icurve=0; icurve<ncurve; icurve++)
  {
    TPhTrack *curve=(TPhTrack *)fTrack[I1]->At(icurve);
    fCurveList->Add(curve);
  }  
  
}


void TPhTrackFinder::RunMulguisin()
{
  int n_roads=0;
  for (int i=0; i<kNPairs; i++) n_roads += fHoughRoadA[i]->GetEntries();
   
  int    InumC=0;
  int   *IdetC=new int[n_roads];
  float *RentC=new float[n_roads];
  float *RetaC=new float[n_roads];
  float *RphiC=new float[n_roads];
	    
  for (int ip=0; ip<kNPairs; ip++)
  {
    int nr=fHoughRoadA[ip]->GetEntries();
    for (int ir=0; ir<nr; ir++)
    {
      TPhHoughRoad *road = (TPhHoughRoad *)fHoughRoadA[ip]->At(ir);  
			              
      float eta=road->GetTH();
      float phi=road->GetQP();
			                
      IdetC[InumC]=1<<ip;
      RentC[InumC]=1/(float)ip;
      RetaC[InumC]=eta;
      RphiC[InumC]=phi;
      InumC++;
				              
    }
  }
   
  TPhMulguisin *mgs = new TPhMulguisin();
  mgs->SetFrame(0);
  mgs->SetMethod(0);
  mgs->SetMode(0);
  mgs->SetRange(0.02,0.05);
  
  mgs->Clustering(InumC,IdetC,RentC,RetaC,RphiC);
  
  int    InumG=mgs->GetInumG();
  int   *IdetG=mgs->GetIdetG();
  float *RentG=mgs->GetRentG();
  float *RetaG=mgs->GetRetaG();
  float *RphiG=mgs->GetRphiG();
	          
  printf("Number of ghosts = %d\n",InumG);
   	      
  //
  //
  //
  
  int    InumJ = 0;
  int   *IdetJ = new int[InumG];
  float *RentJ = new float[InumG];
  float *RetaJ = new float[InumG];
  float *RphiJ = new float[InumG];
	                
  for (int ig=0; ig<InumG; ig++)
  {
    int id=IdetG[ig];
    printf("\tguisin %d : %d\n",ig,id);
    Bool_t good_jet=1;
    //
    // check whether there exists at least one hit on each frame
    //
    for (int j=1; j<=3; j+=2)
    {
      good_jet = good_jet&&(id&(1<<j-1)||id&(1<<j));
    }
    //
    // check whether there are at least 3 hits.
    //
    int nhits=0;
    for (int i=0; i<=3; i++) if (id&(1<<i)) nhits++;
    good_jet = good_jet&&(nhits>=3); 
    //
    // 
    //
    if (good_jet)
    {
      IdetJ[InumJ]=IdetG[ig];
      RentJ[InumJ]=RentG[ig];
      RetaJ[InumJ]=RetaG[ig];
      RphiJ[InumJ]=RphiG[ig];
      InumJ++;
      
      float theta = 2*atan(exp(-RetaG[ig]));
      float phi = RphiG[ig];
    /*  
      TPhRecTrack *seed = new TPhRecTrack();
      
      seed->SetPx(sin(theta)*cos(phi));
      seed->SetPy(sin(theta)*sin(phi));
      seed->SetPz(cos(theta));
      seed->SetXProd(0);
      seed->SetYProd(0);
      seed->SetZProd(0);
      fTrackSeedList->Add(seed);*/
    }
  }
  
  printf("Number of curves = %d\n",InumJ);
    
  
  //  
  // release memory
  //
  
  delete [] IdetC;
  delete [] RentC;
  delete [] RetaC;
  delete [] RphiC;
  
  delete [] IdetJ;
  delete [] RentJ;
  delete [] RetaJ;
  delete [] RphiJ;
  
}

