#include "TPhStuff.h"

#include <stdlib.h>
#include <math.h>

#include "TRandom.h"

ClassImp(TPhStuff)
//////////////////////////////////////////////////////////////////////////
//
// A class containing a mixed bag of things. Mainly routines from CERNLIB
// and C++ converted routines from Numerical Receipes
//
// $Id: TPhStuff.cxx,v 1.3 1999/02/15 21:59:15 baker Exp $
//
//////////////////////////////////////////////////////////////////////////

#define ITMAX 100
#define EPS 3.0e-7
#define TINY 1.0e-20;


void nrerror(char *error_text)
{

	fprintf(stderr,"Numerical Recipes run-time error...\n");
	fprintf(stderr,"%s\n",error_text);
	fprintf(stderr,"...now exiting to system...\n");

	exit(-1);
}



void TPhStuff::gser(Float_t *gamser,Float_t a,Float_t x,Float_t *gln)
{
        int n;
        float sum,del,ap;


        *gln=gammln(a);
        if (x <= 0.0) {
                if (x < 0.0) printf("x less than 0 in routine GSER");
                *gamser=0.0;
                return;
        } else {
                ap=a;
                del=sum=1.0/a;
                for (n=1;n<=ITMAX;n++) {
                        ap += 1.0;
                        del *= x/ap;
                        sum += del;
                        if (fabs(del) < fabs(sum)*EPS) {
                                *gamser=sum*exp(-x+a*log(x)-(*gln));
                                return;
                        }
                }
		// MDB 15FEB99 Suppress error - too verbose
		//   printf("a too large, ITMAX too small in routine GSER");
                return;
        }
}

Float_t TPhStuff::gammln(Float_t xx)
{
        double x,tmp,ser;
        static double cof[6]={76.18009173,-86.50532033,24.01409822,
                -1.231739516,0.120858003e-2,-0.536382e-5};
        int j;

        x=xx-1.0;
        tmp=x+5.5;
        tmp -= (x+0.5)*log(tmp);
        ser=1.0;
        for (j=0;j<=5;j++) {
                x += 1.0;
                ser += cof[j]/x;
        }
        return -tmp+log(2.50662827465*ser);
}

void TPhStuff::gcf(Float_t *gammcf,Float_t a,Float_t x,Float_t *gln)
{

        int n;
        float gold=0.0,g,fac=1.0,b1=1.0;
        float b0=0.0,anf,ana,an,a1,a0=1.0;


        *gln=gammln(a);
        a1=x;
        for (n=1;n<=ITMAX;n++) {
                an=(float) n;
                ana=an-a;
                a0=(a1+a0*ana)*fac;
                b0=(b1+b0*ana)*fac;
                anf=an*fac;
                a1=x*a0+anf*a1;
                b1=x*b0+anf*b1;
                if (a1) {
                        fac=1.0/a1;
                        g=b1*fac;
                        if (fabs((g-gold)/g) < EPS) {
                                *gammcf=exp(-x+a*log(x)-(*gln))*g;
                                return;
                        }
                        gold=g;
                }
        }
	// MDB 15FEB99 Suppress error - too verbose
	//        printf("a too large, ITMAX too small in routine GCF");
}

Float_t TPhStuff::gammq(Float_t a,Float_t x)
{
        float gamser,gammcf,gln;

        if (x < 0.0 || a <= 0.0) printf("Invalid arguments in routine GAMMQ\n");
        if (x < (a+1.0)) {
                gser(&gamser,a,x,&gln);
                return 1.0-gamser;
        } else {
                gcf(&gammcf,a,x,&gln);
                return gammcf;
        }
}

Float_t TPhStuff::GetProb(Float_t chi2, Float_t ndof)
{

        if(chi2 < 0 || chi2 > 1000.0)
        {
                return 0;
        }
        else if(ndof <= 0)
        {
                return 0;
        }
        else
        {
                return gammq(0.5*ndof,0.5*chi2);
        }
}


void ludcmp(Double_t **a,Int_t n,Int_t *indx,Double_t *d,Double_t *vv)
{
	int i,imax,j,k;
	Double_t big,dum,sum,temp;

	*d=1.0;

	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) nrerror("Singular matrix in routine LUDCMP");
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) a[i][j] *= dum;
		}
	}

}

void lubksb(Double_t **a,int n,int *indx,Double_t *b)
{
	int i,ii=0,ip,j;
	Double_t sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}

Int_t TPhStuff::Polint(Float_t xarr[], Float_t yarr[], Int_t n, 
		       Float_t x, Float_t *y, Float_t *dy)
{
  // Given arrays xarr[0..n-1] and yarr[0..n-1], and given a value x, 
  // this routine returns a value y, and
  // an error estimate dy. If P(x) is the polynomial of degree N
  //
  // Routine comes from NR and was adjusted to start from index 0
  // Return:        0 on success
  //               -1 on failure

  Int_t i, m, ns = 1;
  Float_t den, dif, dift, ho, hp, w;
  Float_t *c, *d;

  Float_t  *xa = xarr-1;
  Float_t  *ya = yarr-1;

  dif = fabs(x-xa[1]);
  c = new Float_t[n+1];
  d = new Float_t[n+1];

  for(i = 1; i <= n; i++) { 
    // Here we find the index ns of the closest table entry,
    if ((dift = fabs(x-xa[i])) < dif) {
      ns = i;
      dif = dift;
    }
    // and initialize the tableau of c's and d's.
    c[i] = ya[i]; 
    d[i] = ya[i];
  }
  *y = ya[ns--]; 
  // This is the initial approximation to y.
  for (m = 1; m < n; m++) { 
    // For each column of the tableau,
    for (i = 1;i <= n-m; i++) { 
      // we loop over the current c's and d's and update them
      ho = xa[i]-x;
      hp = xa[i+m]-x;
      w = c[i+1]-d[i];
      // This error can occur only if two input xa's are 
      // (to within roundoff) identical.
      if ((den = ho-hp) == 0.0) {
	printf("Polint: Two input points on grid xa are identical\n");
	return -1;
      }
      den = w/den;
      // Here the c's and d's are updated.
      d[i] = hp*den; 
      c[i] = ho*den;
    }
    *y += (*dy=(2*ns < (n-m) ? c[ns+1] : d[ns--]));
    // After each column in the tableau is completed, we decide 
    // which correction, c or d, we want to add to our accumulating 
    // value of y, i.e., which path to take through the
    // tableau|forking up or down. We do this in such a way as to take 
    // the most \straight line" route through the tableau to its 
    // apex, updating ns accordingly to keep track of where we are. 
    // This route keeps the partial approximations centered (insofar 
    // as possible) on the target x. The last dy added is thus the error 
    // indication.
  }
  delete[] c;
  delete[] d;
  return 0; // all ok
}

Int_t TPhStuff::Locate(Float_t xx_arr[], UInt_t n, Float_t x)
{
  // Given an array xx_arr[0..n-1], and given a value x, returns a value 
  // j such that x is between xx[j] and xx[j+1]. xx must be monotonic, 
  // either increasing or decreasing.
  // 
  // Routine comes from NR and was adjusted to start from index 0
  // Return:   -1 or n-1 when x is out of range

  Int_t ju, jm, jl;
  Int_t ascnd;

  Float_t *xx = xx_arr - 1;

  jl = 0; // Initialize lower
  ju = n+1; // and upper limits.
  ascnd = (xx[n] >= xx[1]);
  while (ju-jl > 1) { // If we are not yet done,
    jm = (ju + jl) >> 1;  // compute a midpoint,
    // and replace either the lower limit
    // or the upper limit, as appropriate.
    if ((x >= xx[jm]) == ascnd)
      jl = jm; 
    else
      ju = jm; 
  } // Repeat until the test condition is satised.
  if (x == xx[1]) 
    return 0; // Then set the output
  else if(x == xx[n]) 
    return n-2;
  else 
    return jl-1;
} 

void TPhStuff::InterPol1D(Float_t xarr[], Float_t yarr[], Int_t n, Int_t dim, 
			  Float_t x, Float_t *y, Float_t *dy)
{
  // Interpolation of order dim-1 in 1D

  Int_t j, k;
  j = Locate(xarr, n, x);
  if(j == -1 || j == n-1) {
    //    printf("Out of range j = %d\n", j);
    return;
  }
  k = TMath::Min(TMath::Max(j-(dim-1)/2, 1), n+1-dim);
  Polint(&xarr[k-1], &yarr[k-1], dim, x, y, dy);
}

void TPhStuff::InterPol2D(Float_t xarr1[], Float_t xarr2[], Float_t *yarr[], 
			  Int_t m, Int_t n, Int_t dim, Float_t x1, Float_t x2, 
			  Float_t *y, Float_t *dy)
{
  // Interpolation of order dim-1 in 2D
  
  Int_t j;
  Float_t *ytmp;

  ytmp = new Float_t[m + 1];
  for(j = 0; j <= m; j++)
    InterPol1D(xarr2, yarr[j], n, dim, x2, &ytmp[j], dy);
  InterPol1D(xarr1, ytmp, m, dim, x1, y, dy);
  delete[] ytmp;
}

void TPhStuff::InvCovar(Float_t *covar, Int_t N)
{
  // This routine is leaking memory like crazy!!! none of the allocated
  // memory is being freeed!

    Double_t **mat;
    Double_t **out;
    Double_t *col;   
    Double_t *vv;   
    Double_t d;
    Int_t i,j;
    Int_t *indx;

    mat  = (Double_t **) malloc(8*(N+1));
    out  = (Double_t **) malloc(8*(N+1));
    vv   = (Double_t *)  malloc(8*(N+1));
    col  = (Double_t *)  malloc(8*(N+1));
    indx = (Int_t *)    malloc(8*(N+1));
        
    for(i=0;i<=N;i++)
    {
        mat[i] = (Double_t *) malloc(8*(N+1));
        out[i] = (Double_t *) malloc(8*(N+1));
	//        printf("%p\n",mat[i]);
    } 
    printf("In: \n");
    for(i=1;i<=N;i++)
    {
        for(j=1;j<=N;j++)
        {
            out[i][j] = 0;
            mat[i][j] = covar[(i-1)*N+(j-1)];
            printf("%.4f ",mat[i][j]);
            
        }
        printf("\n");
    }
    printf("\n");
    ludcmp(mat,N,indx,&d,vv);
  
    printf("call lbksb\n");
  
    for(j=1;j<=N;j++) {
       for(i=1;i<=N;i++) col[i] = 0.0;
       col[j] = 1.0;
       lubksb(mat,N,indx,col);
       for(i=1;i<=N;i++) out[i][j] = col[i];
    }
    printf("Out: \n");
    for(i=1;i<=N;i++)
    {
        for(j=1;j<=N;j++)
        {
            printf("%.5f ",out[i][j]);
            covar[(i-1)*N+(j-1)] = out[i][j];
        }
        printf("\n");
    }
  
    printf("Ende \n");
}
